#!/usr/bin/perl
#------------------------------------------------------------------------------
# Web server logfile analyzer partly based on 'AdvancedWebStatistics.pl'
# Works only from command line. You must use this script as often as
# necessary from your scheduler to update your statistics and
# to generate report results.
# For additional setup instructions see AWScloudMap documentation at:
# http://cloudMap.calcResult.com
#------------------------------------------------------------------------------
# Author: Mike Brockington - 15/04/2008 14:31:12
#            Last revision - 13/02/2010 23:12:40
# Based on AWStats revision 1.892; by eldy; dated: 2007/07/07 11:00:05
# Also contains section from  title-grabber.pl
#     (Feedback/suggestions to:Simon Waight: simonjw@users.sourceforge.net)
#
# Note: If you want to retrieve document titles over SSL you must have OpenSSL and
#       the Net::SSL(eay) Perl Module available.  This code will check that SSL is
#		supported before attempting to retrieve via it.
#
#
# See: http://awstats.sourceforge.net
#      http://cloudMap.calcResult.com
#
# Not sure how accurate the next line is. Currently running under ActivePERL 5.10.0 MJB 19/02/2010
require 5.005;




## All user-modifiable configuration should be placed in a file,
## in the same folder as this program, called 'My Config'.cloudMap.conf
## For example: basicConfig.cloudMap.conf or thisSiteName.cloudMap.conf




#$|=1;
#use warnings;		        # Must be used in test mode only. This reduces process speed a little,
#use diagnostics;	        # Must be used in test mode only. This reduces process speed a lot.
use strict;no strict "refs";
use Time::Local;	        # use Time::Local 'timelocal_nocheck' is faster but not supported by all Time::Local modules
use Socket;
use LWP::UserAgent;         # Needed for the page-title retrieval section
use File::Find;
use File::Basename;
use File::Temp ();
use Archive::Extract ();    # Only needed if log files are kept in a compressed format



#------------------------------------------------------------------------------
# Defines
#------------------------------------------------------------------------------
use vars qw/ $REVISION $VERSION /;
$REVISION='18';
$VERSION="1 (build $REVISION)";

# ----- Constants -----
use vars qw/
$MAX_RANK
$MIN_RANK
$BASE_PERCENTAGE
$FONT_INCREASE
$DEBUG_FORCED $NB_OF_LINES_FOR_BENCHMARK
$LIMIT_FLUSH $NOT_SORTED_RECORD_TOLERANCE
$lastyearbeforeupdate $lastmonthbeforeupdate $lastdaybeforeupdate $lasthourbeforeupdate $lastdatebeforeupdate
/;
$MIN_RANK=-1;                        # Minimum ranking for a URL before it is displayed on the cloudMap.
                                     # Value range is from 0 -> $MAX_RANK
                                     # Helps to weed out miss-spelled URL's and/or reduce the number of URL's displayed in the cloud map.
$MAX_RANK=12;                        # Maximum 'Level' that a cloudMap URL can reach (You will need one CSS class for each level. MJB)
                                     # NB: Count starts at zero, so if MAX_RANK=10 then you will need 11 CSS classes
                                     # If you use the  -pluginmode  option then these classes will NOT be generated for you;
                                     # TIP: Consider running once without  -pluginmode  and copying the CSS block out of the generated file
$BASE_PERCENTAGE = 50;               # Minimum font size for a cloudMap URL, in %
$FONT_INCREASE   = 15;               # Increase above the BASE_PERCENTAGE for each ranking level (Integer)
$DEBUG_FORCED=0;				 	 # Force debug level to log lesser level into debug.log file (Keep this value to 0)
$NB_OF_LINES_FOR_BENCHMARK=8192;	 # Benchmark info are printing every NB_OF_LINES_FOR_BENCHMARK lines (Must be a power of 2)
$LIMIT_FLUSH=5000;					 # Nb of records in data arrays after how we need to flush data on disk
$NOT_SORTED_RECORD_TOLERANCE=20000;	 # Lapse of time to accept a record if not in correct order. 20000 = 2 hour (Default = 20000)
# ----- Running variables -----
use vars qw/
$DIR $PROG $Extension
$configFilePrefix
%MonthNum
$regtruncanchor
$regtruncurl
$regext
$regdefault
$Debug $ShowSteps
$DebugResetDone
$UpdateFor $HeaderHTMLSent
$LastLine $LastLineNumber $LastLineOffset $LastLineChecksum $LastUpdate
$lowerval
$PluginMode
$SKIP_FOLDERS
$TotalPages
$displayCount
$FileConfig $FileSuffix $Host $YearRequired $MonthRequired $DayRequired $HourRequired
$QueryString $SiteConfig $PageCode $PageDir $PerlParsingFormat $retrievalUserAgent
$pos_vh $pos_host $pos_logname $pos_date $pos_tz $pos_method $pos_url $pos_code $pos_size
$pos_referer $pos_agent $pos_query $pos_gzipin $pos_gzipout $pos_compratio $pos_timetaken
$pos_cluster $pos_hostr @pos_extra @RemoveFromTitle
$TotalDifferentPages
$regget $regsent
$lastprocesseddate $lastprocessedyear $lastprocessedmonth $lastprocessedday $lastprocessedhour
/;
$DIR=$PROG=$Extension='';
$configFilePrefix = 'cloudMap';
%MonthNum = ("Jan","01","jan","01","Feb","02","feb","02","Mar","03","mar","03","Apr","04","apr","04","May","05","may","05","Jun","06","jun","06","Jul","07","jul","07","Aug","08","aug","08","Sep","09","sep","09","Oct","10","oct","10","Nov","11","nov","11","Dec","12","dec","12");	# MonthNum must be in english because used to translate log date in apache log files
$regtruncanchor = "";
$regtruncurl = "";
$regext = "";
$regdefault = "";
$Debug = 0;
$ShowSteps = 0;
$DebugResetDone = 0;
$UpdateFor = $HeaderHTMLSent = 0;
$LastLine = $LastLineNumber = $LastLineOffset = $LastLineChecksum = $LastUpdate = 0;
$lowerval = 0;
$PluginMode = 0;
$SKIP_FOLDERS = "";
$displayCount = 0;
$TotalPages = 0;
$TotalDifferentPages = 0;
$regget = $regsent = "";
$lastprocesseddate = $lastprocessedyear = $lastprocessedmonth = $lastprocessedday = $lastprocessedhour = "";
($FileConfig, $FileSuffix, $Host, $YearRequired, $MonthRequired, $DayRequired, $HourRequired,
$QueryString, $SiteConfig, $PageCode, $PageDir, $PerlParsingFormat, $retrievalUserAgent)=
('','','','','','','','','','','','','');
# ----- Plugins variable -----
use vars qw/ $PluginDir $AtLeastOneSectionPlugin /;
$PluginDir='';
$AtLeastOneSectionPlugin=0;
# ----- Time vars -----
use vars qw/
$starttime
$nowtime $tomorrowtime
$nowweekofmonth $nowweekofyear $nowdaymod $nowsmallyear
$nowsec $nowmin $nowhour $nowday $nowmonth $nowyear $nowwday $nowyday $nowns
$StartSeconds $StartMicroseconds
/;
$StartSeconds=$StartMicroseconds=0;
# ----- Variables for config file reading -----
use vars qw/
$FoundNotPageList
/;
$FoundNotPageList=0;
# ----- Config file variables -----
use vars qw/
$StaticExt
$Lang
$MaxRowsInHTMLOutput
$MaxLengthOfShownURL
%BarPng
$BuildReportFormat
$BuildHistoryFormat
$ExtraTrackedRowsLimit
$DatabaseBreak
$SectionsToBeSaved
/;
$StaticExt='html';
$Lang='auto';
$SectionsToBeSaved='all';
$MaxRowsInHTMLOutput=1000;
$MaxLengthOfShownURL=64;            # Characters - default=64 (which sounds a lot, but really isn't)
%BarPng=('vv'=>'vv.png','vu'=>'vu.png','hu'=>'hu.png','vp'=>'vp.png','hp'=>'hp.png',
'he'=>'he.png','hx'=>'hx.png','vh'=>'vh.png','hh'=>'hh.png','vk'=>'vk.png','hk'=>'hk.png');
$BuildReportFormat='html';
$BuildHistoryFormat='text';
$ExtraTrackedRowsLimit=500;
$DatabaseBreak='month';
use vars qw/
$AllowToUpdateStatsFromBrowser $EnableLockForUpdate
$CreateDirDataIfNotExists $KeepBackupOfHistoricFiles $RecursiveDirs
$NbOfLinesParsed $NbOfLinesDropped $NbOfLinesCorrupted $NbOfOldLines $NbOfNewLines
$NbOfLinesShowsteps $NewLinePhase $NbOfLinesForCorruptedLog $PurgeLogFile $ArchiveLogRecords
$ShowDropped $ShowCorrupted $ShowUnknownOrigin
$ShowAuthenticatedUsers $ShowFileSizesStats $ShowScreenSizeStats
$ShowEMailSenders $ShowEMailReceivers $ShowWormsStats $ShowClusterStats
$IncludeInternalLinksInOriginSection
$UpdateStats $URLNotCaseSensitive $URLWithQuery
/;
($AllowToUpdateStatsFromBrowser, $EnableLockForUpdate,
$CreateDirDataIfNotExists, $KeepBackupOfHistoricFiles, $RecursiveDirs,
$NbOfLinesParsed, $NbOfLinesDropped, $NbOfLinesCorrupted, $NbOfOldLines, $NbOfNewLines,
$NbOfLinesShowsteps, $NewLinePhase, $NbOfLinesForCorruptedLog, $PurgeLogFile, $ArchiveLogRecords,
$ShowDropped, $ShowCorrupted, $ShowUnknownOrigin,
$ShowAuthenticatedUsers, $ShowFileSizesStats, $ShowScreenSizeStats,
$ShowEMailSenders, $ShowEMailReceivers, $ShowWormsStats, $ShowClusterStats,
$IncludeInternalLinksInOriginSection,
$UpdateStats, $URLNotCaseSensitive, $URLWithQuery,
)=
(0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0);
use vars qw/
$FirstDayOfWeek $KeyWordsNotSensitive $SaveDatabaseFilesWithPermissionsForEveryone
$WarningMessages $ShowLinksOnUrl
$ShowMonthStats $ShowDaysOfMonthStats $ShowDaysOfWeekStats
$ShowHoursStats $ShowDomainsStats $ShowHostsStats
$ShowRobotsStats $ShowSessionsStats $ShowPagesStats $ShowFileTypesStats
$ShowOSStats $ShowBrowsersStats $ShowOriginStats
$ShowKeyphrasesStats $ShowKeywordsStats $ShowMiscStats $ShowHTTPErrorsStats
/;
(
$FirstDayOfWeek, $KeyWordsNotSensitive, $SaveDatabaseFilesWithPermissionsForEveryone,
$WarningMessages, $ShowLinksOnUrl,
$ShowMonthStats, $ShowDaysOfMonthStats, $ShowDaysOfWeekStats,
$ShowHoursStats, $ShowDomainsStats, $ShowHostsStats,
$ShowRobotsStats, $ShowSessionsStats, $ShowPagesStats, $ShowFileTypesStats,
$ShowOSStats, $ShowBrowsersStats, $ShowOriginStats,
$ShowKeyphrasesStats, $ShowKeywordsStats, $ShowMiscStats, $ShowHTTPErrorsStats,
)=
(1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1);
use vars qw/
$LevelForRobotsDetection $LevelForWormsDetection $LevelForBrowsersDetection $LevelForOSDetection $LevelForRefererAnalyze
$LevelForFileTypesDetection $LevelForSearchEnginesDetection $LevelForKeywordsDetection
/;
(
$LevelForRobotsDetection, $LevelForWormsDetection, $LevelForBrowsersDetection, $LevelForOSDetection, $LevelForRefererAnalyze,
$LevelForFileTypesDetection, $LevelForSearchEnginesDetection, $LevelForKeywordsDetection)=
(2,0,2,2,2,2,2,2);
use vars qw/
$DirLock $DirConfig $DirData $DirLang $ArchiveFileName
$LinksToWhoIs $LinksToIPWhoIs
$LogFile $LogType $LogFormat $LogSeparator $Logo $LogoLink $StyleSheet $SiteDomain $TestDomain
$UseHTTPSLinkForUrl $URLQuerySeparators $URLWithAnchor $ErrorMessages $ShowFlagLinks
/;
($DirLock, $DirConfig, $DirData, $DirLang, $ArchiveFileName,
$LinksToWhoIs, $LinksToIPWhoIs,
$LogFile, $LogType, $LogFormat, $LogSeparator, $Logo, $LogoLink, $StyleSheet, $SiteDomain, $TestDomain,
$UseHTTPSLinkForUrl, $URLQuerySeparators, $URLWithAnchor, $ErrorMessages, $ShowFlagLinks)=
('','','','','','','','','','','','','','','','','','','','','','');
use vars qw/
$color_Background $color_TableBG $color_TableBGRowTitle
$color_TableBGTitle $color_TableBorder $color_TableRowTitle $color_TableTitle
$color_text $color_textpercent $color_titletext $color_weekend $color_link $color_hover $color_other
$color_h $color_k $color_p $color_e $color_x $color_s $color_u $color_v
/;
($color_Background, $color_TableBG, $color_TableBGRowTitle,
$color_TableBGTitle, $color_TableBorder, $color_TableRowTitle, $color_TableTitle,
$color_text, $color_textpercent, $color_titletext, $color_weekend, $color_link, $color_hover, $color_other,
$color_h, $color_k, $color_p, $color_e, $color_x, $color_s, $color_u, $color_v)=
('','','','','','','','','','','','','','','','','','','','','','');
# ---------- Init arrays --------
use vars qw/
@RobotsSearchIDOrder_list1 @RobotsSearchIDOrder_list2 @RobotsSearchIDOrder_listgen
@SearchEnginesSearchIDOrder_list1 @SearchEnginesSearchIDOrder_list2 @SearchEnginesSearchIDOrder_listgen
@BrowsersSearchIDOrder @OSSearchIDOrder @WordsToExtractSearchUrl @WordsToCleanSearchUrl
@WormsSearchIDOrder
@RobotsSearchIDOrder @SearchEnginesSearchIDOrder
@_from_p @_from_h
@_time_p @_time_h @_time_k @_time_nv_p @_time_nv_h @_time_nv_k
@DOWIndex @fieldlib @keylist
/;
@RobotsSearchIDOrder = @SearchEnginesSearchIDOrder = ();
@_from_p = @_from_h = ();
@_time_p = @_time_h = @_time_k = @_time_nv_p = @_time_nv_h = @_time_nv_k = ();
@DOWIndex = @fieldlib = @keylist = ();
use vars qw/
@MiscListOrder %MiscListCalc
%OSFamily %BrowsersFamily @SessionsRange %SessionsAverage
%LanguageCodeMap
@HostAliases
@DefaultFile
@SkipHosts @SkipUserAgents @SkipFiles @SkipExactFiles @SkipReferrers @NotPageFiles
@OnlyHosts @OnlyUserAgents @OnlyFiles
@URLWithQueryWithOnly @URLWithQueryWithout
@ExtraName @ExtraCondition @ExtraStatTypes @MaxNbOfExtra @MinHitExtra
@ExtraFirstColumnTitle @ExtraFirstColumnValues @ExtraFirstColumnFunction @ExtraFirstColumnFormat
@ExtraCodeFilter @ExtraConditionType @ExtraConditionTypeVal
@ExtraFirstColumnValuesType @ExtraFirstColumnValuesTypeVal
@ExtraAddAverageRow @ExtraAddSumRow
/;
@MiscListOrder=('AddToFavourites','JavascriptDisabled','JavaEnabled','DirectorSupport','FlashSupport','RealPlayerSupport','QuickTimeSupport','WindowsMediaPlayerSupport','PDFSupport');
%MiscListCalc=('TotalMisc'=>'','AddToFavourites'=>'u','JavascriptDisabled'=>'hm','JavaEnabled'=>'hm','DirectorSupport'=>'hm','FlashSupport'=>'hm','RealPlayerSupport'=>'hm','QuickTimeSupport'=>'hm','WindowsMediaPlayerSupport'=>'hm','PDFSupport'=>'hm');
%OSFamily=('win'=>'Windows','mac'=>'Macintosh','linux'=>'Linux','bsd'=>'BSD');
%BrowsersFamily=('msie'=>1,'firefox'=>2,'netscape'=>3,'svn'=>4);
@SessionsRange=('0s-30s','30s-2mn','2mn-5mn','5mn-15mn','15mn-30mn','30mn-1h','1h+');
%SessionsAverage=('0s-30s',15,'30s-2mn',75,'2mn-5mn',210,'5mn-15mn',600,'15mn-30mn',1350,'30mn-1h',2700,'1h+',3600);
# HTTP-Accept or Lang parameter => AWScloudMap code to use for lang
# ISO-639-1 or 2 or other       => AWScloudMap-xx.txt where xx is ISO-639-1
%LanguageCodeMap=(
'sq'=>'al','ar'=>'ar','ba'=>'ba','bg'=>'bg','zh-tw'=>'tw','zh'=>'cn','cs'=>'cz',
'de'=>'de','da'=>'dk',
'en'=>'en','et'=>'et','fi'=>'fi','fr'=>'fr','gl'=>'gl',
'es'=>'es','eu'=>'eu','ca'=>'ca',
'el'=>'gr','hu'=>'hu','is'=>'is','in'=>'id','it'=>'it',
'ja'=>'jp','kr'=>'ko','lv'=>'lv','nl'=>'nl','no'=>'nb','nb'=>'nb','nn'=>'nn',
'pl'=>'pl','pt'=>'pt','pt-br'=>'br','ro'=>'ro','ru'=>'ru','sr'=>'sr','sk'=>'sk',
'sv'=>'se','th'=>'th','tr'=>'tr','uk'=>'ua','cy'=>'cy','wlk'=>'cy'
);
@HostAliases=();
@DefaultFile = ();
@SkipHosts = @SkipUserAgents = @NotPageFiles = @SkipFiles = @SkipExactFiles = @SkipReferrers = ();
@OnlyHosts = @OnlyUserAgents = @OnlyFiles = ();
@URLWithQueryWithOnly = @URLWithQueryWithout = ();
@ExtraName = @ExtraCondition = @ExtraStatTypes = @MaxNbOfExtra = @MinHitExtra = ();
@ExtraFirstColumnTitle = @ExtraFirstColumnValues = @ExtraFirstColumnFunction = @ExtraFirstColumnFormat = ();
@ExtraCodeFilter = @ExtraConditionType = @ExtraConditionTypeVal = ();
@ExtraFirstColumnValuesType = @ExtraFirstColumnValuesTypeVal = ();
@ExtraAddAverageRow = @ExtraAddSumRow = ();

use vars qw/
%HTMLOutput %FilterIn %FilterEx
%BadFormatWarning
%MonthNumLib
%ValidHTTPCodes
%TrapInfosForHTTPErrorCodes %NotPageList %DayBytes %DayHits %DayPages %DayVisits
%MaxNbOf %MinHit
%ListOfYears %HistoryAlreadyFlushed %PosInFile %ValueInFile
%val %nextval %egal
%TmpOS %TmpRefererServer %TmpRobot %TmpBrowser
/;
%HTMLOutput = %FilterIn = %FilterEx = ();
%BadFormatWarning = ();
%MonthNumLib = ();
%ValidHTTPCodes = ();
%TrapInfosForHTTPErrorCodes=(); $TrapInfosForHTTPErrorCodes{404}=1;	# TODO Add this in config file
%NotPageList=();
%DayBytes = %DayHits = %DayPages = %DayVisits = ();
%MaxNbOf = %MinHit = ();
%ListOfYears = %HistoryAlreadyFlushed = %PosInFile = %ValueInFile = ();
%val = %nextval = %egal = ();
%TmpOS = %TmpRefererServer = %TmpRobot = %TmpBrowser = ();
use vars qw/
%FirstTime %LastTime
%MonthHostsKnown %MonthHostsUnknown
%MonthUnique %MonthVisits
%MonthPages %MonthHits %MonthBytes
%MonthNotViewedPages %MonthNotViewedHits %MonthNotViewedBytes
%_session %_browser_h
%_domener_p %_domener_h %_domener_k %_errors_k
%_filetypes_h %_filetypes_k %_filetypes_gz_in %_filetypes_gz_out
%_host_p %_host_h %_host_k %_host_l %_host_s %_host_u
%_waithost_e %_waithost_l %_waithost_s %_waithost_u
%_keyphrases %_keywords %_os_h %_pagesrefs_p %_pagesrefs_h %_robot_h %_robot_k %_robot_l %_robot_r
%_worm_h %_worm_k %_worm_l %_login_h %_login_p %_login_k %_login_l %_screensize_h
%_misc_p %_misc_h %_misc_k
%_cluster_p %_cluster_h %_cluster_k
%_url_p %_url_title %_url_x
/;
&Init_HashArray();
# ---------- Init Regex --------
use vars qw/ $regclean1 $regclean2 $regdate /;
$regclean1=qr/<(recnb|\/td)>/i;
$regclean2=qr/<\/?[^<>]+>/i;
$regdate=qr/(\d\d\d\d)(\d\d)(\d\d)(\d\d)(\d\d)(\d\d)/;

my $MetaRobot=0;	# meta robots
my $MetaKeywords=0;	# meta keywords
my $helpfound=0;
my $nohosts=0;
my $overwritedata=0;
my $hostname="";
my $useHTTPS=0;
my @archivedKeys=();
my @archivedHits=();
my $aliasCounter = 0;
my $pageTitle = "";
# Create a user agent (browser) object:
my $ua = new LWP::UserAgent;

my $FILE_START_MARKER = "BEGIN_DATA";
my $FILE_END_MARKER   = "END_DATA";

# String to appear if we can't retrieve it properly.
# NB: sub sanitisePageTitle will replace this with a shortened URL for display purposes
my $MISSING_TITLE_STRING = "missing";
my $EMPTY_TITLE_STRING   = "empty";


# Throttle HTTP requests - help avoid DoS-like results if on a quick network.
# Number is the number of seconds to pause between requests. Set to zero for
# no throttling.
my $throttleRequestsTime = 1;

# LWP settings
# UserAgent string passed in all server requests.
# You should add this to SkipUserAgents in your ?????.cloudMap.conf file if you want to ignore hits from this program.
my $retrievalUserAgent = "cloudMap_v$VERSION";
# Put a sensible e-mail address here:
my $spiderOwner = "AWSspider\@calcResult.com";

# Timeout (in seconds) for each HTTP request (increase on slow connections)
my $getTimeOut = 4;
# Proxy server to use when doing http/s - leave blank if you don't have one
#my $proxyServer = "http://my.proxy.server:port/";
#my $proxyServer = "http://proxy.example.com:8080";
my $proxyServer = "";
# Hosts not to use a proxy for
my @hostsNoProxy = ("localhost","10.0.0.6");
# Make sure we don't download multi-megabyte files! We only need the <title>
my $maxDocSizeBytes = 1024; # number is bytes


# PROTOCOL CODES
use vars qw/ %httpcodelib %ftpcodelib /;

# DEFAULT MESSAGES
use vars qw/ @Message /;
@Message=(
'First visit',
'Last visit',
'',
'Pages-URL',
'Hits',
'Jan',
'Feb',
'Mar',
'Apr',
'May',
'Jun',
'Jul',
'Aug',
'Sep',
'Oct',
'Nov',
'Dec',
'Bandwidth',
'Top',
'dd mmm yyyy - HH:MM',
'KB',
'MB',
'GB',
'Bytes',
'Mails',
'Size',
'First',
'Last',
'most-viewed pages listed, out of',
'logged'
);


# Set up a few items so that we can do HTML page retrieval:
# User agent name:
$ua->agent($retrievalUserAgent);
# set user-agent owners e-mail address:
$ua->from($spiderOwner);
# set timeout for requests:
$ua->timeout($getTimeOut);
if ($proxyServer)
{
	# set proxy for access to external sites
	$ua->proxy(["http","https"],$proxyServer);
	# avoid proxy for these hosts
	$ua->no_proxy(@hostsNoProxy);
}
# set maximum size of document to retrieve (in bytes)
$ua->max_size($maxDocSizeBytes);
if(!($ua->is_protocol_supported('https')) && $useHTTPS)
{
	print "\nSorry, SSL is not supported on this machine.\n";
	exit();
}



#------------------------------------------------------------------------------
# Function to solve problem with openvms
#------------------------------------------------------------------------------
sub file_filt (@)
{
    my @retval;
    foreach my $fl (@_)
    {
        $fl =~ tr/^//d;
        push @retval, $fl;
    }
    return sort @retval;
}


#------------------------------------------------------------------------------
# Function:		Write the output header of HTML page
# Parameters:	None
# Input:		%HTMLOutput $Lang $StyleSheet $PageCode $PageDir
# Output:		$HeaderHTMLSent=1
# Return:		None
#------------------------------------------------------------------------------
sub html_head
{
	my $dir=$PageDir?'right':'left';
	if (scalar keys %HTMLOutput)
	{
        my $periodtitle=" ($YearRequired";
        $periodtitle.=($MonthRequired ne 'all'?"-$MonthRequired":"");
        $periodtitle.=($DayRequired ne ''?"-$DayRequired":"");
        $periodtitle.=($HourRequired ne ''?"-$HourRequired":"");
        $periodtitle.=")";
		# Write head section
		if ($BuildReportFormat eq 'xhtml' || $BuildReportFormat eq 'xml')
		{
			if ($PageCode) { print "<?xml version=\"1.0\" encoding=\"$PageCode\"?>\n"; }
			else { print "<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>\n"; };
            print "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">";
			print "\n<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"$Lang\">\n";
		} else {
			# HTML 4:
			# print "<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\n";

			# HTML 5:
			print "<!DOCTYPE html>";
			print "\n<html lang='$Lang'".($PageDir?" dir='rtl'":"").">";
		}
		print "\n<head>";

        my $endtag='>';
        if ($BuildReportFormat eq 'xhtml' || $BuildReportFormat eq 'xml') { $endtag=' />'; }

        print "\n<meta name=\"generator\" content=\"AWScloudMap $VERSION from config file AWScloudMap.$SiteConfig.conf (http://cloudMap.calcResult.com)\"$endtag";

		if ($MetaRobot)
		{
		    print "\n<meta name=\"robots\" content=\"noindex,nofollow\"$endtag";
		}

		if ($BuildReportFormat eq 'xhtml' || $BuildReportFormat eq 'xml')
		{
		    print ($ENV{'HTTP_USER_AGENT'}=~/MSIE|Googlebot/i?"<meta http-equiv=\"content-type\" content=\"text/html; charset=".($PageCode?$PageCode:"iso-8859-1")."\" />\n":"<meta http-equiv=\"content-type\" content=\"text/xml; charset=".($PageCode?$PageCode:"iso-8859-1")."\"$endtag\n");
		} else {
		    print "\n<meta http-equiv=\"content-type\" content=\"text/html; charset=".($PageCode?$PageCode:"iso-8859-1")."\"$endtag";
		}

		print "\n<meta http-equiv=\"description\" content=\"".ucfirst($PROG)." - Advanced Web Statistics for $SiteDomain$periodtitle\"$endtag";
		if ($MetaKeywords) { print "\n<meta http-equiv=\"keywords\" content=\"$SiteDomain, free, advanced, realtime, web, server, logfile, log, analyzer, analysis, statistics, stats, perl, analyse, performance, hits, visits\"$endtag"; }
		print "\n<title>CloudMap $SiteDomain$periodtitle</title>";

		if ($StyleSheet)
		{
			print "\n<link rel=\"stylesheet\" href=\"$StyleSheet\" />";
		} else {
    		# A STYLE section must be in head section. Do not use " for number in a style section
    		print "\n<style type=\"text/css\">";
    		if ($BuildReportFormat eq 'xhtml' || $BuildReportFormat eq 'xml')
    		{
    		    print ($ENV{'HTTP_USER_AGENT'}=~/Firebird/i?"<!--\n":"<![CDATA[\n");
    		} else {
    		    # print "\n<!--";   This was never needed for STYLE sections - only for JavaScript
    		}

            my $thisPercentage = $BASE_PERCENTAGE;
            my $stopPoint = $MAX_RANK + 1;
            for (my $i=0; $i<$stopPoint; $i++)
            {
                print "\na.cloudLevel$i  { font-size:".$thisPercentage."%; line-height:1.25em; }";
                $thisPercentage += $FONT_INCREASE;
            }

    		if ($BuildReportFormat eq 'xhtml' || $BuildReportFormat eq 'xml')
    		{
    		    print ($ENV{'HTTP_USER_AGENT'}=~/Firebird/i?"//-->\n":"]]>\n");
    		} else {
    		    # print "\n//-->";
    		}
    		print "\n</style>";
        }
        &get_html_header();
		print "\n</head>\n";
	}
	$HeaderHTMLSent++;
}


#------------------------------------------------------------------------------
# Function:		Write end of HTML page to std output
# Parameters:	None
# Input:		None
# Output:		None
# Return:		None
#------------------------------------------------------------------------------
sub get_html_end
{
	if ($DirConfig)
	{
	    # Read HTML from external template:
		if ($DirConfig && $DirConfig !~ /[\\\/]$/) { $DirConfig .= "/"; }
		debug( "Template file: '$DirConfig$SiteConfig.end.template'", 6);
		if (open(EXT_TEMPLATE, "$DirConfig$SiteConfig.end.template"))
		{
        	&Print_File( *EXT_TEMPLATE );
        	close EXT_TEMPLATE;
        	return;
        }
	}
	print "\n<!-- No additional footer info added. Template file not found? -->";
	print "\n</body>";
	print "\n</html>\n";
}


#------------------------------------------------------------------------------
# Function:		Write start of HTML page to std output
# Parameters:	None
# Input:		None
# Output:		None
# Return:		None
#------------------------------------------------------------------------------
sub get_html_start
{
	if ($DirConfig)
	{
	    # Read HTML from external template:
		if ($DirConfig && $DirConfig !~ /[\\\/]$/) { $DirConfig .= "/"; }
		debug( "Template file: '$DirConfig$SiteConfig.start.template'", 6);
		if (open(EXT_TEMPLATE, "$DirConfig$SiteConfig.start.template"))
		{
        	&Print_File( *EXT_TEMPLATE );
        	close EXT_TEMPLATE;
        	return;
        }
	}
	print "\n<!-- No BODY info added. Template file not found? -->";
	print "\n<body>";
}


#------------------------------------------------------------------------------
# Function:		Write end of HTML page to std output
# Parameters:	None
# Input:		None
# Output:		None
# Return:		None
#------------------------------------------------------------------------------
sub get_html_header
{
	if ($DirConfig)
	{
	    # Read HTML from external template:
		if ($DirConfig && $DirConfig !~ /[\\\/]$/) { $DirConfig .= "/"; }
		debug( "Template file: '$DirConfig$SiteConfig.header.template'", 6);
		if (open(EXT_TEMPLATE, "$DirConfig$SiteConfig.header.template"))
		{
        	&Print_File( *EXT_TEMPLATE );
        	close EXT_TEMPLATE;
        	return;
        }
	}
	print "\n<!-- No additional header info added. Template file not found? -->";
}


#------------------------------------------------------------------------------
# Function:		Write HTML template to std-out
# Parameters:	None
# Input:		None
# Output:		None
# Return:		None
#------------------------------------------------------------------------------
sub Print_File
{
    my $fileHandle = shift;
	print "\n";
   	while (<$fileHandle>)
   	{
		print $_;
	}
}


#------------------------------------------------------------------------------
# Function:		Write error message and exit
# Parameters:	$message $secondmessage $thirdmessage $donotshowsetupinfo
# Input:		$HeaderHTMLSent %HTMLOutput $LogSeparator $LogFormat
# Output:		None
# Return:		None
#------------------------------------------------------------------------------
sub error
{
	my $message=shift||''; if (scalar keys %HTMLOutput) { $message =~ s/\</&lt;/g; $message =~ s/\>/&gt;/g; }
	my $secondmessage=shift||'';
	my $thirdmessage=shift||'';
	my $donotshowsetupinfo=shift||0;

	if (! $HeaderHTMLSent && scalar keys %HTMLOutput) 	{ print "<html><body>\n"; $HeaderHTMLSent=1; }
	debug("$message $secondmessage $thirdmessage",1);
	my $tagbold=''; my $tagunbold=''; my $tagbr=''; my $tagfontred=''; my $tagfontgrey=''; my $tagunfont='';
	if (scalar keys %HTMLOutput)
	{
		$tagbold='<b>'; $tagunbold='</b>'; $tagbr='<br>';
		$tagfontred='<span style="color: #880000">';
		$tagfontgrey='<span style="color: #888888">';
		$tagunfont='</span>';
	}
	if (! $ErrorMessages && $message =~ /^Format error$/i)
	{
		# Files seems to have bad format
		if (scalar keys %HTMLOutput) { print "\n<br><br>"; }
		if ($message !~ $LogSeparator)
		{
			# Bad LogSeparator parameter
			print "\n${tagfontred}AWScloudMap did not find the ${tagbold}LogSeparator${tagunbold} in these log records.${tagbr}${tagunfont}";
		} else {
			# Bad LogFormat parameter
			print "\nAWScloudMap did not find any valid log lines that matched the ${tagbold}LogFormat${tagunbold} parameter, in the first ${NbOfLinesForCorruptedLog} non-commented lines read from this log.${tagbr}";
			print "\n${tagfontred}The log file ${tagbold}$thirdmessage${tagunbold} must have a bad format or ${tagbold}LogFormat${tagunbold} parameter setup does not match this format.${tagbr}${tagbr}${tagunfont}";
			print "\nThe AWScloudMap ${tagbold}LogFormat${tagunbold} parameter is:";
			print "\n${tagbold}$LogFormat${tagunbold}${tagbr}";
			print "\nThis means each line in the web server log file needs to have:";
			if ($LogFormat == 1)
			{
				print "\n${tagbold}\"combined log format\"${tagunbold} like this:${tagbr}";
				print (scalar keys %HTMLOutput?"$tagfontgrey<i>":"");
				print "111.22.33.44 - - [10/Jan/2001:02:14:14 +0200] \"GET / HTTP/1.1\" 200 1234 \"http://www.fromserver.com/from.htm\" \"Mozilla/4.0 (compatible; MSIE 5.01; Windows NT 5.0)\"\n";
				print (scalar keys %HTMLOutput?"</i>$tagunfont${tagbr}${tagbr}\n":"");
			}
			if ($LogFormat == 2)
			{
				print "\n${tagbold}\"MSIE Extended W3C log format\"${tagunbold} like this:${tagbr}\n";
				print (scalar keys %HTMLOutput?"$tagfontgrey<i>":"");
				print "date time c-ip c-username cs-method cs-uri-sterm sc-status sc-bytes cs-version cs(User-Agent) cs(Referer)\n";
				print (scalar keys %HTMLOutput?"</i>$tagunfont${tagbr}${tagbr}\n":"");
			}
			if ($LogFormat == 3)
			{
				print "\n${tagbold}\"WebStar native log format\"${tagunbold}${tagbr}\n";
			}
			if ($LogFormat == 4)
			{
				print "\n${tagbold}\"common log format\"${tagunbold} like this:${tagbr}\n";
				print (scalar keys %HTMLOutput?"$tagfontgrey<i>":"");
				print "111.22.33.44 - - [10/Jan/2001:02:14:14 +0200] \"GET / HTTP/1.1\" 200 1234\n";
				print (scalar keys %HTMLOutput?"</i>$tagunfont${tagbr}${tagbr}\n":"");
			}
			if ($LogFormat == 6)
			{
				print "\n${tagbold}\"Lotus Notes/Lotus Domino\"${tagunbold}${tagbr}\n";
				print (scalar keys %HTMLOutput?"$tagfontgrey<i>":"");
				print "111.22.33.44 - Firstname Middlename Lastname [10/Jan/2001:02:14:14 +0200] \"GET / HTTP/1.1\" 200 1234 \"http://www.fromserver.com/from.htm\" \"Mozilla/4.0 (compatible; MSIE 5.01; Windows NT 5.0)\"\n";
				print (scalar keys %HTMLOutput?"</i></span>${tagbr}${tagbr}\n":"");
			}
			if ($LogFormat !~ /^[1-6]$/)
			{
				print "\nthe following personalized log format:${tagbr}\n";
				print (scalar keys %HTMLOutput?"$tagfontgrey<i>":"");
				print "$LogFormat\n";
				print (scalar keys %HTMLOutput?"</i>$tagunfont${tagbr}${tagbr}\n":"");
			}
			print "\nAnd this is an example of records AWScloudMap found in your log file (the record number $NbOfLinesForCorruptedLog in your log):\n";
			print (scalar keys %HTMLOutput?"<br>$tagfontgrey<i>":"");
			print "$secondmessage";
			print (scalar keys %HTMLOutput?"</i>$tagunfont${tagbr}${tagbr}":"");
			print "\n";
		}
		#print "Note: If the first $NbOfLinesForCorruptedLog lines in your log files are wrong because of a worm or virus attack, you can increase the NbOfLinesForCorruptedLog parameter in config file.\n";
	} else {
		print (scalar keys %HTMLOutput?"<br>$tagfontred\n":"");
		print ($ErrorMessages?"$ErrorMessages":"Error: $message");
		print (scalar keys %HTMLOutput?"\n</span><br>":"");
		print "\n";
	}
	if (! $ErrorMessages && ! $donotshowsetupinfo)
	{
		if ($message =~ /Couldn.t open config file/i)
		{
			my $dir=$DIR;
			if ($dir =~ /^\./)
			{
			    $dir.='/../..';
		    } else {
		        $dir =~ s/[\\\/]?wwwroot[\/\\]cgi-bin[\\\/]?//;  ## This is out-of-date MJB
	        }
			print "\n${tagbr}";
			print "\n- ${tagbold}Did you use correct config parameter ?${tagunbold}${tagbr}";
			print "\nExample: If your config file is mysite.cloudMap.conf, use -config=mysite";
			print "\n- ${tagbold}Did you create your config file 'AWScloudMap.$SiteConfig.conf' ?${tagunbold}${tagbr}";
			print "\n${tagbr}";
		} else {
		    print "\n${tagbr}${tagbold}Setup (".($FileConfig?"'".$FileConfig."'":"Config")." file, web server or permissions) may be wrong.${tagunbold}${tagbr}";
	    }
		print "\nCheck config file, permissions and AWScloudMap documentation (in 'docs' directory).";
	}
	# Remove lock if not a lock message
	if ($EnableLockForUpdate && $message !~ /lock file/) { &Lock_Update(0); }
	if (scalar keys %HTMLOutput) { print "\n</body></html>\n"; }
	exit 1;
}

#------------------------------------------------------------------------------
# Function:		Write a warning message
# Parameters:	$message
# Input:		$HeaderHTMLSent $WarningMessage %HTMLOutput $PluginMode
# Output:		None
# Return:		None
#------------------------------------------------------------------------------
sub warning
{
	my $messagestring=shift;

	debug("$messagestring",1);
	if ($WarningMessages)
	{
        if (! $HeaderHTMLSent && !$PluginMode) { &html_head; }
		if (scalar keys %HTMLOutput)
		{
			$messagestring =~ s/\n/\<br\>/g;
			print "\n$messagestring<br>";
		} else {
			print "\n$messagestring";
		}
	}
}

#------------------------------------------------------------------------------
# Function:     Write debug message and exit
# Parameters:   $string $level
# Input:        %HTMLOutput  $Debug=required level  $DEBUG_FORCED=required level forced
# Output:		None
# Return:		None
#------------------------------------------------------------------------------
sub debug
{
	my $level = $_[1] || 1;

	if ($level <= $DEBUG_FORCED)
	{
		my $debugstring = $_[0];
		if (! $DebugResetDone) { open(DEBUG_FORCEDFILE,"debug.log"); close DEBUG_FORCEDFILE; chmod 0666,"debug.log"; $DebugResetDone=1; }
		open(DEBUG_FORCEDFILE,">>debug.log");
		print DEBUG_FORCEDFILE localtime(time)." - $$ - DEBUG $level - $debugstring\n";
		close DEBUG_FORCEDFILE;
	}
	if ($level <= $Debug)
	{
		my $debugstring = $_[0];
		if (scalar keys %HTMLOutput) { $debugstring =~ s/^ /&nbsp;&nbsp; /; $debugstring .= "<br>"; }
		print localtime(time)." - DEBUG $level - $debugstring\n";
	}
}

#------------------------------------------------------------------------------
# Function:     Optimize an array removing duplicate entries
# Parameters:	@Array notcasesensitive=0|1
# Input:        None
# Output:		None
# Return:		None
#------------------------------------------------------------------------------
sub OptimizeArray
{
	my $array=shift;
	my @arrayunreg=map{if (/\(\?[-\w]*:(.*)\)/) { $1 } } @$array;
	my $notcasesensitive=shift;
	my $searchlist=0;
	debug("OptimizeArray (notcasesensitive=$notcasesensitive)",4);
	while ($searchlist>-1 && @arrayunreg)
	{
		my $elemtoremove=-1;
		OPTIMIZELOOP:
		foreach my $i ($searchlist..(scalar @arrayunreg)-1)
		{
			# Search if $i elem is already treated by another elem
			foreach my $j (0..(scalar @arrayunreg)-1)
			{
				if ($i == $j) { next; }
				my $parami=$notcasesensitive?lc($arrayunreg[$i]):$arrayunreg[$i];
				my $paramj=$notcasesensitive?lc($arrayunreg[$j]):$arrayunreg[$j];
				debug(" Compare $i ($parami) to $j ($paramj)",4);
				if (index($parami,$paramj)>-1) {
					debug(" Elem $i ($arrayunreg[$i]) already treated with elem $j ($arrayunreg[$j])",4);
					$elemtoremove=$i;
					last OPTIMIZELOOP;
				}
			}
		}
		if ($elemtoremove > -1)
		{
			debug(" Remove elem $elemtoremove - $arrayunreg[$elemtoremove]",4);
			splice @arrayunreg, $elemtoremove, 1;
			$searchlist=$elemtoremove;
		} else {
			$searchlist=-1;
		}
	}
	if ($notcasesensitive) { return map{qr/$_/i} @arrayunreg; }
	return map{qr/$_/} @arrayunreg;
}


#------------------------------------------------------------------------------
# Function:     Check if parameter is in SkipHosts array
# Parameters:	host @SkipHosts (a NOT case sensitive precompiled regex array)
# Return:		0 Not found, 1 Found
#------------------------------------------------------------------------------
sub SkipHost
{
	foreach (@SkipHosts) { if ($_[0] =~ /$_/) { return 1; } }
	0; # Not in @SkipHosts
}

#------------------------------------------------------------------------------
# Function:     Check if parameter is in SkipReferrers array
# Parameters:	host @SkipReferrers (a NOT case sensitive precompiled regex array)
# Return:		0 Not found, 1 Found
#------------------------------------------------------------------------------
sub SkipReferrer
{
	foreach (@SkipReferrers) { if ($_[0] =~ /$_/) { return 1; } }
	0; # Not in @SkipReferrers
}

#------------------------------------------------------------------------------
# Function:     Check if parameter is in SkipUserAgents array
# Parameters:	useragent @SkipUserAgents (a NOT case sensitive precompiled regex array)
# Return:		0 Not found, 1 Found
#------------------------------------------------------------------------------
sub SkipUserAgent
{
	my $thisAgent = shift;
	foreach (@SkipUserAgents) { if ($thisAgent =~ /$_/) { return 1; } }
	if ($thisAgent eq $retrievalUserAgent)  { return 1; }  # Check against the user-agent set for the Page-Title retrieval
	0; # Not in @SkipUserAgent
}

#------------------------------------------------------------------------------
# Function:     Check if parameter is in SkipFiles array
# Parameters:	url @SkipFiles (a NOT case sensitive precompiled regex array)
# Return:		0 Not found, 1 Found
#------------------------------------------------------------------------------
sub SkipFile
{
	my $thisFile = shift;
	my $arrayLength = scalar(@SkipFiles);
    if($SKIP_FOLDERS)
    {
	    if ($thisFile =~ /\/$/) # Trailing Slash
	    {
    	    debug("URL: '$thisFile' Rejected by 'SKIP_FOLDERS = $SKIP_FOLDERS'", 1);
	        return 1;  # Ignore this URL
        }
    }
    for (my $i=0; $i<$arrayLength; $i++)
    {
	    if ($thisFile =~ $SkipFiles[$i])
	    {
    	    debug("URL: '$thisFile' Rejected by partial match with:$SkipFiles[$i]", 1);
	        return 1;  # Ignore this URL
        }
	}
	$arrayLength = scalar(@SkipExactFiles);
    for (my $i=0; $i<$arrayLength; $i++)
    {
	    if ($thisFile eq $SkipExactFiles[$i])
	    {
    	    debug("URL: '$thisFile' Rejected by exact match with: $SkipExactFiles[$i]", 1);
	        return 1;  # Ignore this URL
        }
	}
	0; # Okay to use.
}


#------------------------------------------------------------------------------
# Function:     Check if parameter is in OnlyHosts array
# Parameters:	host @OnlyHosts (a NOT case sensitive precompiled regex array)
# Return:		0 Not found, 1 Found
#------------------------------------------------------------------------------
sub OnlyHost
{
	foreach (@OnlyHosts) { if ($_[0] =~ /$_/) { return 1; } }
	0; # Not in @OnlyHosts
}

#------------------------------------------------------------------------------
# Function:     Check if parameter is in OnlyUserAgents array
# Parameters:	useragent @OnlyUserAgents (a NOT case sensitive precompiled regex array)
# Return:		0 Not found, 1 Found
#------------------------------------------------------------------------------
sub OnlyUserAgent
{
	foreach (@OnlyUserAgents) { if ($_[0] =~ /$_/) { return 1; } }
	0; # Not in @OnlyHosts
}

#------------------------------------------------------------------------------
# Function:     Check if parameter is in NotPageFiles array
# Parameters:	url @NotPageFiles (a NOT case sensitive precompiled regex array)
# Return:		0 Not found, 1 Found
#------------------------------------------------------------------------------
sub NotPageFile
{
	foreach (@NotPageFiles) { if ($_[0] =~ /$_/) { return 1; } }
	0; # Not in @NotPageFiles
}

#------------------------------------------------------------------------------
# Function:     Check if parameter is in OnlyFiles array
# Parameters:	url @OnlyFiles (a NOT case sensitive precompiled regex array)
# Return:		0 Not found, 1 Found
#------------------------------------------------------------------------------
sub OnlyFile
{
	foreach (@OnlyFiles) { if ($_[0] =~ /$_/) { return 1; } }
	0; # Not in @OnlyFiles
}

#------------------------------------------------------------------------------
# Function:     Return 1 if a date exists
# Parameters:	$day $month $year
# Return:		1 if date exists else 0
#------------------------------------------------------------------------------
sub DateIsValid
{
	my ($day, $month, $year) = @_;
	debug("DateIsValid for $day $month $year",4);
	if ($day < 1)  { return 0; }
	if ($day > 31) { return 0; }
	if ($month==4 || $month==6 || $month==9 || $month==11)
	{
		if ($day > 30) { return 0; }
	}
	elsif ($month==2) {
		my $leapyear=($year%4==0?1:0);						# A leap year every 4 years
		if ($year%100==0 && $year%400!=0) { $leapyear=0; }	# Except if year is 100x and not 400x
		if ($day > (28+$leapyear)) { return 0; }
	}
	return 1;
}

#------------------------------------------------------------------------------
# Function:     Return string of visit duration
# Parameters:	$starttime $endtime
# Input:        None
# Output:		None
# Return:		A string that identify the visit duration range
#------------------------------------------------------------------------------
sub GetSessionRange
{
	my $starttime = my $endtime;
	if (shift =~ /$regdate/o) { $starttime = Time::Local::timelocal($6,$5,$4,$3,$2-1,$1); }
	if (shift =~ /$regdate/o) { $endtime = Time::Local::timelocal($6,$5,$4,$3,$2-1,$1); }
	my $delay=$endtime-$starttime;
	debug("GetSessionRange $endtime - $starttime = $delay",4);
	if ($delay <= 30)   { return $SessionsRange[0]; }
	if ($delay <= 120)  { return $SessionsRange[1]; }
	if ($delay <= 300)  { return $SessionsRange[2]; }
	if ($delay <= 900)  { return $SessionsRange[3]; }
	if ($delay <= 1800) { return $SessionsRange[4]; }
	if ($delay <= 3600) { return $SessionsRange[5]; }
	return $SessionsRange[6];
}

#------------------------------------------------------------------------------
# Function:     Read config file
# Parameters:	None or configdir to scan
# Input:        $DIR $PROG $SiteConfig
# Output:		Global variables
# Return:		-
#------------------------------------------------------------------------------
sub Read_Config
{
	my $configdir=shift;
	my @PossibleConfigDir=();
	# Check if the config file exists in common possible directories :
	# Windows :                   				"$DIR" (same dir as AWScloudMap.pl)
	# Linux :	                                "/etc/cloudMap", "/usr/local/etc/cloudMap", "/etc", "/etc/opt/cloudMap"

	if ($configdir)
	{
		@PossibleConfigDir=("$configdir");
	}else {
	    @PossibleConfigDir=("$DIR","./","/etc/cloudMap","/usr/local/etc/cloudMap","/etc","/etc/opt/cloudMap");
    }

	# Open the config file:
	$FileConfig=$FileSuffix='';
	my $searchdir="";
	foreach (@PossibleConfigDir)
	{
		$searchdir=$_;
		if ($searchdir && $searchdir !~ /[\\\/]$/) { $searchdir .= "/"; }
		if (open(CONFIG,"$searchdir$SiteConfig.$configFilePrefix.conf"))
        {
            $FileConfig="$searchdir$SiteConfig.$configFilePrefix.conf";
            $FileSuffix=".$SiteConfig";
            $DirConfig = $searchdir;
            last;
        }
		if (open(CONFIG,"$searchdir.$configFilePrefix.conf"))
		        { $FileConfig="$searchdir.$configFilePrefix.conf"; $FileSuffix=''; last; }
	}
	if (! $FileConfig) { error("Couldn't open config file \"$searchdir$SiteConfig.$configFilePrefix.conf\" nor \"$searchdir.$configFilePrefix.conf\" after searching in path \"".join(',',@PossibleConfigDir)."\": $!"); }

	# Analyze config file content and close it
	&Parse_Config( *CONFIG , 1 , $FileConfig);
	close CONFIG;

	# If parameter NotPageList not found, init for backward compatibility
	if (! $FoundNotPageList)
	{
		%NotPageList=('css'=>1,'js'=>1,'class'=>1,'gif'=>1,'jpg'=>1,'jpeg'=>1,'png'=>1,'bmp'=>1,'ico'=>1,'swf'=>1);
	}
	# If parameter ValidHTTPCodes empty, init for backward compatibility
	if (! scalar keys %ValidHTTPCodes) { $ValidHTTPCodes{"200"}=$ValidHTTPCodes{"304"}=1; }
}

#------------------------------------------------------------------------------
# Function:     Parse content of a config file
# Parameters:	opened file handle, depth level, file name
# Input:        -
# Output:		Global variables
# Return:		-
#------------------------------------------------------------------------------
sub Parse_Config
{
    my ( $confighandle ) = $_[0];
	my $level = $_[1];
	my $configFile = $_[2];
	my $versionnum=0;
	my $confLineNum=0;

	if ($level > 10) { error("$PROG can't read down more than 10 level of includes. Check that no 'included' config files include their parent config file (this cause infinite loop)."); }

   	while (<$confighandle>)
   	{
		chomp $_; s/\r//;
		$confLineNum++;

		if ($_ =~ /^\s*$/) { next; }

		# Check includes
		if ($_ =~ /^Include "([^\"]+)"/ || $_ =~ /^#include "([^\"]+)"/)
		{
		    # #include kept for backward compatibility
		    my $includeFile = $1;
		    # Expand __var__ by values
		    while ($includeFile =~ /__([^\s_]+(?:_[^\s_]+)*)__/) {
		        my $var=$1; $includeFile =~ s/__${var}__/$ENV{$var}/g;
		    }
			debug("Found an include : $includeFile",2);
		    if ( $includeFile !~ /^[\\\/]/ ) {
			    # Correct relative include files
				if ($FileConfig =~ /^(.*[\\\/])[^\\\/]*$/) { $includeFile = "$1$includeFile"; }
			}
			if ($level > 1) {
				warning("Warning: Perl versions before 5.6 cannot handle nested includes");
				next;
			}
		    if ( open( CONFIG_INCLUDE, $includeFile ) )
		    {
				&Parse_Config( *CONFIG_INCLUDE , $level+1, $includeFile);
				close( CONFIG_INCLUDE );
		    } else {
				error("Could not open include file: $includeFile" );
		    }
			next;
		}

		# Remove comments
		if ($_ =~ /^\s*#/) { next; }
		$_ =~ s/\s#.*$//;

		# Extract param and value
		my ($param,$value)=split(/=/,$_,2);
		$param =~ s/^\s+//; $param =~ s/\s+$//;

		# If not a param=value, try with next line
		if (! $param)         { warning("Warning: Syntax error line $confLineNum in file '$configFile'. Config line is ignored."); next; }
		if (! defined $value) { warning("Warning: Syntax error line $confLineNum in file '$configFile'. Config line is ignored."); next; }

		if ($value)
		{
			$value =~ s/^\s+//;
			$value =~ s/\s+$//;
			$value =~ s/^\"//;
			$value =~ s/\";?$//;
			# Replace __MONENV__ with value of environnement variable MONENV
			# Must be able to replace __VAR_1____VAR_2__
			while ($value =~ /__([^\s_]+(?:_[^\s_]+)*)__/) { my $var=$1; $value =~ s/__${var}__/$ENV{$var}/g; }
		}

		# Initialize parameter for (param,value)
		if ($param =~ /^LogFile/)
		{
			if ($QueryString !~ /logfile=([^\s&]+)/i) { $LogFile=$value; }
			next;
		}
		if ($param =~ /^LogType/)
		{
			if ($value =~ /([WSMFC])/i)
			{
			    $LogType=$1;
            	debug("\$logtype = $1 matched", 1);
		    }else{
            	debug("\$logtype NOT matched in '$value'", 1);
		    }
			next;
		}
		if ($param =~ /^SiteDomain/)
		{
			# No regex test as SiteDomain is always exact value
			$SiteDomain=$value;
			next;
		}
		if ($param =~ /^TestDomain/)
		{
			# No regex test as TestDomain is always exact value
			$TestDomain=$value;
			next;
		}
		if ($param =~ /^MinRank/)
		{
			if($value > -1)
			{
    			$MIN_RANK =$value;
    		}
			next;
		}
		if ($param =~ /^MaxRank/)
		{
			if($value > -1)
			{
    			$MAX_RANK =$value;
    		}
			next;
		}
		if ($param =~ /^BasePercentage/)
		{
			if($value > 10)
			{
    			$BASE_PERCENTAGE =$value;
    		}
			next;
		}
		if ($param =~ /^FontIncrease/)
		{
			if($value > 0)
			{
    			$FONT_INCREASE =$value;
    		}
			next;
		}
		if ($param =~ /^MissingTitleString/)
		{
			if($value)
			{
    			$MISSING_TITLE_STRING =$value;
    		}
			next;
		}
		if ($param =~ /^EmptyTitleString/)
		{
			if($value)
			{
    			$EMPTY_TITLE_STRING =$value;
    		}
			next;
		}
		if ($param =~ /^SkipFolders/)
		{
			if($value)
			{
    			$SKIP_FOLDERS = $value;
    			if($value =~ /false/i)
    			{
        			$SKIP_FOLDERS = "";  # Set to false
        		}
    		}
			next;
		}
		if ($param =~ /^RemoveFromTitle/)
		{
			# No regex test as RemoveFromTitle is always (a set of) exact values
			# This is a list of items in the config, so:
			push @RemoveFromTitle, $value;
			next;
		}
		if ($param =~ /^HostAliases/)
		{
            @HostAliases=();
			foreach my $elem (split(/\s+/,$value))
			{
			    if ($elem =~ s/^\@//)
			    {
			        # If list of hostaliases in a file
			        open(DATAFILE,"<$elem") || error("Failed to open file '$elem' declared in HostAliases parameter");
			        my @val=map(/^(.*)$/i,<DATAFILE>);
			        push @HostAliases, map{qr/^$_$/i} @val;
			        close(DATAFILE);
			    } else {
				    if ($elem =~ /^REGEX\[(.*)\]$/i) { $elem=$1; }
				    else { $elem='^'.quotemeta($elem).'$'; }
				    if ($elem) { push @HostAliases, qr/$elem/i; }
				}
			}
			next;
			}
		# Special optional setup params
		if ($param =~ /^DefaultFile/)
		{
		    @DefaultFile=();
			foreach my $elem (split(/\s+/,$value))
			{
				# No REGEX for this option
				#if ($elem =~ /^REGEX\[(.*)\]$/i) { $elem=$1; }
				#else { $elem='^'.quotemeta($elem).'$'; }
				if ($elem) { push @DefaultFile,$elem; }
			}
			next;
			}
		if ($param =~ /^SkipHosts/)
		{
		    @SkipHosts=();
			foreach my $elem (split(/\s+/,$value))
			{
				if ($elem =~ /^REGEX\[(.*)\]$/i) { $elem=$1; }
				else { $elem='^'.quotemeta($elem).'$'; }
				if ($elem) { push @SkipHosts, qr/$elem/i; }
			}
			next;
			}
		if ($param =~ /^SkipReferrersBlackList/ && $value)
		{
			open (BLACKLIST, "<$value") || die "Failed to open blacklist: $!\n";
			while (<BLACKLIST>) {
				chomp;
				my $elem = $_;
				$elem =~ s/ //;
				$elem =~ s/\#.*//;
				if ($elem) { push @SkipReferrers, qr/$elem/i; }
			}
			next;
			close (BLACKLIST);
			}
		if ($param =~ /^SkipUserAgents/)
		{
		    @SkipUserAgents=();
			foreach my $elem (split(/\s+/,$value))
			{
				if ($elem =~ /^REGEX\[(.*)\]$/i)
				{
				    $elem=$1;
			    }else{
			        $elem='^'.quotemeta($elem).'$';
		        }
				if ($elem) { push @SkipUserAgents, qr/$elem/i; }
			}
			next;
			}
		if ($param =~ /^SkipFiles/)
		{
		    @SkipFiles=();
			foreach my $elem (split(/\s+/,$value))
			{
				if ($elem =~ /^REGEX\[(.*)\]$/i)
				{
				    $elem=$1;
		        }
				if ($elem) { push @SkipFiles, qr/$elem/i; }
				debug("SkipFiles: \$elem = $elem", 1);
			}
			next;
		}
		if ($param =~ /^SkipExactFiles/)
		{
		    @SkipExactFiles=();
			foreach my $elem (split(/\s+/,$value))
			{
				if ($elem) { push @SkipExactFiles, $elem; }
			}
			next;
		}
		if ($param =~ /^OnlyHosts/) {
		    @OnlyHosts=();
			foreach my $elem (split(/\s+/,$value))	{
				if ($elem =~ /^REGEX\[(.*)\]$/i) { $elem=$1; }
				else { $elem='^'.quotemeta($elem).'$'; }
				if ($elem) { push @OnlyHosts, qr/$elem/i; }
			}
			next;
			}
		if ($param =~ /^OnlyUserAgents/) {
		    @OnlyUserAgents=();
			foreach my $elem (split(/\s+/,$value))	{
				if ($elem =~ /^REGEX\[(.*)\]$/i) { $elem=$1; }
				else { $elem='^'.quotemeta($elem).'$'; }
				if ($elem) { push @OnlyUserAgents, qr/$elem/i; }
			}
			next;
			}
		if ($param =~ /^OnlyFiles/) {
		    @OnlyFiles=();
			foreach my $elem (split(/\s+/,$value))	{
				if ($elem =~ /^REGEX\[(.*)\]$/i) { $elem=$1; }
				else { $elem='^'.quotemeta($elem).'$'; }
				if ($elem) { push @OnlyFiles, qr/$elem/i; }
			}
			next;
			}
		if ($param =~ /^NotPageFiles/) {
		    @NotPageFiles=();
			foreach my $elem (split(/\s+/,$value))	{
				if ($elem =~ /^REGEX\[(.*)\]$/i) { $elem=$1; }
				else { $elem='^'.quotemeta($elem).'$'; }
				if ($elem) { push @NotPageFiles, qr/$elem/i; }
			}
			next;
			}
		if ($param =~ /^NotPageList/) {
            %NotPageList=();
			foreach (split(/\s+/,$value))	{ $NotPageList{$_}=1; }
			$FoundNotPageList=1;
			next;
			}
		if ($param =~ /^ValidHTTPCodes/) {
            %ValidHTTPCodes=();
			foreach (split(/\s+/,$value))	{ $ValidHTTPCodes{$_}=1; }
			next;
			}
		if ($param =~ /^URLWithQueryWithOnlyFollowingParameters$/)	{
			@URLWithQueryWithOnly=split(/\s+/,$value);
			next;
			}
		if ($param =~ /^URLWithQueryWithoutFollowingParameters$/)	{
			@URLWithQueryWithout=split(/\s+/,$value);
			next;
			}

 		# Extra parameters
 		if ($param =~ /^ExtraSectionName(\d+)/)			        { $ExtraName[$1]=$value; next; }
 		if ($param =~ /^ExtraSectionCodeFilter(\d+)/)  	        { @{$ExtraCodeFilter[$1]}=split(/\s+/,$value); next; }
 		if ($param =~ /^ExtraSectionCondition(\d+)/)  	        { $ExtraCondition[$1]=$value; next; }
 		if ($param =~ /^ExtraSectionStatTypes(\d+)/)            { $ExtraStatTypes[$1]=$value; next; }
 		if ($param =~ /^ExtraSectionFirstColumnTitle(\d+)/) 	{ $ExtraFirstColumnTitle[$1]=$value; next; }
 		if ($param =~ /^ExtraSectionFirstColumnValues(\d+)/) 	{ $ExtraFirstColumnValues[$1]=$value; next; }
 		if ($param =~ /^ExtraSectionFirstColumnFunction(\d+)/) 	{ $ExtraFirstColumnFunction[$1]=$value; next; }
 		if ($param =~ /^ExtraSectionFirstColumnFormat(\d+)/) 	{ $ExtraFirstColumnFormat[$1]=$value; next; }
 		if ($param =~ /^ExtraSectionAddAverageRow(\d+)/) 	    { $ExtraAddAverageRow[$1]=$value; next; }
 		if ($param =~ /^ExtraSectionAddSumRow(\d+)/) 	        { $ExtraAddSumRow[$1]=$value; next; }
 		if ($param =~ /^MaxNbOfExtra(\d+)/) 			        { $MaxNbOfExtra[$1]=$value; next; }
 		if ($param =~ /^MinHitExtra(\d+)/) 				        { $MinHitExtra[$1]=$value; next; }
		# Other parameter checks we need to put after MaxNbOfExtra and MinHitExtra
 		if ($param =~ /^MaxNbOf(\w+)/) 	                        { $MaxNbOf{$1}=$value; next; }
 		if ($param =~ /^MinHit(\w+)/) 	                        { $MinHit{$1}=$value; next; }
		# Check if this is a known parameter
#		if (! $ConfOk{$param}) { error("Unknown config parameter '$param' found line $confLineNum in file \"configFile\""); }
		# If parameters was not found previously, defined variable with name of param to value
		$$param=$value;
	}

    if( ! $TestDomain)
    {
    	# If nothing is defined, we fall back to the live site's URL:
    	$TestDomain=$SiteDomain;
    }

	debug("Config file read was \"$configFile\" (level $level)");
}



#------------------------------------------------------------------------------
# Function:     Get the messages for a specified language
# Parameters:	LanguageId
# Input:		$DirLang $DIR
# Output:		$Message table is defined in memory
# Return:		None
#------------------------------------------------------------------------------
sub Read_Language_Data
{
	# Look for lang files in common possible directories :
	my @PossibleLangDir=("$DirLang","$DIR/lang","/usr/share/AWScloudMap/lang");

	my $FileLang='';
	foreach (@PossibleLangDir)
	{
		my $searchdir=$_;
		if ($searchdir && (!($searchdir =~ /\/$/)) && (!($searchdir =~ /\\$/)) ) { $searchdir .= "/"; }
		if (open(LANG,"${searchdir}cloudMap-$_[0].txt")) { $FileLang="${searchdir}cloudMap-$_[0].txt"; last; }
	}
	# If file not found, default to english:
	if (! $FileLang)
	{
		foreach (@PossibleLangDir)
		{
			my $searchdir=$_;
			if ($searchdir && (!($searchdir =~ /\/$/)) && (!($searchdir =~ /\\$/)) ) { $searchdir .= "/"; }
			if (open(LANG,"${searchdir}awstats-en.txt")) { $FileLang="${searchdir}awstats-en.txt"; last; }
		}
	}
	debug("Call to Read_Language_Data [FileLang=\"$FileLang\"]");
	if ($FileLang)
	{
		my $i = 0;
		binmode LANG;	# Might avoid 'Malformed UTF-8 errors'
		my $cregcode=qr/^PageCode=[\t\s\"\']*([\w-]+)/i;
		my $cregdir=qr/^PageDir=[\t\s\"\']*([\w-]+)/i;
		my $cregmessage=qr/^Message\d+=/i;
		while (<LANG>)
		{
			chomp $_; s/\r//;
			if ($_ =~ /$cregcode/o) { $PageCode = $1; }
			if ($_ =~ /$cregdir/o)  { $PageDir = $1; }
			if ($_ =~ s/$cregmessage//o)
			{
				$_ =~ s/^#.*//;								# Remove comments
				$_ =~ s/\s+#.*//;							# Remove comments
				$_ =~ tr/\t /  /s;							# Change all blanks into " " (space)
				$_ =~ s/^\s+//; $_ =~ s/\s+$//;
				$_ =~ s/^\"//; $_ =~ s/\"$//;
				$Message[$i] = $_;                  # This means that the language files can't have items removed arbitrarily. MJB
				$i++;
			}
		}
		close(LANG);
	} else {
		warning("Warning: Couldn't find language files for \"$_[0]\", defaulting to English.");
	}
	# Some language string changes
	if ($LogType eq 'M')
	{
	    # For mail
		$Message[0]=$Message[26];
		$Message[1]=$Message[27];
		$Message[4]=$Message[24];
		$Message[17]=$Message[25];
	}
}


#------------------------------------------------------------------------------
# Function:     Substitute date tags in a string by value
# Parameters:	String
# Input:		All global variables
# Output:		Change on some global variables
# Return:		String
#------------------------------------------------------------------------------
sub Substitute_Tags
{
    my $SourceString=shift;
	debug("Call to Substitute_Tags on '$SourceString'");

	my %MonthNumLibEn = ("01","Jan","02","Feb","03","Mar","04","Apr","05","May","06","Jun","07","Jul","08","Aug","09","Sep","10","Oct","11","Nov","12","Dec");

	while ($SourceString =~ /%([ymdhwYMDHWNSO]+)-(\(\d+\)|\d+)/)
	{
		# Accept tag %xx-dd and %xx-(dd)
		my $timetag="$1";
		my $timephase=quotemeta("$2");
		my $timephasenb="$2"; $timephasenb=~s/[^\d]//g;
		debug(" Found a time tag '$timetag' with a phase of '$timephasenb' hour in log file name",1);
		# Get older time
		my ($oldersec,$oldermin,$olderhour,$olderday,$oldermonth,$olderyear,$olderwday,$olderyday) = localtime($starttime-($timephasenb*3600));
		my $olderweekofmonth=int($olderday/7);
		my $olderweekofyear=int(($olderyday-1+6-($olderwday==0?6:$olderwday-1))/7)+1; if ($olderweekofyear > 52) { $olderweekofyear = 1; }
		my $olderdaymod=$olderday%7;
		$olderwday++;
		my $olderns=Time::Local::timegm(0,0,0,$olderday,$oldermonth,$olderyear);
		if ($olderdaymod <= $olderwday) { if (($olderwday != 7) || ($olderdaymod != 0)) { $olderweekofmonth=$olderweekofmonth+1; } }
		if ($olderdaymod >  $olderwday) { $olderweekofmonth=$olderweekofmonth+2; }
		# Change format of time variables
		$olderweekofmonth = "0$olderweekofmonth";
		if ($olderweekofyear < 10) { $olderweekofyear = "0$olderweekofyear"; }
		if ($olderyear < 100) { $olderyear+=2000; } else { $olderyear+=1900; }
		my $oldersmallyear=$olderyear;$oldersmallyear =~ s/^..//;
		if (++$oldermonth < 10) { $oldermonth = "0$oldermonth"; }
		if ($olderday < 10) { $olderday = "0$olderday"; }
		if ($olderhour < 10) { $olderhour = "0$olderhour"; }
		if ($oldermin < 10) { $oldermin = "0$oldermin"; }
		if ($oldersec < 10) { $oldersec = "0$oldersec"; }
		# Replace tag with new value
		if ($timetag eq 'YYYY') { $SourceString =~ s/%YYYY-$timephase/$olderyear/ig; next; }
		if ($timetag eq 'YY')   { $SourceString =~ s/%YY-$timephase/$oldersmallyear/ig; next; }
		if ($timetag eq 'MM')   { $SourceString =~ s/%MM-$timephase/$oldermonth/ig; next; }
		if ($timetag eq 'MO')   { $SourceString =~ s/%MO-$timephase/$MonthNumLibEn{$oldermonth}/ig; next; }
		if ($timetag eq 'DD')   { $SourceString =~ s/%DD-$timephase/$olderday/ig; next; }
		if ($timetag eq 'HH')   { $SourceString =~ s/%HH-$timephase/$olderhour/ig; next; }
		if ($timetag eq 'NS')   { $SourceString =~ s/%NS-$timephase/$olderns/ig; next; }
		if ($timetag eq 'WM')   { $SourceString =~ s/%WM-$timephase/$olderweekofmonth/g; next; }
		if ($timetag eq 'Wm')   { my $olderweekofmonth0=$olderweekofmonth-1; $SourceString =~ s/%Wm-$timephase/$olderweekofmonth0/g; next; }
		if ($timetag eq 'WY')   { $SourceString =~ s/%WY-$timephase/$olderweekofyear/g; next; }
		if ($timetag eq 'Wy')   { my $olderweekofyear0=sprintf("%02d",$olderweekofyear-1); $SourceString =~ s/%Wy-$timephase/$olderweekofyear0/g; next; }
		if ($timetag eq 'DW')   { $SourceString =~ s/%DW-$timephase/$olderwday/g; next; }
		if ($timetag eq 'Dw')   { my $olderwday0=$olderwday-1; $SourceString =~ s/%Dw-$timephase/$olderwday0/g; next; }
		# If unknown tag
		error("Unknown tag '\%$timetag' in parameter.");
	}
	# Replace %YYYY %YY %MM %DD %HH with current value. Kept for backward compatibility.
	$SourceString =~ s/%YYYY/$nowyear/ig;
	$SourceString =~ s/%YY/$nowsmallyear/ig;
	$SourceString =~ s/%MM/$nowmonth/ig;
	$SourceString =~ s/%MO/$MonthNumLibEn{$nowmonth}/ig;
	$SourceString =~ s/%DD/$nowday/ig;
	$SourceString =~ s/%HH/$nowhour/ig;
	$SourceString =~ s/%NS/$nowns/ig;
	$SourceString =~ s/%WM/$nowweekofmonth/g;
	my $nowweekofmonth0=$nowweekofmonth-1; $SourceString =~ s/%Wm/$nowweekofmonth0/g;
	$SourceString =~ s/%WY/$nowweekofyear/g;
	my $nowweekofyear0=$nowweekofyear-1; $SourceString =~ s/%Wy/$nowweekofyear0/g;
	$SourceString =~ s/%DW/$nowwday/g;
	my $nowwday0=$nowwday-1; $SourceString =~ s/%Dw/$nowwday0/g;

	debug("Finished Substitute_Tags on '$SourceString'");
	return $SourceString;
}


#------------------------------------------------------------------------------
# Function:     Check if all parameters are correctly defined. If not set them to default.
# Parameters:	None
# Input:		All global variables
# Output:		Change on some global variables
# Return:		None
#------------------------------------------------------------------------------
sub Check_Config
{
	debug("Call to Check_Config");

	# Show initial values of main parameters before check
	if ($Debug)
	{
		debug(" LogFile='$LogFile'",2);
		debug(" LogType='$LogType'",2);
		debug(" LogFormat='$LogFormat'",2);
		debug(" LogSeparator='$LogSeparator'",2);
		debug(" DirData='$DirData'",2);
		debug(" NotPageList ".(join(',',keys %NotPageList)),2);
		debug(" ValidHTTPCodes ".(join(',',keys %ValidHTTPCodes)),2);
		debug(" BuildReportFormat=$BuildReportFormat",2);
		debug(" BuildHistoryFormat=$BuildHistoryFormat",2);
		debug(" URLWithQueryWithOnlyFollowingParameters=".(join(',',@URLWithQueryWithOnly)),2);
		debug(" URLWithQueryWithoutFollowingParameters=".(join(',',@URLWithQueryWithout)),2);
	}

	# Main section
	if ($HTMLOutput{'urlcloudmap'} )
	{
    	$LogFile = "placeholder";    # ToDo: investigate whether this is really required.
    	                             # No reason to care where the initial log files are when running a report
    }else{
    	$LogFile=&Substitute_Tags($LogFile);
    	if (! $LogFile)   { error("LogFile parameter is not defined in config/domain file"); }
    }
	if ($LogType !~ /[WSMFC]/i) { $LogType='W'; }
	$LogFormat =~ s/\\//g;
	if (! $LogFormat) { error("LogFormat parameter is not defined in config/domain file"); }
	if ($LogFormat =~ /^\d$/ && $LogFormat !~ /[0-6]/)  { error("LogFormat parameter is wrong in config/domain file. Value is '$LogFormat' (should be 0,1,2,3,4,5 or a 'personalized AWScloudMap log format string')"); }
	$LogSeparator||="\\s";
	$DirData||='.';
	if (! $SiteDomain)                              { error("SiteDomain parameter not defined in your config/domain file. You must edit it for using this version of AWScloudMap."); }
	if ($AllowToUpdateStatsFromBrowser !~ /[0-1]/) 	{ $AllowToUpdateStatsFromBrowser=0; }
	# Optional setup section
	if (! $SectionsToBeSaved) 						{ $SectionsToBeSaved='all'; }
	if ($EnableLockForUpdate !~ /[0-1]/)           	{ $EnableLockForUpdate=0; }
	if ($CreateDirDataIfNotExists !~ /[0-1]/)      	{ $CreateDirDataIfNotExists=0; }
	if ($RecursiveDirs !~ /[0-1]/)      	        { $RecursiveDirs=0; }
	if ($BuildReportFormat !~ /html|xhtml|xml/i) 	{ $BuildReportFormat='html'; }
	if ($BuildHistoryFormat !~ /text|xml/) 			{ $BuildHistoryFormat='text'; }
	if ($SaveDatabaseFilesWithPermissionsForEveryone !~ /[0-1]/)	{ $SaveDatabaseFilesWithPermissionsForEveryone=0; }
	if ($PurgeLogFile !~ /[0-1]/)                 	{ $PurgeLogFile=0; }
	if ($KeepBackupOfHistoricFiles !~ /[0-1]/)     	{ $KeepBackupOfHistoricFiles=0; }
	$DefaultFile[0]||='index.html';
	if ($URLNotCaseSensitive !~ /[0-1]/)           	{ $URLNotCaseSensitive=0; }
	if ($URLWithAnchor !~ /[0-1]/)                 	{ $URLWithAnchor=0; }
	$URLQuerySeparators =~ s/\s//g;
	if (! $URLQuerySeparators)                 		{ $URLQuerySeparators='?;'; }
	if ($URLWithQuery !~ /[0-1]/)                 	{ $URLWithQuery=0; }
	if ($WarningMessages !~ /[0-1]/)              	{ $WarningMessages=1; }
	if ($NbOfLinesForCorruptedLog !~ /^\d+/ || $NbOfLinesForCorruptedLog<1)	{ $NbOfLinesForCorruptedLog=50; }
	# Optional extra setup section
	foreach my $extracpt (1..@ExtraName-1)
	{
		if ($ExtraStatTypes[$extracpt] !~ /[PHBL]/)  { $ExtraStatTypes[$extracpt]='PHBL'; }
		if ($MaxNbOfExtra[$extracpt] !~ /^\d+$/ || $MaxNbOfExtra[$extracpt]<1) { $MaxNbOfExtra[$extracpt]=20; }
		if ($MinHitExtra[$extracpt] !~ /^\d+$/ || $MinHitExtra[$extracpt]<1) { $MinHitExtra[$extracpt]=1; }
		if (! $ExtraFirstColumnValues[$extracpt]) { error("Extra section number $extracpt is defined without ExtraSectionFirstColumnValues$extracpt parameter"); }
		if (! $ExtraFirstColumnFormat[$extracpt]) { $ExtraFirstColumnFormat[$extracpt] = '%s'; }
	}
	# Optional appearance setup section
	if ($MaxRowsInHTMLOutput !~ /^\d+/ || $MaxRowsInHTMLOutput<1)     { $MaxRowsInHTMLOutput=1000; }
	my @maxnboflist=('Domain','HostsShown','LoginShown','RobotShown','WormsShown','PageShown','OsShown','BrowsersShown','ScreenSizesShown','RefererShown','KeyphrasesShown','KeywordsShown','EMailsShown');
	my @maxnboflistdefaultval=(10,10,10,10,5,10,10,10,5,10,10,10,20);
	foreach my $i (0..(@maxnboflist-1))
	{
		if (! $MaxNbOf{$maxnboflist[$i]} || $MaxNbOf{$maxnboflist[$i]} !~ /^\d+$/ || $MaxNbOf{$maxnboflist[$i]}<1) 	{ $MaxNbOf{$maxnboflist[$i]}=$maxnboflistdefaultval[$i]; }
	}
	my @minhitlist=('Domain','Host','Login','Robot','Worm','File','Os','Browser','ScreenSize','Refer','Keyphrase','Keyword','EMail');
	my @minhitlistdefaultval=(1,1,1,1,1,1,1,1,1,1,1,1,1);
	foreach my $i (0..(@minhitlist-1))
	{
		if (! $MinHit{$minhitlist[$i]} || $MinHit{$minhitlist[$i]} !~ /^\d+$/ || $MinHit{$minhitlist[$i]}<1) 	{ $MinHit{$minhitlist[$i]}=$minhitlistdefaultval[$i]; }
	}
	if ($FirstDayOfWeek !~ /[01]/)               	{ $FirstDayOfWeek=1; }
	if ($ShowLinksOnUrl !~ /[01]/)               	{ $ShowLinksOnUrl=1; }
	if ($MaxLengthOfShownURL !~ /^\d+/ || $MaxLengthOfShownURL<1) { $MaxLengthOfShownURL=64; }
	$Logo||='AWScloudMapLogo.png';
	$LogoLink||='http://cloudMap.calcResult.com';

	# Convert extra sections data into @ExtraConditionType, @ExtraConditionTypeVal...
	foreach my $extranum (1..@ExtraName-1)
	{
		my $part=0;
		foreach my $conditioncouple (split(/\s*\|\|\s*/, $ExtraCondition[$extranum]))
		{
	 		my ($conditiontype, $conditiontypeval)=split(/[,:]/,$conditioncouple,2);
	 		$ExtraConditionType[$extranum][$part]=$conditiontype;
			if ($conditiontypeval =~ /^REGEX\[(.*)\]$/i) { $conditiontypeval=$1; }
			#else { $conditiontypeval=quotemeta($conditiontypeval); }
	 		$ExtraConditionTypeVal[$extranum][$part]=qr/$conditiontypeval/i;
			$part++;
	 	}
		$part=0;
		foreach my $rowkeycouple (split(/\s*\|\|\s*/, $ExtraFirstColumnValues[$extranum]))
		{
	 		my ($rowkeytype, $rowkeytypeval)=split(/[,:]/,$rowkeycouple,2);
	 		$ExtraFirstColumnValuesType[$extranum][$part]=$rowkeytype;
			if ($rowkeytypeval =~ /^REGEX\[(.*)\]$/i) { $rowkeytypeval=$1; }
			#else { $rowkeytypeval=quotemeta($rowkeytypeval); }
	 		$ExtraFirstColumnValuesTypeVal[$extranum][$part]=qr/$rowkeytypeval/i;
			$part++;
	 	}
	}

	# Show definitive value for major parameters
	if ($Debug)
	{
		debug(" LogFile='$LogFile'",2);
		debug(" LogFormat='$LogFormat'",2);
		debug(" LogSeparator='$LogSeparator'",2);
		debug(" DirData='$DirData'",2);
		debug(" SiteDomain='$SiteDomain'",2);
		debug(" TestDomain='$TestDomain'",2);
		debug(" RemoveFromTitle='@RemoveFromTitle'",2);
		foreach (keys %MaxNbOf) { debug(" MaxNbOf{$_}=$MaxNbOf{$_}",2); }
		foreach (keys %MinHit)  { debug(" MinHit{$_}=$MinHit{$_}",2); }
		foreach my $extranum (1..@ExtraName-1)
		{
			debug(" ExtraCodeFilter[$extranum] is array ".join(',',@{$ExtraCodeFilter[$extranum]}),2);
			debug(" ExtraConditionType[$extranum] is array ".join(',',@{$ExtraConditionType[$extranum]}),2);
			debug(" ExtraConditionTypeVal[$extranum] is array ".join(',',@{$ExtraConditionTypeVal[$extranum]}),2);
			debug(" ExtraFirstColumnFunction[$extranum] is array ".join(',',@{$ExtraFirstColumnFunction[$extranum]}),2);
			debug(" ExtraFirstColumnValuesType[$extranum] is array ".join(',',@{$ExtraFirstColumnValuesType[$extranum]}),2);
			debug(" ExtraFirstColumnValuesTypeVal[$extranum] is array ".join(',',@{$ExtraFirstColumnValuesTypeVal[$extranum]}),2);
		}
	}

	# Deny URLWithQueryWithOnlyFollowingParameters and URLWithQueryWithoutFollowingParameters both set
	if (@URLWithQueryWithOnly && @URLWithQueryWithout)
	{
		error("URLWithQueryWithOnlyFollowingParameters and URLWithQueryWithoutFollowingParameters can't be both set at the same time");
	}

	# Deny LogFile if contains a pipe and PurgeLogFile || ArchiveLogRecords set on
	if (($PurgeLogFile || $ArchiveLogRecords) && $LogFile =~ /\|\s*$/)
	{
		error("A pipe in log file name is not allowed if PurgeLogFile and ArchiveLogRecords are not set to 0");
	}
	# Check if DirData is OK
	if (! -d $DirData)
	{
		if ($CreateDirDataIfNotExists)
		{
			debug("Making directory $DirData",2);
			my $mkdirok=mkdir "$DirData", 0766;
			if (! $mkdirok) { error("$PROG failed to create directory DirData (DirData=\"$DirData\", CreateDirDataIfNotExists=$CreateDirDataIfNotExists)."); }
		} else {
			error("AWScloudMap database directory defined in config file by 'DirData' parameter ($DirData) does not exist or is not writable, and can't be created.");
		}
	}
	if ($LogType eq 'S') { $NOT_SORTED_RECORD_TOLERANCE=1000000; }
}


#------------------------------------------------------------------------------
# Function:     Return a checksum for an array of string
# Parameters:	Array of string
# Input:		None
# Output:		None
# Return: 		Checksum number
#------------------------------------------------------------------------------
sub CheckSum
{
	my $string=shift;
	my $checksum=0;
#	use MD5;
# 	$checksum = MD5->hexhash($string);
	my $i=0; my $j=0;
	while ($i < length($string)) {
		my $c=substr($string,$i,1);
		$checksum+=(ord($c)<<(8*$j));
		if ($j++ > 3) { $j=0; }
		$i++;
	}
 	return $checksum;
}



#------------------------------------------------------------------------------
# Function:		Read history file and create or update tmp history file
# Parameters:	year,month,day,hour,withupdate,withpurge,part_to_load[,lastlinenb,lastlineoffset,lastlinechecksum]
# Input:		$DirData $PROG $FileSuffix $LastLine $DatabaseBreak
# Output:		None
# Return:		Tmp history file name created/updated or '' if withupdate is 0
#------------------------------------------------------------------------------
sub Read_History_With_TmpUpdate
{
	my $year=sprintf("%04i",shift||0);
	my $month=sprintf("%02i",shift||0);
	my $day=shift; if ($day ne '')  { $day=sprintf("%02i",$day); }
	my $hour=shift; if ($hour ne '') { $hour=sprintf("%02i",$hour); }
	my $withupdate=shift||0;
	my $withpurge=shift||0;
	my $part=shift||'';

	my ($date,$filedate)=('','');
	if    ($DatabaseBreak eq 'month') { $date=sprintf("%04i%02i",$year,$month); $filedate=sprintf("%02i%04i",$month,$year); }
	elsif ($DatabaseBreak eq 'year')  { $date=sprintf("%04i%",$year); $filedate=sprintf("%04i",$year); }
	elsif ($DatabaseBreak eq 'day')   { $date=sprintf("%04i%02i%02i",$year,$month,$day); $filedate=sprintf("%02i%04i%02i",$month,$year,$day); }
	elsif ($DatabaseBreak eq 'hour')  { $date=sprintf("%04i%02i%02i%02i",$year,$month,$day,$hour); $filedate=sprintf("%02i%04i%02i%02i",$month,$year,$day,$hour); }

# print "\n\$date:".$date;

	my $xml=($BuildHistoryFormat eq 'xml'?1:0);
	my $xmleb='</table><nu>';  ## !! MJB
	my $xmlrb='<tr><td>';

	my $lastLineNum=shift||0;
	my $lastlineoffset=shift||0;
	my $lastlinechecksum=shift||0;

	my %allsections=('DATA'=>1);

	my $order=(scalar keys %allsections)+1;
	my $withread=0;

	# Variable used to read old format history files
	my $readvisitorforbackward=0;

	debug("Call to Read_History_With_TmpUpdate [$year,$month,$day,$hour,withupdate=$withupdate,withpurge=$withpurge,part=$part,lastlinenb=$lastLineNum,lastlineoffset=$lastlineoffset,lastlinechecksum=$lastlinechecksum]");
	debug("date=$date");

	my %SectionsToLoad = ();
	my $order=1;
	$SectionsToLoad{'time'}=$order++;	# Needed to count TotalPages, TotalBandwidth
	$SectionsToLoad{'DATA'}=$order++;	# Always loaded

	# Define SectionsToSave (which sections to save)
	my %SectionsToSave = ();
	if ($withupdate)
	{
		%SectionsToSave=%allsections;
	}

    debug(" List of sections marked for load : ".join(' ',(sort { $SectionsToLoad{$a} <=> $SectionsToLoad{$b} } keys %SectionsToLoad)),1);
    debug(" List of sections marked for save : ".join(' ',(sort { $SectionsToSave{$a} <=> $SectionsToSave{$b} } keys %SectionsToSave)),1);

	# Define value for filetowrite and filetoread (Month before Year kept for backward compatibility)
	my $filetowrite='';
	my $filetoread='';
	if ($HistoryAlreadyFlushed{"$year$month$day$hour"} && -s "$DirData/$PROG$filedate$FileSuffix.tmp.$$")
	{
		# tmp history file was already flushed
		$filetoread="$DirData/$PROG$filedate$FileSuffix.tmp.$$";
		$filetowrite="$DirData/$PROG$filedate$FileSuffix.tmp.$$.bis";
	} else {
		$filetoread="$DirData/$PROG$filedate$FileSuffix.txt";
		$filetowrite="$DirData/$PROG$filedate$FileSuffix.tmp.$$";
	}
	debug(" History file to read is '$filetoread'",1);
#	print "\n History file to read is '$filetoread'";

	# Is there an old data file to read / can we open the file for read
	if (-s $filetoread ) { $withread=1; }

	# Open files
	if ($withread)
	{
		open(HISTORY,$filetoread) || error("Couldn't open file \"$filetoread\" for read: $!","","");
		binmode HISTORY;	# Avoid premature EOF due to history files corrupted with \cZ or bin chars
	}
	if ($withupdate)
	{
		open(HISTORYTMP,">$filetowrite") || error("Couldn't open file \"$filetowrite\" for write: $!");
		binmode HISTORYTMP;
		Save_History("header",$year,$month,$date);
	}

	# Loop on read file
	my $readxml=0;
	if ($withread)
	{
		my $countlines=0;
		my $versionnum=0;
		my @field=();
		while (<HISTORY>)
		{
			chomp $_; s/\r//;
			$countlines++;

			# Test if it's xml
			if (! $readxml && $_ =~ /^<xml/)
			{
				$readxml=1;
				debug(" Data file format is 'xml'",1);
				next;
			}

			# Analyze fields
			@field=split(/\s+/,($readxml?CleanFromTags($_):$_));
			if (! $field[0]) { next; }

			# BEGIN_GENERAL
			if ($field[0] eq 'BEGIN_GENERAL')
			{
				debug(" Begin of GENERAL section");
				next; # Don't need to do anything with this log-line
			}
			if ($field[0] eq 'LastLine' || $field[0] eq "${xmlrb}LastLine")
			{
				if (! $LastLine || $LastLine < int($field[1])) { $LastLine=int($field[1]); };
				if ($field[2]) { $LastLineNumber=int($field[2]); }
				if ($field[3]) { $LastLineOffset=int($field[3]); }
				if ($field[4]) { $LastLineChecksum=int($field[4]); }
				next;
			}
			if ($field[0] eq 'FirstTime' || $field[0] eq "${xmlrb}FirstTime")      { if (! $FirstTime{$date} || $FirstTime{$date} > int($field[1])) { $FirstTime{$date}=int($field[1]); }; next; }
			if ($field[0] eq 'LastTime' || $field[0] eq "${xmlrb}LastTime")        { if (! $LastTime{$date} || $LastTime{$date} < int($field[1])) { $LastTime{$date}=int($field[1]); }; next; }
			if ($field[0] eq 'LastUpdate' || $field[0] eq "${xmlrb}LastUpdate")
			{
				if ($LastUpdate < $field[1])
				{
					$LastUpdate=int($field[1]);
					#$LastUpdateLinesRead=int($field[2]);
					#$LastUpdateNewLinesRead=int($field[3]);
					#$LastUpdateLinesCorrupted=int($field[4]);
				};
				next;
			}
			if (($field[0] eq 'END_GENERAL' || $field[0] eq "${xmleb}END_GENERAL"))
			{
				debug(" End of GENERAL section");

				if (! scalar %SectionsToLoad) { debug(" Stop reading history file. Got all we need."); last; }
				next;
			}

			# BEGIN_DATA
			if ($field[0] eq $FILE_START_MARKER)
			{
				debug("Starting to write DATA section", 1);
				$field[0]='';
				my $count=0;my $countloaded=0;
				do
				{
					if ($field[0])
					{
						$count++;
						if ($SectionsToLoad{'DATA'})
						{
							my $loadrecord=0;
							if ($withupdate)
							{
								$loadrecord=1;
							} else {
								if ($HTMLOutput{'main'})
								{
									if ($MonthRequired eq 'all') { $loadrecord=1; }
									else {
										if ($countloaded < $MaxNbOf{'PageShown'} && $field[1] >= $MinHit{'File'}) { $loadrecord=1; }
									}
								} else {
								    # This is for $HTMLOutput = urlcloudmap
									if ($MonthRequired eq 'all' )
									{
										if ((!$FilterIn{'url'} || $field[0] =~ /$FilterIn{'url'}/)
										 && (!$FilterEx{'url'} || $field[0] !~ /$FilterEx{'url'}/)) { $loadrecord=1; }
									} else {
										if ((!$FilterIn{'url'} || $field[0] =~ /$FilterIn{'url'}/)
										 && (!$FilterEx{'url'} || $field[0] !~ /$FilterEx{'url'}/)
										 && $field[1] >= $MinHit{'File'}) { $loadrecord=1; }
										$TotalDifferentPages++;
									}
								}
							}
							if ($loadrecord)
							{
								if ($field[1]) { $_url_p{$field[0]}+=$field[1]; }
								my $fieldArrayLength = scalar(@field);
								my $thisPageTitle = '';
                                for (my $i=2; $i<$fieldArrayLength; $i++)
                                {
    								$thisPageTitle .= $field[$i]." ";
                                }
								$_url_title{$field[0]} = $thisPageTitle;
								$countloaded++;
							}
						}
					}
					$_=<HISTORY>;
					chomp $_; s/\r//;
					@field=split(/\s+/,($readxml?CleanFromTags($_):$_)); $countlines++;
				}
				until ($field[0] eq 'END_DATA' || $field[0] eq "${xmleb}END_DATA" || ! $_);
				if ($field[0] ne 'END_DATA' && $field[0] ne "${xmleb}END_DATA") { error("History file \"$filetoread\" is corrupted (End of section SIDER not found).\nRestore a recent backup of this file (data for this month will be restored to backup date), remove it (data for month will be lost), or remove the corrupted section in file (data for at least this section will be lost).","","",1); }
# print"\n End of DATA section ($count entries, $countloaded loaded)\n";
				debug(" End of DATA section ($count entries, $countloaded loaded)");
				delete $SectionsToLoad{'DATA'};
				# WE DO NOT SAVE SECTION NOW BECAUSE VALUES CAN BE CHANGED AFTER READING VISITOR
				#if ($SectionsToSave{'sider'}) {
				#	Save_History('sider',$year,$month,$date); delete $SectionsToSave{'sider'};
				#	if ($withpurge) { %_url_p=(); %_url_k=(); %_url_e=(); %_url_x=(); }
				#}
				if (! scalar %SectionsToLoad) { debug(" Stop reading history file. Got all we need."); last; }
				next;
			}

			# BEGIN_PLUGINS
           	if ($AtLeastOneSectionPlugin && $field[0] =~ /^BEGIN_PLUGIN_(\w+)$/i)
           	{
          	    my $pluginname=$1;
          	    my $found=0;
   				if (! scalar %SectionsToLoad) { debug(" Stop reading history file. Got all we need."); last; }
                # The plugin for this section was not loaded
   				if (! $found)
   				{
					do {
						$_=<HISTORY>;
						chomp $_; s/\r//;
						@field=split(/\s+/,($readxml?CleanFromTags($_):$_)); $countlines++;
					}
					until ($field[0] eq "END_PLUGIN_$pluginname" || $field[0] eq "${xmleb}END_PLUGIN_$pluginname" || ! $_);
   				}
   				next;
            }

		}
	}

	if ($withupdate)
	{
		# Process rest of data saved in 'wait' arrays (data for hosts that are not in history file or no history file found)
		# This can change some values for DATA section
		debug(" Processing data in 'wait' arrays",3);
		foreach (keys %_waithost_e)
		{
			debug("  Visit in 'wait' array for $_ is a new visit",4);
			my $newtimehosts=($_waithost_s{$_}?$_waithost_s{$_}:$_host_s{$_});
			my $newtimehostl=($_waithost_l{$_}?$_waithost_l{$_}:$_host_l{$_});
			$newtimehosts =~ /^(\d\d\d\d\d\d\d\d)/; $DayVisits{$1}++;
			if ($_waithost_s{$_})
			{
				# There was also a second session in processed log
				$_session{GetSessionRange($newtimehosts,$newtimehostl)}++;
			}
		}
	}

	debug(" Check and write all unwrote sections: ".join(',',keys %SectionsToSave),2);
	foreach my $key (sort { $SectionsToSave{$a} <=> $SectionsToSave{$b} } keys %SectionsToSave)
	{
		debug("Writing section: $key",1);
		Save_History("$key",$year,$month,$date,$lastLineNum,$lastlineoffset,$lastlinechecksum);
	}
	%SectionsToSave=();

	# Update offset in map section and last data in general section then close files
	if ($withupdate)
	{
		# Update offset of sections in the MAP section
		foreach (sort { $PosInFile{$a} <=> $PosInFile{$b} } keys %ValueInFile)
		{
			debug(" Update offset of section $_=$ValueInFile{$_} in file at offset $PosInFile{$_}");
			if ($PosInFile{"$_"}) {
				seek(HISTORYTMP,$PosInFile{"$_"},0); print HISTORYTMP $ValueInFile{"$_"};
			}
		}
		# Save last data in general sections
		debug(" Update MonthHostsKnown=$MonthHostsKnown{$year.$month} in file at offset $PosInFile{MonthHostsKnown}");
		seek(HISTORYTMP,$PosInFile{"MonthHostsKnown"},0); print HISTORYTMP $MonthHostsKnown{$year.$month};
		debug(" Update MonthHostsUnknown=$MonthHostsUnknown{$year.$month} in file at offset $PosInFile{MonthHostsUnknown}");
		seek(HISTORYTMP,$PosInFile{"MonthHostsUnknown"},0); print HISTORYTMP $MonthHostsUnknown{$year.$month};
		close(HISTORYTMP) || error("Failed to write temporary history file");
	}
	if ($withread)
	{
		close(HISTORY) || error("Command for pipe '$filetoread' failed");
	}

	# Purge data
	if ($withpurge) { &Init_HashArray(); }

	# If update, rename tmp file bis into tmp file or set HistoryAlreadyFlushed
	if ($withupdate)
	{
		if ($HistoryAlreadyFlushed{"$year$month$day$hour"})
		{
			debug("Rename tmp history file bis '$filetoread' to '$filetowrite'");
			if (rename($filetowrite,$filetoread)==0)
			{
				error("Failed to update tmp history file $filetoread");
			}
		} else {
			$HistoryAlreadyFlushed{"$year$month$day$hour"}=1;
		}

		if (! $ListOfYears{"$year"} || $ListOfYears{"$year"} lt "$month") { $ListOfYears{"$year"}="$month"; }
	}

	# For backward compatibility, if LastLine does not exist, set to LastTime
	$LastLine||=$LastTime{$date};

	return ($withupdate?"$filetowrite":"");
}

#------------------------------------------------------------------------------
# Function:		Save a part of history file
# Parameters:	sectiontosave,year,month,breakdate[,lastlinenb,lastlineoffset,lastlinechecksum]
# Input:		$VERSION HISTORYTMP $nowyear $nowmonth $nowday $nowhour $nowmin $nowsec $LastLineNumber $LastLineOffset $LastLineChecksum
# Output:		None
# Return:		None
#------------------------------------------------------------------------------
sub Save_History
{
	my $sectiontosave=shift||'';
	my $year         =shift||'';
	my $month        =shift||'';
	my $breakdate    =shift||'';

	my $xml=($BuildHistoryFormat eq 'xml'?1:0);
	my ($xmlbb,$xmlbs,$xmlbe,$xmlhb,$xmlhs,$xmlhe,$xmlrb,$xmlrs,$xmlre,$xmleb,$xmlee)=('','','','','','','','','','','');
	if ($xml) { ($xmlbb,$xmlbs,$xmlbe,$xmlhb,$xmlhs,$xmlhe,$xmlrb,$xmlrs,$xmlre,$xmleb,$xmlee)=("</comment><nu>\n",'</nu><recnb>','</recnb><table>','<tr><th>','</th><th>','</th></tr>','<tr><td>','</td><td>','</td></tr>','</table><nu>',"\n</nu></section>" ); }
	else { $xmlbs=' '; $xmlhs=' '; $xmlrs=' '; }

	my $lastLineNum=shift||0;
	my $lastlineoffset=shift||0;
	my $lastlinechecksum=shift||0;
	if (! $lastLineNum)
	{
		$lastLineNum=$LastLineNumber;
		$lastlineoffset=$LastLineOffset;
		$lastlinechecksum=$LastLineChecksum;
	}
	my $bunchOfSpaces="                    ";
	my %keysinkeylist=();

	# Header
	if ($sectiontosave eq 'header')
	{
		print HISTORYTMP "AWScloudMap DATA FILE Version:$VERSION\n";
		print HISTORYTMP "# If you remove this file, all statistics for year:$year /month:$month will be lost/reset.\n";
		print HISTORYTMP "\n";
	}

	if ($sectiontosave eq 'DATA')
	{
		print HISTORYTMP "\n";
		print HISTORYTMP "# URL - Hits - Page Title\n";
		$ValueInFile{$sectiontosave}=tell HISTORYTMP;
		print HISTORYTMP "${xmlbb}$FILE_START_MARKER${xmlbs}".(scalar keys %_url_p)."${xmlbe}\n";
		&BuildKeyListNoSort($MaxNbOf{'PageShown'},$MinHit{'File'},\%_url_p,\%_url_p);
		%keysinkeylist=();
		foreach (@keylist)
		{
			$keysinkeylist{$_}=1;
			my $newkey=$_;
			$newkey =~ s/([^:])\/\//$1\//g;		# Because some targeted url may have been recorded with two slashes: (Ex: //rep//file.htm). We must keep http://rep/file.htm
			sleep $throttleRequestsTime;
			print HISTORYTMP "${xmlrb}".Generate_Alias_List_Entry($newkey, int($_url_p{$_}||0) )."${xmlre}\n";
		}
		foreach (keys %_url_p)
		{
			if ($keysinkeylist{$_}) { next; }
			my $newkey=$_;
			$newkey =~ s/([^:])\/\//$1\//g;		# Because some targeted url may have been recorded with two slashes: (Ex: //rep//file.htm). We must keep http://rep/file.htm
			print HISTORYTMP "${xmlrb}".Generate_Alias_List_Entry($newkey, int($_url_p{$_}||0) )."${xmlre}\n";
			# print HISTORYTMP "${xmlrb}$newkey ".int($_url_p{$_}||0)."${xmlre}\n";
		}
		print HISTORYTMP "${xmleb}END_DATA${xmlee}\n";
	}
	%keysinkeylist=();
}

#--------------------------------------------------------------------
# Function:     Rename all tmp history file into history
# Parameters:   None
# Input:        $DirData $PROG $FileSuffix
#               $KeepBackupOfHistoricFile $SaveDatabaseFilesWithPermissionsForEveryone
# Output:       None
# Return:       1 Ok, 0 at least one error (tmp files are removed)
#--------------------------------------------------------------------
sub Rename_All_Tmp_History
{
	my $pid=$$;
	my $renameok=1;

	debug("Call to Rename_All_Tmp_History (FileSuffix=$FileSuffix)");

	opendir(DIR,"$DirData");

	my $datemask;
	if    ($DatabaseBreak eq 'month') { $datemask='\d\d\d\d\d\d'; }
	elsif ($DatabaseBreak eq 'year')  { $datemask='\d\d\d\d'; }
	elsif ($DatabaseBreak eq 'day')   { $datemask='\d\d\d\d\d\d\d\d'; }
	elsif ($DatabaseBreak eq 'hour')  { $datemask='\d\d\d\d\d\d\d\d\d\d'; }
	debug("Scan for temp history files to rename into DirData='$DirData' with mask='$datemask'");

	my $regfilesuffix=quotemeta($FileSuffix);
	foreach (grep /^$PROG($datemask)$regfilesuffix\.tmp\.$pid$/, file_filt sort readdir DIR)
	{
		/^$PROG($datemask)$regfilesuffix\.tmp\.$pid$/;
		if ($renameok) {	# No rename error yet
			debug(" Rename new tmp history file $PROG$1$FileSuffix.tmp.$$ into $PROG$1$FileSuffix.txt",1);
			if (-s "$DirData/$PROG$1$FileSuffix.tmp.$$") {		# Rename tmp files if size > 0
				if ($KeepBackupOfHistoricFiles) {
					if (-s "$DirData/$PROG$1$FileSuffix.txt")
					{
					    # History file already exists. We backup it
						debug("  Make a backup of old history file into $PROG$1$FileSuffix.bak before",1);
						if (rename("$DirData/$PROG$1$FileSuffix.txt", "$DirData/$PROG$1$FileSuffix.bak")==0) {
							warning("Warning: Failed to make a backup of \"$DirData/$PROG$1$FileSuffix.txt\" into \"$DirData/$PROG$1$FileSuffix.bak\".");
						}
						if ($SaveDatabaseFilesWithPermissionsForEveryone) {
							chmod 0666,"$DirData/$PROG$1$FileSuffix.bak";
						}
					}
					else {
						debug("  No need to backup old history file",1);
					}
				}
				if (rename("$DirData/$PROG$1$FileSuffix.tmp.$$", "$DirData/$PROG$1$FileSuffix.txt")==0) {
					$renameok=0;	# At least one error in renaming working files
					# Remove tmp file
					unlink "$DirData/$PROG$1$FileSuffix.tmp.$$";
					warning("Warning: Failed to rename \"$DirData/$PROG$1$FileSuffix.tmp.$$\" into \"$DirData/$PROG$1$FileSuffix.txt\".\nWrite permissions on \"$PROG$1$FileSuffix.txt\" might be wrong".($ENV{'GATEWAY_INTERFACE'}?" for an 'update from web'":"")." or file might be opened.");
					next;
				}
				if ($SaveDatabaseFilesWithPermissionsForEveryone) {
					chmod 0666,"$DirData/$PROG$1$FileSuffix.txt";
				}
			}
		}
		else {				# Because of rename error, we remove all remaining tmp files
			unlink "$DirData/$PROG$1$FileSuffix.tmp.$$";
		}
	}
	close DIR;
	return $renameok;
}

#------------------------------------------------------------------------------
# Function:     Return time elapsed since last call in miliseconds
# Parameters:	0|1 (0 reset counter, 1 no reset)
# Input:		None
# Output:		None
# Return:		Number of miliseconds elapsed since last call
#------------------------------------------------------------------------------
sub GetDelaySinceStart
{
	if (shift) { $StartSeconds=0; }	# Reset chrono
	my ($newseconds, $newmicroseconds)=(time(),0);
	# Plugin call : Return seconds and milliseconds
	if (! $StartSeconds) { $StartSeconds=$newseconds; $StartMicroseconds=$newmicroseconds; }
	return (($newseconds-$StartSeconds)*1000+int(($newmicroseconds-$StartMicroseconds)/1000));
}

#------------------------------------------------------------------------------
# Function:     Reset all variables whose name start with _ because a new month start
# Parameters:	None
# Input:        $YearRequired All variables whose name start with _
# Output:       All variables whose name start with _
# Return:		None
#------------------------------------------------------------------------------
sub Init_HashArray
{
	debug("Call to Init_HashArray");
	# Reset global hash arrays
	%FirstTime = %LastTime = ();
	%MonthHostsKnown = %MonthHostsUnknown = ();
	%MonthVisits = %MonthUnique = ();
	%MonthPages = %MonthHits = %MonthBytes = ();
	%MonthNotViewedPages = %MonthNotViewedHits = %MonthNotViewedBytes = ();
	%DayPages = %DayHits = %DayBytes = %DayVisits = ();
	# Reset all arrays with name beginning by _
	for (my $ix=0; $ix<6; $ix++)  { $_from_p[$ix]=0; $_from_h[$ix]=0; }
	for (my $ix=0; $ix<24; $ix++) { $_time_h[$ix]=0; $_time_k[$ix]=0; $_time_p[$ix]=0; $_time_nv_h[$ix]=0; $_time_nv_k[$ix]=0; $_time_nv_p[$ix]=0; }
	# Reset all hash arrays with name beginning by _
	%_session = %_browser_h = ();
	%_domener_p = %_domener_h = %_domener_k = %_errors_k = ();
	%_filetypes_h = %_filetypes_k = %_filetypes_gz_in = %_filetypes_gz_out = ();
	%_host_p = %_host_h = %_host_k = %_host_l = %_host_s = %_host_u = ();
	%_waithost_e = %_waithost_l = %_waithost_s = %_waithost_u = ();
	%_keyphrases = %_keywords = %_os_h = %_pagesrefs_p = %_pagesrefs_h = %_robot_h = %_robot_k = %_robot_l = %_robot_r = ();
	%_worm_h = %_worm_k = %_worm_l = %_login_p = %_login_h = %_login_k = %_login_l = %_screensize_h = ();
 	%_misc_p = %_misc_h = %_misc_k = ();
	%_cluster_p = %_cluster_h = %_cluster_k = ();
	%_url_p = %_url_title = %_url_x = ();
 	for (my $ix=1; $ix < @ExtraName; $ix++)
 	{
 		%{'_section_' . $ix . '_h'} = %{'_section_' . $ix . '_o'} = %{'_section_' . $ix . '_k'}	=
 		%{'_section_' . $ix . '_l'} = %{'_section_' . $ix . '_p'} = ();
 	}
}

#------------------------------------------------------------------------------
# Function:     Change word separators of a keyphrase string into space and
#               remove bad coded chars
# Parameters:	stringtodecode
# Input:        None
# Output:       None
# Return:		decodedstring
#------------------------------------------------------------------------------
sub ChangeWordSeparatorsIntoSpace
{
	$_[0] =~ s/%0[ad]/ /ig;				# LF CR
	$_[0] =~ s/%2[02789abc]/ /ig;		# space " ' ( ) * + ,
	$_[0] =~ s/%3a/ /ig;				# :
	$_[0] =~ tr/\+\'\(\)\"\*,:/        /s;	# "&" and "=" must not be in this list
}


#------------------------------------------------------------------------------
# Function:		Transforms special chars by entities as needed in XML/XHTML
# Parameters:	stringtoencode
# Return:		encodedstring
#------------------------------------------------------------------------------
sub XMLEncode
{
	if ($BuildReportFormat ne 'xhtml' && $BuildReportFormat ne 'xml') { return shift; }
	my $string = shift;
	$string =~ s/&/&amp;/g;
	$string =~ s/</&lt;/g;
	$string =~ s/>/&gt;/g;
    $string =~ s/\"/&quot;/g;
    $string =~ s/\'/&apos;/g;
	return $string;
}


#------------------------------------------------------------------------------
# Function:     Decode an precompiled regex value to a common regex value
# Parameters:   compiledregextodecode
# Input:        None
# Output:       None
# Return:		standardregex
#------------------------------------------------------------------------------
sub UnCompileRegex
{
	shift =~ /\(\?[-\w]*:(.*)\)/;
	return $1;
}

#------------------------------------------------------------------------------
# Function:     Clean a string of all chars that are not char or _ - \ / . \s
# Parameters:   stringtoclean, full
# Input:        None
# Output:       None
# Return:		cleanedstring
#------------------------------------------------------------------------------
sub Sanitize
{
	my $stringtoclean=shift;
	my $full=shift||0;
	if ($full)
	{
	    $stringtoclean =~ s/[^\w]//g;
    } else {
	    $stringtoclean =~ s/[^\w\-\\\/\.:\s]//g;
	}
	return $stringtoclean;
}


#------------------------------------------------------------------------------
# Function:     Shorten the given URL to be easier to display on screen
# Parameters:   stringtoclean
# Input:        None
# Output:       None
# Return:		File-name only - all path info removed
#------------------------------------------------------------------------------
sub formatURL
{
	my $stringtoclean = shift;
	my @sections = split(/[\/\\]/, $stringtoclean);

    if(scalar(@sections) > 0)
    {
        my $retVal = pop(@sections);
    	return $retVal;
    }else{
    	return $stringtoclean;
    }
}

#------------------------------------------------------------------------------
# Function:     Clean a string of unwanted portions, eg Site Name
# Parameters:   stringtoclean, actualURL
# Input:        None
# Output:       None
# Return:		cleanedstring, or cleaned-up version of $actualURL if the string is not meaningful
#------------------------------------------------------------------------------
sub sanitisePageTitle
{
	my $stringtoclean=shift;
	my $actualURL=shift;

	if($stringtoclean eq $MISSING_TITLE_STRING || $stringtoclean eq $MISSING_TITLE_STRING." ") # The second test was required due to a trailing space getting included
	{
        return formatURL($actualURL);   # Just the file name
	}

    if(length($stringtoclean) > $MaxLengthOfShownURL)
    {
        debug("About to trim string: $stringtoclean, original length:".length($stringtoclean)." chars",5);
        $stringtoclean = substr($stringtoclean, 0, ($MaxLengthOfShownURL + 2));
    	$stringtoclean =~ s/ $//;	        # Remove a final space character if it exists, as it would prevent the next regex from doing what is intended:
    	$stringtoclean =~ s/ \w+$/&hellip\;/;	    # Replace last space character (and any following 'word characters') with an ellipsis char
    }
	$stringtoclean =~ s/ /&nbsp\;/g;	# Replace any remaining space characters with non-breaking spaces
	                                    # Do this _after_ we have counted the number of characters in the string, as we are adding five chars per space found
	return $stringtoclean;
}

#------------------------------------------------------------------------------
# Function:     Clean tags in a string
# Parameters:   stringtodecode
# Input:        None
# Output:       None
# Return:		decodedstring
#------------------------------------------------------------------------------
sub CleanFromTags
{
	my $stringtoclean=shift;
	$stringtoclean =~ s/$regclean1/ /g;	# Replace <recnb> or </td> with space
	$stringtoclean =~ s/$regclean2//g;	# Remove <xxx>
	return $stringtoclean;
}


#------------------------------------------------------------------------------
# Function:		Format value in bytes in a string (Bytes, Kb, Mb, Gb)
# Parameters:   bytes (integer value or "0.00")
# Input:        None
# Output:       None
# Return:       "x.yz MB" or "x.yy KB" or "x Bytes" or "0"
#------------------------------------------------------------------------------
sub Format_Bytes
{
	my $bytes = shift||0;
	my $fudge = 1;
	# Do not use exp/log function to calculate 1024power, function make segfault on some unix/perl versions
	if ($bytes >= ($fudge << 30)) { return sprintf("%.2f", $bytes/1073741824)." $Message[22]"; }
	if ($bytes >= ($fudge << 20)) { return sprintf("%.2f", $bytes/1048576)." $Message[21]"; }
	if ($bytes >= ($fudge << 10)) { return sprintf("%.2f", $bytes/1024)." $Message[20]"; }
	if ($bytes < 0) { $bytes="?"; }
	return int($bytes).(int($bytes)?" $Message[23]":"");
}

#------------------------------------------------------------------------------
# Function:		Format a number
# Parameters:   number
# Input:        None
# Output:       None
# Return:       "999 999 999 999"
#------------------------------------------------------------------------------
sub Format_Number
{
	my $number = shift||0;
	$number=~s/(\d)(\d\d\d)$/$1 $2/;
	$number=~s/(\d)(\d\d\d\s\d\d\d)$/$1 $2/;
	$number=~s/(\d)(\d\d\d\s\d\d\d\s\d\d\d)$/$1 $2/;
	return $number;
}


#------------------------------------------------------------------------------
# Function:		Format a date according to Message[19] (country date format)
# Parameters:   String date YYYYMMDDHHMMSS
#               Option 0=LastUpdate and LastTime date
#                      1=Arrays date except daymonthvalues
#                      2=daymonthvalues date (only year month and day)
# Input:        $Message[19]
# Output:       None
# Return:       Date with format defined by Message[19] and option
#------------------------------------------------------------------------------
sub Format_Date
{
	my $date=shift;
	my $option=shift||0;
	my $year=substr("$date",0,4);
	my $month=substr("$date",4,2);
	my $day=substr("$date",6,2);
	my $hour=substr("$date",8,2);
	my $min=substr("$date",10,2);
	my $sec=substr("$date",12,2);
	my $dateformat=$Message[19];
	if ($option == 2) {
		$dateformat =~ s/^[^ymd]+//g;
		$dateformat =~ s/[^ymd]+$//g;
	}
	$dateformat =~ s/yyyy/$year/g;
	$dateformat =~ s/yy/$year/g;
	$dateformat =~ s/mmm/$MonthNumLib{$month}/g;
	$dateformat =~ s/mm/$month/g;
	$dateformat =~ s/dd/$day/g;
	$dateformat =~ s/HH/$hour/g;
	$dateformat =~ s/MM/$min/g;
	$dateformat =~ s/SS/$sec/g;
	return "$dateformat";
}


#------------------------------------------------------------------------------
# Function:     Add a val from sorting tree
# Parameters:   keytoadd keyval [firstadd]
# Input:        None
# Output:       None
# Return:       None
#------------------------------------------------------------------------------
sub AddInTree
{
	my $keytoadd=shift;
	my $keyval=shift;
	my $firstadd=shift||0;
	if ($firstadd==1) {			# Val is the first one
		debug("  firstadd",4);
		$val{$keyval}=$keytoadd;
		$lowerval=$keyval;
		debug("  lowerval=$lowerval, nb elem val=".(scalar keys %val).", nb elem egal=".(scalar keys %egal).".",4);
		return;
	}
	if ($val{$keyval}) { 		# Val is already in tree
		debug("  val is already in tree",4);
		$egal{$keytoadd}=$val{$keyval};
		$val{$keyval}=$keytoadd;
		debug("  lowerval=$lowerval, nb elem val=".(scalar keys %val).", nb elem egal=".(scalar keys %egal).".",4);
		return;
	}
	if ($keyval <= $lowerval) {	# Val is a new one lower (should happens only when tree is not full)
		debug("  keytoadd val=$keyval is lower or equal to lowerval=$lowerval",4);
		$val{$keyval}=$keytoadd;
		$nextval{$keyval}=$lowerval;
		$lowerval=$keyval;
		debug("  lowerval=$lowerval, nb elem val=".(scalar keys %val).", nb elem egal=".(scalar keys %egal).".",4);
		return;
	}
	# Val is a new one higher
	debug("  keytoadd val=$keyval is higher than lowerval=$lowerval",4);
	$val{$keyval}=$keytoadd;
	my $valcursor=$lowerval;	# valcursor is value just before keyval
	while ($nextval{$valcursor} && ($nextval{$valcursor} < $keyval)) { $valcursor=$nextval{$valcursor}; }
	if ($nextval{$valcursor}) {	# keyval is between valcursor and nextval{valcursor}
		$nextval{$keyval}=$nextval{$valcursor};
	}
	$nextval{$valcursor}=$keyval;
	debug("  lowerval=$lowerval, nb elem val=".(scalar keys %val).", nb elem egal=".(scalar keys %egal).".",4);
}

#------------------------------------------------------------------------------
# Function:     Remove a val from sorting tree
# Parameters:   None
# Input:        $lowerval %val %egal
# Output:       None
# Return:       None
#------------------------------------------------------------------------------
sub Removelowerval
{
	my $keytoremove=$val{$lowerval};	# This is lower key
	debug("   remove for lowerval=$lowerval: key=$keytoremove",4);
	if ($egal{$keytoremove}) {
		$val{$lowerval}=$egal{$keytoremove};
		delete $egal{$keytoremove};
	}
	else {
		delete $val{$lowerval};
		$lowerval=$nextval{$lowerval};	# Set new lowerval
	}
	debug("   new lower value=$lowerval, val size=".(scalar keys %val).", egal size=".(scalar keys %egal),4);
}


#------------------------------------------------------------------------------
# Function:     Build @keylist array, Not Sorted
# Parameters:   Size max for @keylist array,
#               Min value in hash for select,
#               Hash used for select,
#               Hash used for order
# Input:        None
# Output:       None
# Return:       @keylist response array
#------------------------------------------------------------------------------
sub BuildKeyListNoSort
{
	debug("In sub BuildKeyListNoSort()",1);
	my $ArraySize=shift||error("System error. Call to BuildKeyListNoSort function with incorrect value for first param","","",1);
	my $MinValue=shift||error("System error. Call to BuildKeyListNoSort function with incorrect value for second param","","",1);
	my $hashforselect=shift;
	my $hashfororder=shift;
	debug("BuildKeyList($ArraySize,$MinValue,$hashforselect has size=".(scalar keys %$hashforselect).",$hashfororder has size=".(scalar keys %$hashfororder).")",1);
	delete $hashforselect->{0};delete $hashforselect->{''};		# This is to protect from infinite loop when hash array has an incorrect null key
	my $count=0;
	$lowerval=0;	# Global because used in AddInTree and Removelowerval
	%val=();
	%nextval=();
	%egal=();
	foreach my $key (keys %$hashforselect)
	{
		if ($count < $ArraySize)
		{
			if ($hashforselect->{$key} >= $MinValue)
			{
				$count++;
				debug("  Add in tree entry $count : $key (value=".($hashfororder->{$key}||0).", tree not full)",4);
				AddInTree($key,$hashfororder->{$key}||0,$count);
			}
			next;
		}
		$count++;
		if (($hashfororder->{$key}||0)<=$lowerval) { next; }
		debug("  Add in tree entry $count : $key (value=".($hashfororder->{$key}||0)." > lowerval=$lowerval)",4);
		AddInTree($key,$hashfororder->{$key}||0);
		debug("  Removelower in tree",4);
		Removelowerval();
	}

	# Build key list (but don't sort it)
	debug("  Build key list and sort it. lowerval=$lowerval, nb elem val=".(scalar keys %val).", nb elem egal=".(scalar keys %egal).".",3);
	my %notsortedkeylist=();
	foreach my $key (values %val) {	$notsortedkeylist{$key}=1; }
	foreach my $key (values %egal) { $notsortedkeylist{$key}=1; }
	@keylist=(keys %notsortedkeylist);

	debug("  BuildKeyList End (keylist size=".(@keylist).")",3);
	return;
}

#------------------------------------------------------------------------------
# Function:     Lock or unlock update
# Parameters:   status (1 to lock, 0 to unlock)
# Input:        $DirLock (if status=0) $PROG $FileSuffix
# Output:       $DirLock (if status=1)
# Return:       None
#------------------------------------------------------------------------------
sub Lock_Update
{
	my $status=shift;
	my $lock="$PROG$FileSuffix.lock";
	if ($status)
	{
		# We stop if there is at least one lock file wherever it is
		foreach my $key ($ENV{"TEMP"},$ENV{"TMP"},"/tmp","/",".")
		{
			my $newkey =$key;
			$newkey =~ s/[\\\/]$//;
			if (-f "$newkey/$lock") { error("An AWScloudMap update process seems to be already running for this config file. Try later.\nIf this is not true, remove manually lock file '$newkey/$lock'.","","",1); }
		}
		# Set lock where we can
		foreach my $key ($ENV{"TEMP"},$ENV{"TMP"},"/tmp","/",".")
		{
			if (! -d "$key") { next; }
			$DirLock=$key;
			$DirLock =~ s/[\\\/]$//;
			debug("Update lock file $DirLock/$lock is set");
			open(LOCK,">$DirLock/$lock") || error("Failed to create lock file $DirLock/$lock","","",1);
			print LOCK "AWScloudMap update started by process $$ at $nowyear-$nowmonth-$nowday $nowhour:$nowmin:$nowsec\n";
			close(LOCK);
			last;
		}
	} else {
		# Remove lock
		debug("Update lock file $DirLock/$lock is removed");
		unlink("$DirLock/$lock");
	}
	return;
}

#------------------------------------------------------------------------------
# Function:     Signal handler to call Lock_Update to remove lock file
# Parameters:   Signal name
# Input:        None
# Output:       None
# Return:       None
#------------------------------------------------------------------------------
sub SigHandler
{
	my $signame = shift;
	print ucfirst($PROG)." process (ID $$) interrupted by signal $signame.\n";
	&Lock_Update(0);
	exit 1;
}

#------------------------------------------------------------------------------
# Function:     Convert an IPAddress array into a decimal integer
# Parameters:   IPAddressArray
# Input:        None
# Output:       None
# Return:       Int
#------------------------------------------------------------------------------
sub Convert_IP_To_Decimal
{
	my ($IPAddress) = @_;
	my @ip_seg_arr = split(/\./,$IPAddress);
	my $decimal_ip_address = 256 * 256 *256 * $ip_seg_arr[0] + 256 * 256 * $ip_seg_arr[1] + 256 * $ip_seg_arr[2] + $ip_seg_arr[3];
	return($decimal_ip_address);
}

#------------------------------------------------------------------------------
# Function:     Test there is at least one value in list that is not null
# Parameters:   List of values
# Input:        None
# Output:       None
# Return:       1 if there is at least one not null value, otherwise 0
#------------------------------------------------------------------------------
sub AtLeastOneNotNull
{
	debug(" Call to AtLeastOneNotNull (".join('-',@_).")",3);
	foreach my $val (@_) { if ($val) { return 1; } }
	return 0;
}


#------------------------------------------------------------------------------
# Function:     Write url info
# Parameters:   $url
# Input:        %Aliases $MaxLengthOfShownURL $ShowLinksOnUrl $SiteDomain $UseHTTPSLinkForUrl
# Output:       URL link
# Return:       None
#------------------------------------------------------------------------------
sub ShowCloudMapInfo
{
	my $cnt=shift;
	my $url=shift;
	my $maxHits=shift;
	my $htmlTitle=shift;
	my $strDisplay='';
	my $rank=int ((($cnt / $maxHits) * $MAX_RANK) + 0.5 ); # Need to add the 0.5 to achieve rounding rather than just truncation

	if($rank < $MIN_RANK)
	{
	    # Don't want to display low-traffic pages
	    return;
	}
    $displayCount++;
    if($htmlTitle)
    {
        $strDisplay = sanitisePageTitle($htmlTitle, $url);
        debug("strDisplay using page title: \$strDisplay=$strDisplay", 1);
    }else{
        $strDisplay=formatURL($url);
        debug("strDisplay using formatted URL: \$strDisplay=$strDisplay", 1);
    }
    debug("Link Title: \$htmlTitle=$htmlTitle",2);

	if ($ShowLinksOnUrl)
	{
		my $newkey=$url;
		if ($LogType eq 'W' || $LogType eq 'S')
		{
		    # Web or streaming log file
			if ($newkey =~ /^http(s|):/i)
			{
			    # URL seems to be extracted from a proxy log file
				print "\n<a href=\"".XMLEncode("$newkey")."\" class=\"cloudLevel$rank\" title=\"Rank = $rank\" >".XMLEncode($strDisplay)."</a>\&nbsp;";
			} elsif ($newkey =~ /^\//) {
			    # URL seems to be an url extracted from a web or wap server log file
				$newkey =~ s/^\/$SiteDomain//i;
				# Define urlprot
				my $urlprot='http';
				if ($UseHTTPSLinkForUrl && $newkey =~ /^$UseHTTPSLinkForUrl/) { $urlprot='https'; }
				print "\n<a href=\"".XMLEncode("$urlprot://$SiteDomain$newkey")."\" class=\"cloudLevel$rank\" title=\"Rank = $rank\" >".XMLEncode($strDisplay)."</a> \&nbsp;";
			} else {
				print XMLEncode($strDisplay);
			}
		} elsif ($LogType eq 'F') {
		    # Ftp log file
			print XMLEncode($strDisplay);
		} else {
		    # Other type log file
			print XMLEncode($strDisplay);
		}
	} else {
		print XMLEncode($strDisplay);
	}
}

#------------------------------------------------------------------------------
# Function:     Define value for PerlParsingFormat (used for regex log record parsing)
# Parameters:   $LogFormat
# Input:        -
# Output:       $pos_xxx, @pos_extra, @fieldlib, $PerlParsingFormat
# Return:       -
#------------------------------------------------------------------------------
sub DefinePerlParsingFormat
{
    my $LogFormat=shift;
    $pos_vh = $pos_host = $pos_logname = $pos_date = $pos_tz = $pos_method = $pos_url = $pos_code = $pos_size = -1;
    $pos_referer = $pos_agent = $pos_query = $pos_gzipin = $pos_gzipout = $pos_compratio = -1;
    $pos_cluster = $pos_hostr = -1;
    @pos_extra=();
	@fieldlib=();
    $PerlParsingFormat='';
	# Log record examples:
	# Apache combined: 62.161.78.73 user - [dd/mmm/yyyy:hh:mm:ss +0000] "GET / HTTP/1.1" 200 1234 "http://www.from.com/from.htm" "Mozilla/4.0 (compatible; MSIE 5.01; Windows NT 5.0)"
	# Apache combined (408 error): my.domain.com - user [09/Jan/2001:11:38:51 -0600] "OPTIONS /mime-tmp/xxx file.doc HTTP/1.1" 408 - "-" "-"
	# Apache combined (408 error): 62.161.78.73 user - [dd/mmm/yyyy:hh:mm:ss +0000] "-" 408 - "-" "-"
	# Apache common_with_mod_gzip_info1: %h %l %u %t \"%r\" %>s %b mod_gzip: %{mod_gzip_compression_ratio}npct.
	# Apache common_with_mod_gzip_info2: %h %l %u %t \"%r\" %>s %b mod_gzip: %{mod_gzip_result}n In:%{mod_gzip_input_size}n Out:%{mod_gzip_output_size}n:%{mod_gzip_compression_ratio}npct.
	# Apache deflate: %h %l %u %t \"%r\" %>s %b \"%{Referer}i\" \"%{User-Agent}i\" (%{ratio}n)
	# IIS: 2000-07-19 14:14:14 62.161.78.73 - GET / 200 1234 HTTP/1.1 Mozilla/4.0+(compatible;+MSIE+5.01;+Windows+NT+5.0) http://www.from.com/from.htm
	# WebStar: 05/21/00	00:17:31	OK  	200	212.242.30.6	Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt)	http://www.cover.dk/	"www.cover.dk"	:Documentation:graphics:starninelogo.white.gif	1133
	# Squid extended: 12.229.91.170 - - [27/Jun/2002:03:30:50 -0700] "GET http://www.callistocms.com/images/printable.gif HTTP/1.1" 304 354 "-" "Mozilla/5.0 Galeon/1.0.3 (X11; Linux i686; U;) Gecko/0" TCP_REFRESH_HIT:DIRECT
	debug("Call To DefinePerlParsingFormat (LogType='$LogType', LogFormat='$LogFormat')");
	if ($LogFormat =~ /^[1-6]$/)
	{
	    # Pre-defined log format
		if ($LogFormat eq '1' || $LogFormat eq '6')
		{
		    # Same as "%h %l %u %t \"%r\" %>s %b \"%{Referer}i\" \"%{User-Agent}i\"".
			# %u (user) is "([^\\[]+)" instead of "[^ ]+" because can contain space (Lotus Notes). referer and ua might be "".
			$PerlParsingFormat="([^ ]+) [^ ]+ ([^\\[]+) \\[([^ ]+) [^ ]+\\] \\\"([^ ]+) ([^ ]+) [^\\\"]+\\\" ([\\d|-]+) ([\\d|-]+) \\\"(.*?)\\\" \\\"([^\\\"]*)\\\"";
			$pos_host=0;$pos_logname=1;$pos_date=2;$pos_method=3;$pos_url=4;$pos_code=5;$pos_size=6;$pos_referer=7;$pos_agent=8;
			@fieldlib=('host','logname','date','method','url','code','size','referer','ua');
		} elsif ($LogFormat eq '2') {
		    # Same as "date time c-ip cs-username cs-method cs-uri-stem sc-status sc-bytes cs-version cs(User-Agent) cs(Referer)"
			$PerlParsingFormat="(\\S+ \\S+) (\\S+) (\\S+) (\\S+) (\\S+) ([\\d|-]+) ([\\d|-]+) \\S+ (\\S+) (\\S+)";
			$pos_date=0;$pos_host=1;$pos_logname=2;$pos_method=3;$pos_url=4;$pos_code=5;$pos_size=6;$pos_agent=7;$pos_referer=8;
			@fieldlib=('date','host','logname','method','url','code','size','ua','referer');
		} elsif ($LogFormat eq '3') {
			$PerlParsingFormat="([^\\t]*\\t[^\\t]*)\\t([^\\t]*)\\t([\\d|-]*)\\t([^\\t]*)\\t([^\\t]*)\\t([^\\t]*)\\t[^\\t]*\\t([^\\t]*)\\t([\\d]*)";
			$pos_date=0;$pos_method=1;$pos_code=2;$pos_host=3;$pos_agent=4;$pos_referer=5;$pos_url=6;$pos_size=7;
			@fieldlib=('date','method','code','host','ua','referer','url','size');
		} elsif ($LogFormat eq '4') {
		    # Same as "%h %l %u %t \"%r\" %>s %b"
			# %u (user) is "(.+)" instead of "[^ ]+" because can contain space (Lotus Notes).
			$PerlParsingFormat="([^ ]+) [^ ]+ (.+) \\[([^ ]+) [^ ]+\\] \\\"([^ ]+) ([^ ]+) [^\\\"]+\\\" ([\\d|-]+) ([\\d|-]+)";
			$pos_host=0;$pos_logname=1;$pos_date=2;$pos_method=3;$pos_url=4;$pos_code=5;$pos_size=6;
			@fieldlib=('host','logname','date','method','url','code','size');
		}
	} else {
	    # Personalized log format
		my $LogFormatString=$LogFormat;
		# Replacement for Notes format string that are not Apache
		$LogFormatString =~ s/%vh/%virtualname/g;
		# Replacement for Apache format string
		$LogFormatString =~ s/%v(\s)/%virtualname$1/g; $LogFormatString =~ s/%v$/%virtualname/g;
		$LogFormatString =~ s/%h(\s)/%host$1/g; $LogFormatString =~ s/%h$/%host/g;
		$LogFormatString =~ s/%l(\s)/%other$1/g; $LogFormatString =~ s/%l$/%other/g;
		$LogFormatString =~ s/\"%u\"/%lognamequot/g;
		$LogFormatString =~ s/%u(\s)/%logname$1/g; $LogFormatString =~ s/%u$/%logname/g;
		$LogFormatString =~ s/%t(\s)/%time1$1/g; $LogFormatString =~ s/%t$/%time1/g;
		$LogFormatString =~ s/\"%r\"/%methodurl/g;
		$LogFormatString =~ s/%>s/%code/g;
		$LogFormatString =~ s/%b(\s)/%bytesd$1/g;	$LogFormatString =~ s/%b$/%bytesd/g;
		$LogFormatString =~ s/\"%{Referer}i\"/%refererquot/g;
		$LogFormatString =~ s/\"%{User-Agent}i\"/%uaquot/g;
		$LogFormatString =~ s/%{mod_gzip_input_size}n/%gzipin/g;
		$LogFormatString =~ s/%{mod_gzip_output_size}n/%gzipout/g;
		$LogFormatString =~ s/%{mod_gzip_compression_ratio}n/%gzipratio/g;
		$LogFormatString =~ s/\(%{ratio}n\)/%deflateratio/g;
		# Replacement for an IIS and ISA format string
		$LogFormatString =~ s/cs-uri-query/%query/g;	# Must be before cs-uri
		$LogFormatString =~ s/date\stime/%time2/g;
		$LogFormatString =~ s/c-ip/%host/g;
		$LogFormatString =~ s/cs-username/%logname/g;
		$LogFormatString =~ s/cs-method/%method/g;		# GET, POST, RETR STOR
		$LogFormatString =~ s/cs-uri-stem/%url/g; $LogFormatString =~ s/cs-uri/%url/g;
		$LogFormatString =~ s/sc-status/%code/g;
		$LogFormatString =~ s/sc-bytes/%bytesd/g;
		$LogFormatString =~ s/cs-version/%other/g;		# Protocol
		$LogFormatString =~ s/cs\(User-Agent\)/%ua/g; $LogFormatString =~ s/c-agent/%ua/g;
		$LogFormatString =~ s/cs\(Referer\)/%referer/g; $LogFormatString =~ s/cs-referred/%referer/g;
		$LogFormatString =~ s/sc-authenticated/%other/g;
		$LogFormatString =~ s/s-svcname/%other/g;
		$LogFormatString =~ s/s-computername/%other/g;
		$LogFormatString =~ s/r-host/%virtualname/g;
		$LogFormatString =~ s/cs-host/%virtualname/g;
		$LogFormatString =~ s/r-ip/%other/g;
		$LogFormatString =~ s/r-port/%other/g;
		$LogFormatString =~ s/time-taken/%other/g;
		$LogFormatString =~ s/cs-bytes/%other/g;
		$LogFormatString =~ s/cs-protocol/%other/g;
		$LogFormatString =~ s/cs-transport/%other/g;
		$LogFormatString =~ s/s-operation/%method/g;	# GET, POST, RETR STOR
		$LogFormatString =~ s/cs-mime-type/%other/g;
		$LogFormatString =~ s/s-object-source/%other/g;
		$LogFormatString =~ s/s-cache-info/%other/g;
		$LogFormatString =~ s/cluster-node/%cluster/g;
		# Added for MMS
		$LogFormatString =~ s/protocol/%protocolmms/g;	# cs-method might not be available
		$LogFormatString =~ s/c-status/%codemms/g;		# c-status used when sc-status not available
		debug(" LogFormatString=$LogFormatString");
		# $LogFormatString has an AWScloudMap format, so we can generate PerlParsingFormat variable
		my $i = 0;
		my $LogSeparatorWithoutStar = $LogSeparator;
		   $LogSeparatorWithoutStar =~ s/[\*\+]//g;
		foreach my $f (split(/\s+/,$LogFormatString))
		{
			# Add separator for next field
			if ($PerlParsingFormat) { $PerlParsingFormat.="$LogSeparator"; }
			# Special for logname
			if ($f =~ /%lognamequot$/)
			{
				$pos_logname = $i; $i++; push @fieldlib, 'logname';
				$PerlParsingFormat .= "\\\"?([^\\\"]*)\\\"?";
				# logname can be "value", "" and - in same log (Lotus notes)
			}
			# Date format
			elsif ($f =~ /%time1$/ || $f =~ /%time1b$/)
			{
			    # [dd/mmm/yyyy:hh:mm:ss +0000] or [dd/mmm/yyyy:hh:mm:ss],  time1b kept for backward compatibility
				$pos_date = $i;	$i++; push @fieldlib, 'date';
				$pos_tz = $i; $i++; push @fieldlib, 'tz';
				$PerlParsingFormat .= "\\[([^$LogSeparatorWithoutStar]+)( [^$LogSeparatorWithoutStar]+)?\\]";
			}
			elsif ($f =~ /%time2$/) {
			    # yyyy-mm-dd hh:mm:ss
				$pos_date = $i;	$i++; push @fieldlib, 'date';
				$PerlParsingFormat .= "([^$LogSeparatorWithoutStar]+\\s[^$LogSeparatorWithoutStar]+)";	# Need \s for Exchange log files
			}
			elsif ($f =~ /%time3$/) {
			    # mon d hh:mm:ss  or  mon  d hh:mm:ss  or  mon dd hh:mm:ss yyyy  or  day mon dd hh:mm:ss  or  day mon dd hh:mm:ss yyyy
				$pos_date = $i;	$i++; push @fieldlib, 'date';
				$PerlParsingFormat .= "(?:\\w\\w\\w )?(\\w\\w\\w \\s?\\d+ \\d\\d:\\d\\d:\\d\\d(?: \\d\\d\\d\\d)?)";
			}
			elsif ($f =~ /%time4$/) {
			    # ddddddddddddd
				$pos_date = $i;	$i++; push @fieldlib, 'date';
				$PerlParsingFormat .= "(\\d+)";
			}
			# Special for methodurl and methodurlnoprot
			elsif ($f =~ /%methodurl$/) {
				$pos_method = $i; $i++; push @fieldlib, 'method';
				$pos_url = $i; $i++; push @fieldlib, 'url';
				$PerlParsingFormat .= "\\\"([^$LogSeparatorWithoutStar]+) ([^$LogSeparatorWithoutStar]+) [^\\\"]+\\\"";
			}
			elsif ($f =~ /%methodurlnoprot$/) {
				$pos_method = $i; $i++; push @fieldlib, 'method';
				$pos_url = $i; $i++; push @fieldlib, 'url';
				$PerlParsingFormat .= "\\\"([^$LogSeparatorWithoutStar]+) ([^$LogSeparatorWithoutStar]+)\\\"";
			}
			# Common command tags
			elsif ($f =~ /%virtualnamequot$/) {
				$pos_vh = $i; $i++; push @fieldlib, 'vhost';
				$PerlParsingFormat .= "\\\"([^$LogSeparatorWithoutStar]+)\\\"";
			}
			elsif ($f =~ /%virtualname$/) {
				$pos_vh = $i; $i++; push @fieldlib, 'vhost';
				$PerlParsingFormat .= "([^$LogSeparatorWithoutStar]+)";
			}
			elsif ($f =~ /%host_r$/) {
				$pos_hostr = $i; $i++; push @fieldlib, 'hostr';
				$PerlParsingFormat .= "([^$LogSeparatorWithoutStar]+)";
			}
			elsif ($f =~ /%host$/) {
				$pos_host = $i; $i++; push @fieldlib, 'host';
				$PerlParsingFormat .= "([^$LogSeparatorWithoutStar]+)";
			}
			elsif ($f =~ /%logname$/) {
				$pos_logname = $i; $i++; push @fieldlib, 'logname';
				$PerlParsingFormat .= "([^$LogSeparatorWithoutStar]+)";
			}
			elsif ($f =~ /%method$/) {
				$pos_method = $i; $i++; push @fieldlib, 'method';
				$PerlParsingFormat .= "([^$LogSeparatorWithoutStar]+)";
			}
			elsif ($f =~ /%url$/) {
				$pos_url = $i; $i++; push @fieldlib, 'url';
				$PerlParsingFormat .= "([^$LogSeparatorWithoutStar]+)";
			}
			elsif ($f =~ /%query$/) {
				$pos_query = $i; $i++; push @fieldlib, 'query';
				$PerlParsingFormat .= "([^$LogSeparatorWithoutStar]+)";
			}
			elsif ($f =~ /%code$/) {
				$pos_code = $i; $i++; push @fieldlib, 'code';
				$PerlParsingFormat .= "([^$LogSeparatorWithoutStar]+)";
			}
			elsif ($f =~ /%bytesd$/) {
				$pos_size = $i; $i++; push @fieldlib, 'size';
				$PerlParsingFormat .= "([^$LogSeparatorWithoutStar]+)";
			}
			elsif ($f =~ /%refererquot$/) {
				$pos_referer = $i; $i++; push @fieldlib, 'referer';
				$PerlParsingFormat .= "\\\"([^\\\"]*)\\\""; 		# referer might be ""
			}
			elsif ($f =~ /%referer$/) {
				$pos_referer = $i; $i++; push @fieldlib, 'referer';
				$PerlParsingFormat .= "([^$LogSeparatorWithoutStar]+)";
			}
			elsif ($f =~ /%uaquot$/) {
				$pos_agent = $i; $i++; push @fieldlib, 'ua';
				$PerlParsingFormat .= "\\\"([^\\\"]*)\\\"";			# ua might be ""
			}
			elsif ($f =~ /%uabracket$/) {
				$pos_agent = $i; $i++; push @fieldlib, 'ua';
				$PerlParsingFormat .= "\\\[([^\\\]]*)\\\]"; 		# ua might be []
			}
			elsif ($f =~ /%ua$/) {
				$pos_agent = $i; $i++; push @fieldlib, 'ua';
				$PerlParsingFormat .= "([^$LogSeparatorWithoutStar]+)";
			}
			elsif ($f =~ /%gzipin$/ ) {
				$pos_gzipin=$i;$i++; push @fieldlib, 'gzipin';
				$PerlParsingFormat .= "([^$LogSeparatorWithoutStar]+)";
			}
			elsif ($f =~ /%gzipout/ ) {		# Compare $f to /%gzipout/ and not to /%gzipout$/ like other fields
				$pos_gzipout=$i;$i++; push @fieldlib, 'gzipout';
				$PerlParsingFormat .= "([^$LogSeparatorWithoutStar]+)";
			}
			elsif ($f =~ /%gzipratio/ ) {	# Compare $f to /%gzipratio/ and not to /%gzipratio$/ like other fields
				$pos_compratio=$i;$i++; push @fieldlib, 'gzipratio';
				$PerlParsingFormat .= "([^$LogSeparatorWithoutStar]+)";
			}
			elsif ($f =~ /%deflateratio/ ) {	# Compare $f to /%deflateratio/ and not to /%deflateratio$/ like other fields
				$pos_compratio=$i;$i++; push @fieldlib, 'deflateratio';
				$PerlParsingFormat .= "([^$LogSeparatorWithoutStar]+)";
			}
			elsif ($f =~ /%email_r$/) {
				$i; $i++; push @fieldlib, 'email_r';
				$PerlParsingFormat .= "([^$LogSeparatorWithoutStar]+)";
			}
			elsif ($f =~ /%email$/) {
				$i; $i++; push @fieldlib, 'email';
				$PerlParsingFormat .= "([^$LogSeparatorWithoutStar]+)";
			}
			elsif ($f =~ /%cluster$/) {
				$pos_cluster = $i; $i++; push @fieldlib, 'clusternb';
				$PerlParsingFormat .= "([^$LogSeparatorWithoutStar]+)";
			}
			elsif ($f =~ /%timetaken$/) {
				$pos_timetaken = $i; $i++; push @fieldlib, 'timetaken';
				$PerlParsingFormat .= "([^$LogSeparatorWithoutStar]+)";
			}
			# Special for protocolmms, used for method if method not already found (for MMS)
			elsif ($f =~ /%protocolmms$/) {
				if ($pos_method < 0)
				{
					$pos_method = $i; $i++; push @fieldlib, 'method';
					$PerlParsingFormat .= "([^$LogSeparatorWithoutStar]+)";
				}
			}
			# Special for codemms, used for code only if code not already found (for MMS)
			elsif ($f =~ /%codemms$/) {
				if ($pos_code < 0)
				{
					$pos_code = $i; $i++; push @fieldlib, 'code';
					$PerlParsingFormat .= "([^$LogSeparatorWithoutStar]+)";
				}
			}
			# Extra tag
			elsif ($f =~ /%extra(\d+)$/) {
				$pos_extra[$1] = $i; $i++; push @fieldlib, "extra$1";
				$PerlParsingFormat .= "([^$LogSeparatorWithoutStar]+)";
			}
			# Other tag
			elsif ($f =~ /%other$/) {
				$PerlParsingFormat .= "[^$LogSeparatorWithoutStar]+";
			}
			elsif ($f =~ /%otherquot$/) {
				$PerlParsingFormat .= "\\\"[^\\\"]*\\\"";
			}
			# Unknown tag (no parenthesis)
			else {
				$PerlParsingFormat .= "[^$LogSeparatorWithoutStar]+";
			}
		}
		if (! $PerlParsingFormat) { error("No recognized format tag in personalized LogFormat string"); }
	}
	if ($pos_host < 0) { error("Your personalized LogFormat does not include all fields required by AWScloudMap (Add \%host in your LogFormat string)."); }
	if ($pos_date < 0) { error("Your personalized LogFormat does not include all fields required by AWScloudMap (Add \%time1 or \%time2 in your LogFormat string)."); }
	if ($pos_method < 0) { error("Your personalized LogFormat does not include all fields required by AWScloudMap (Add \%methodurl or \%method in your LogFormat string)."); }
	if ($pos_url < 0) { error("Your personalized LogFormat does not include all fields required by AWScloudMap (Add \%methodurl or \%url in your LogFormat string)."); }
	if ($pos_code < 0) { error("Your personalized LogFormat does not include all fields required by AWScloudMap (Add \%code in your LogFormat string)."); }
	if ($pos_size < 0) { error("Your personalized LogFormat does not include all fields required by AWScloudMap (Add \%bytesd in your LogFormat string)."); }
	$PerlParsingFormat=qr/^$PerlParsingFormat/;
	debug(" PerlParsingFormat is $PerlParsingFormat");
}


#------------------------------------------------------------------------------
# MAIN
#------------------------------------------------------------------------------
($DIR=$0) =~ s/([^\/\\]+)$//; ($PROG=$1) =~ s/\.([^\.]*)$//; $Extension=$1;
$DIR||='.'; $DIR =~ s/([^\/\\])[\\\/]+$/$1/;

$starttime=time();

# Get current time (time when AWScloudMap was started)
($nowsec,$nowmin,$nowhour,$nowday,$nowmonth,$nowyear,$nowwday,$nowyday) = localtime($starttime);
$nowweekofmonth=int($nowday/7);
$nowweekofyear=int(($nowyday-1+6-($nowwday==0?6:$nowwday-1))/7)+1; if ($nowweekofyear > 52) { $nowweekofyear = 1; }
$nowdaymod=$nowday%7;
$nowwday++;
$nowns=Time::Local::timegm(0,0,0,$nowday,$nowmonth,$nowyear);
if ($nowdaymod <= $nowwday) { if (($nowwday != 7) || ($nowdaymod != 0)) { $nowweekofmonth=$nowweekofmonth+1; } }
if ($nowdaymod >  $nowwday) { $nowweekofmonth=$nowweekofmonth+2; }
# Change format of time variables
$nowweekofmonth="0$nowweekofmonth";
if ($nowweekofyear < 10) { $nowweekofyear = "0$nowweekofyear"; }
if ($nowyear < 100) { $nowyear+=2000; } else { $nowyear+=1900; }
$nowsmallyear=$nowyear;$nowsmallyear =~ s/^..//;
if (++$nowmonth < 10) { $nowmonth = "0$nowmonth"; }
if ($nowday < 10) { $nowday = "0$nowday"; }
if ($nowhour < 10) { $nowhour = "0$nowhour"; }
if ($nowmin < 10) { $nowmin = "0$nowmin"; }
if ($nowsec < 10) { $nowsec = "0$nowsec"; }
$nowtime=int($nowyear.$nowmonth.$nowday.$nowhour.$nowmin.$nowsec);
# Get tomorrow time (will be used to discard some record with corrupted date (future date))
my ($tomorrowsec,$tomorrowmin,$tomorrowhour,$tomorrowday,$tomorrowmonth,$tomorrowyear) = localtime($starttime+86400);
if ($tomorrowyear < 100) { $tomorrowyear+=2000; } else { $tomorrowyear+=1900; }
if (++$tomorrowmonth < 10) { $tomorrowmonth = "0$tomorrowmonth"; }
if ($tomorrowday < 10) { $tomorrowday = "0$tomorrowday"; }
if ($tomorrowhour < 10) { $tomorrowhour = "0$tomorrowhour"; }
if ($tomorrowmin < 10) { $tomorrowmin = "0$tomorrowmin"; }
if ($tomorrowsec < 10) { $tomorrowsec = "0$tomorrowsec"; }
$tomorrowtime=int($tomorrowyear.$tomorrowmonth.$tomorrowday.$tomorrowhour.$tomorrowmin.$tomorrowsec);

# Allowed options (ToDo:Not yet used/checked)
my @AllowedArgs=('config',
'logtype','logfile','output','update',
'pluginmode',
'urlcloudmap',
'lang','month','year','debug',
);

# Parse input parameters
$QueryString='';
# AWScloudMap use GATEWAY_INTERFACE to known if ran as CLI or CGI.
# AWScloudMap_DEL_GATEWAY_INTERFACE can be set to force AWScloudMap to be ran as CLI even from a web page.
	# Prepare QueryString
	for (0..@ARGV-1)
	{
		# TODO Check if ARGV is in @AllowedArgs
		if ($QueryString) { $QueryString .= '&amp;'; }
		my $NewLinkParams=$ARGV[$_]; $NewLinkParams =~ s/^-+//;
		$QueryString .= "$NewLinkParams";
	}

	# Update with no report by default when run from command line
	$UpdateStats=1;

	if ($QueryString =~ /config=([^&]+)/i)				{ $SiteConfig=&Sanitize("$1"); }
	if ($QueryString =~ /configdir=([^&]+)/i)			{ $DirConfig=&Sanitize("$1"); }
	# All filters
	if ($QueryString =~ /output=urlcloudmap:([^&]+)/i)	{ $FilterIn{'url'}="$1"; }			# Filter on URL list can be defined with output=urldetail:filter to reduce number of lines read and showed
    # Config parameters
	if ($QueryString =~ /logfile=([^&]+)/i)				{ $LogFile="$1"; }
    $LogFile =~ s/"$//;  # If the logfile parameter has been wrapped in double-quotes, the closing set gets included, so need to be removed here.
                         # No idea why the initial set don't need to be dealt with, probably a side effect of sub Substitute_Tags.
                         # We also lose any trailing slash along the way, though that does not seem to be a problem.
	# If show options
	if ($QueryString =~ /showsteps/i) 					{ $ShowSteps=1; $QueryString=~s/showsteps[^&]*//i; }
	if ($QueryString =~ /showcorrupted/i) 				{ $ShowCorrupted=1; $QueryString=~s/showcorrupted[^&]*//i; }
	if ($QueryString =~ /showdropped/i)					{ $ShowDropped=1; $QueryString=~s/showdropped[^&]*//i; }
	if ($QueryString =~ /showunknownorigin/i)			{ $ShowUnknownOrigin=1; $QueryString=~s/showunknownorigin[^&]*//i; }

if ($QueryString =~ /(^|&|&amp;)debug=(\d+)/i)			{ $Debug=$2; }
if ($QueryString =~ /(^|&|&amp;)databasebreak=(\w+)/i)	{ $DatabaseBreak=$2; }
if ($QueryString =~ /(^|&|&amp;)updatefor=(\d+)/i)		{ $UpdateFor=$2; }
if ($QueryString =~ /(^|&|&amp;)LIMIT_FLUSH=(\d+)/i)	{ $LIMIT_FLUSH=$2; }
if ($QueryString =~ /(^|&|&amp;)pluginmode/i) 			{ $PluginMode=1; } # 0 = Full File output, 1 = Single <div> only, no CSS or <head>
# Get/Define output
if ($QueryString =~ /(^|&|&amp;)output(=[^&]*|)(.*)(&|&amp;)output(=[^&]*|)(&|$)/i) { error("Only 1 output option is allowed","","",1); }
if ($QueryString =~ /(^|&|&amp;)output(=[^&]*|)(&|$)/i)
{
	# At least one output expected. We define %HTMLOutput
	my $outputlist="$2";
	if ($outputlist)
	{
		$outputlist =~ s/^=//;
		foreach my $outputparam (split(/,/,$outputlist))
		{
			$outputparam=~s/:(.*)$//;
			if ($outputparam) { $HTMLOutput{lc($outputparam)}="$1"||1; }
		}
	}
	# If no update
	if ($QueryString !~ /update/i) { $UpdateStats=0; }
	# If no output defined, used default value
	if (! scalar keys %HTMLOutput) { $HTMLOutput{'main'}=1; }
}

# Remove -output option with no = from QueryString
$QueryString=~s/(^|&|&amp;)output(&|$)/$1/i; $QueryString=~s/&+$//;

# Check year, month, day, hour parameters
if ($QueryString =~ /(^|&|&amp;)month=(year)/i) { error("month=year is a deprecated option. Use month=all instead."); }
if ($QueryString =~ /(^|&|&amp;)year=(\d\d\d\d)/i)
{
    $YearRequired=sprintf("%04d",$2);
} else {
    $YearRequired="$nowyear";
}
if ($QueryString =~ /(^|&|&amp;)month=(\d{1,2})/i) { $MonthRequired=sprintf("%02d",$2); }
elsif ($QueryString =~ /(^|&|&amp;)month=(all)/i)  { $MonthRequired='all'; }
else { $MonthRequired="$nowmonth"; }
if ($QueryString =~ /(^|&|&amp;)day=(\d{1,2})/i)   { $DayRequired=sprintf("%02d",$2); }	# day is a hidden option. Must not be used (Make results not understandable). Available for users that rename history files with day.
else { $DayRequired=''; }
if ($QueryString =~ /(^|&|&amp;)hour=(\d{1,2})/i)  { $HourRequired=sprintf("%02d",$2); }	# hour is a hidden option. Must not be used (Make results not understandable). Available for users that rename history files with day.
else { $HourRequired=''; }

# ToDo:
# Check parameter validity

# Print AWScloudMap and Perl version
if ($Debug)
{
	debug(ucfirst($PROG)." - $VERSION - Perl $^X $]",1);
	debug("DIR=$DIR PROG=$PROG Extension=$Extension",2);
	debug("QUERY_STRING=$QueryString",1);
	debug("HTMLOutput=".join(',',keys %HTMLOutput),1);
	debug("YearRequired=$YearRequired, MonthRequired=$MonthRequired",2);
	debug("DayRequired=$DayRequired, HourRequired=$HourRequired",2);
	debug("UpdateFor=$UpdateFor",2);
	debug("PluginMode=$PluginMode",2);
	debug("DirConfig=$DirConfig",2);
}

# Force SiteConfig if AWScloudMap_FORCE_CONFIG is defined
if ($ENV{'AWScloudMap_CONFIG'}) { $ENV{'AWScloudMap_FORCE_CONFIG'}=$ENV{'AWScloudMap_CONFIG'}; } # For backward compatibility
if ($ENV{'AWScloudMap_FORCE_CONFIG'})
{
	debug("AWScloudMap_FORCE_CONFIG parameter is defined to '".$ENV{'AWScloudMap_FORCE_CONFIG'}."'. $PROG will use this as config value.");
	$SiteConfig=&Sanitize($ENV{'AWScloudMap_FORCE_CONFIG'});
}

if(! $SiteConfig)
{
	print "\nAWScloudMap is a special-purpose logfile analyzer, that generates 'Cloud Map' diagrams.\n";
	print "AWScloudMap comes with ABSOLUTELY NO WARRANTY. \n";
	print "It is free software distributed under the GNU General Public License\n";
	print "This program is based on AWStats.pl, which is seperately available,\n";
	print "also under the GNU General Public License (GPL).\n";
	print "\n";
	print "See exampleUsage.txt for command-line examples.\n";
	print "\n";
	print "  This runs $PROG in command line to update statistics of a web site, from\n";
	print "   the log file defined in AWScloudMap config file (with -update option), or build\n";
	print "   an HTML report (with the -output option).\n";
	print "  First, $PROG tries to read $PROG.virtualhostname.conf as the config file.\n";
	print "  If not found, $PROG tries to read $PROG.conf\n";
	print "  Note 1: Config files ($PROG.virtualhostname.conf or $PROG.conf) must be\n";
	print "   in /etc/AWScloudMap, /usr/local/etc/AWScloudMap, /etc or the same folder\n";
	print "   as the AWScloudMap.pl script itself.\n";
	print "  Note 2: If AWScloudMap_FORCE_CONFIG environment variable is defined, AWScloudMap will\n";
	print "   use it as the \"config\" value, whatever is the value on command line or URL.\n";
	print "   See AWScloudMap documentation for all setup instrutions.\n";
	print "\n";
	print "Options to update statistics:\n";
	print "  -update        to update statistics (default)\n";
	print "  -showsteps     to add benchmark information every $NB_OF_LINES_FOR_BENCHMARK lines processed\n";
	print "  -showcorrupted to add output for each corrupted line found, with reason\n";
	print "  -showdropped   to add output for each dropped line found, with reason\n";
	print "  -updatefor=n   to stop the update process after parsing 'n' lines\n";
	print "  Be careful to process log files in chronological order when updating statistics.\n";
	print "\n";
	print "Options to show statistics:\n";
	print "  -output=urlcloudMap \n";
	print "               to output main HTML report (no update made except with -update)\n";
	print "  -lang=LL     to output an HTML report in language LL (en,de,es,fr,it,nl,...)\n";
	print "  -month=MM    to output an HTML report for the month MM\n";
	print "  -year=YYYY   to output an HTML report for the year YYYY\n";
	print "\n";
	print "Other options:\n";
	print "  -debug=X     to add debug information less than level X (speed of operation will be reduced)\n";
	print "\n";
	print "Currently supports:\n";
	print "  Web/Ftp/streaming server logs\n";
	print "  Load balanced log files\n";
	print "  Report by day/month/year\n";
	print "More information can be found at http://cloudMap.calcResult.com\n";
	exit 2;
}
$SiteConfig||=&Sanitize($ENV{'SERVER_NAME'});
$ENV{'AWScloudMap_CURRENT_CONFIG'}=$SiteConfig;

# Read config file (SiteConfig must be defined)
&Read_Config($DirConfig);

# Check language
if ($QueryString =~ /(^|&|&amp;)lang=([^&]+)/i)	{ $Lang="$2"; }
if (! $Lang || $Lang eq 'auto')
{
    # If lang not defined or forced to auto
	my $langlist=$ENV{'HTTP_ACCEPT_LANGUAGE'}||''; $langlist =~ s/;[^,]*//g;
	debug("Search an available language among HTTP_ACCEPT_LANGUAGE=$langlist",1);
	foreach my $code (split(/,/,$langlist))
	{
	    # Search for a valid lang in priority
		if ($LanguageCodeMap{$code})
		{
		    $Lang=$LanguageCodeMap{$code};
		    debug(" Will try to use Lang=$Lang",1);
		    last;
		}
		$code =~ s/-.*$//;
		if ($LanguageCodeMap{$code})
		{
		    $Lang=$LanguageCodeMap{$code};
		    debug(" Will try to use Lang=$Lang",1);
		    last;
	    }
	}
}
if (! $Lang || $Lang eq 'auto')
{
	debug(" No language defined or available. Will use Lang=en",1);
	$Lang='en';
}

# Check and correct bad parameters
&Check_Config();

# Load Message files, Reference data files and Plugins
	&Read_Language_Data($Lang);
		my %datatoload=();
		my ($filedomains,$filemime,$filerobots,$fileworms,$filebrowser,$fileos,$filese)=('domains','mime','robots','worms','browsers','operating_systems','search_engines');
        my ($filestatushttp)=('status_http');
		if ($LevelForBrowsersDetection eq 'allphones') { $filebrowser='browsers_phone'; }


# Init other parameters
$NB_OF_LINES_FOR_BENCHMARK--;
if (! $DirData || $DirData =~ /^\./)
{
    if (! $DirData || $DirData eq '.')
    {
        $DirData="$DIR";
    } elsif ($DIR && $DIR ne '.') {
        # If not defined or chosen to '.' value then DirData is current dir
        $DirData="$DIR/$DirData";
    }
}
$DirData||='.';		# If current dir not defined then we put it to '.'
$DirData =~ s/[\\\/]+$//;

if ($FirstDayOfWeek == 1) { @DOWIndex = (1,2,3,4,5,6,0); }
else { @DOWIndex = (0,1,2,3,4,5,6); }

# Print html header
if (! $HeaderHTMLSent && !$PluginMode) { &html_head; }

%MonthNumLib = ("01","$Message[5]","02","$Message[6]","03","$Message[7]","04","$Message[8]","05","$Message[9]","06","$Message[10]","07","$Message[11]","08","$Message[12]","09","$Message[13]","10","$Message[14]","11","$Message[15]","12","$Message[16]");

# Build ListOfYears list with all existing years
($lastyearbeforeupdate,$lastmonthbeforeupdate,$lastdaybeforeupdate,$lasthourbeforeupdate,$lastdatebeforeupdate)=(0,0,0,0,0);
my $datemask='';
if    ($DatabaseBreak eq 'month') { $datemask='(\d\d)(\d\d\d\d)'; }
elsif ($DatabaseBreak eq 'year')  { $datemask='(\d\d\d\d)'; }
elsif ($DatabaseBreak eq 'day')   { $datemask='(\d\d)(\d\d\d\d)(\d\d)'; }
elsif ($DatabaseBreak eq 'hour')  { $datemask='(\d\d)(\d\d\d\d)(\d\d)(\d\d)'; }
debug("Scan for last history files into DirData='$DirData' with mask='$datemask'");
opendir(DIR,"$DirData");
my $regfilesuffix=quotemeta($FileSuffix);
foreach (grep /^$PROG$datemask$regfilesuffix\.txt(|\.gz)$/i, file_filt sort readdir DIR)
{
	/^$PROG$datemask$regfilesuffix\.txt(|\.gz)$/i;
	if (! $ListOfYears{"$2"} || "$1" gt $ListOfYears{"$2"})
	{
		# ListOfYears contains max month found
		$ListOfYears{"$2"}="$1";
	}
	my $rangestring=($2||"").($1||"").($3||"").($4||"");
	if ($rangestring gt $lastdatebeforeupdate)
	{
		# We are on a new max for mask
		$lastyearbeforeupdate=($2||"");
		$lastmonthbeforeupdate=($1||"");
		$lastdaybeforeupdate=($3||"");
		$lasthourbeforeupdate=($4||"");
		$lastdatebeforeupdate=$rangestring;
	}
}
close DIR;

# If at least one file found, get value for LastLine
if ($lastyearbeforeupdate)
{
	# Read 'general' section of last history file for LastLine
	&Read_History_With_TmpUpdate($lastyearbeforeupdate,$lastmonthbeforeupdate,$lastdaybeforeupdate,$lasthourbeforeupdate,0,0,"general");
}
# Warning if lastline in future
if ($LastLine > ($nowtime + 20000))
{
	warning("WARNING: LastLine parameter in history file is '$LastLine' so in future. May be you need to manually correct the line LastLine in some AWScloudMap*.$SiteConfig.conf files.");
}
# Force LastLine
if ($QueryString =~ /lastline=(\d{14})/i)
{
	$LastLine=$1;
}
if ($Debug)
{
	debug("Last year=$lastyearbeforeupdate - Last month=$lastmonthbeforeupdate");
	debug("Last day=$lastdaybeforeupdate - Last hour=$lasthourbeforeupdate");
	debug("LastLine=$LastLine");
	debug("LastLineNumber=$LastLineNumber");
	debug("LastLineOffset=$LastLineOffset");
	debug("LastLineChecksum=$LastLineChecksum");
}

# Init vars
&Init_HashArray();


#------------------------------------------
# UPDATE PROCESS
#------------------------------------------
my $lastLineNum=0;
my $lastlineoffset=0;
my $lastlineoffsetnext=0;

debug("\$UpdateStats = $UpdateStats",2);
if ($UpdateStats)
{
	if (! scalar keys %HTMLOutput)
	{
		print "\n\nAWScloudMap version $VERSION \nis about to Create/Update summary data files using config-file: \"$FileConfig\", \nusing data in: $LogFile ";
	}

	$lastprocessedyear=$lastyearbeforeupdate||0;
	$lastprocessedmonth=$lastmonthbeforeupdate||0;
	$lastprocessedday=$lastdaybeforeupdate||0;
	$lastprocessedhour=$lasthourbeforeupdate||0;
	$lastprocesseddate='';
	if    ($DatabaseBreak eq 'month') { $lastprocesseddate=sprintf("%04i%02i",$lastprocessedyear,$lastprocessedmonth); }
	elsif ($DatabaseBreak eq 'year')  { $lastprocesseddate=sprintf("%04i%",$lastprocessedyear); }
	elsif ($DatabaseBreak eq 'day')   { $lastprocesseddate=sprintf("%04i%02i%02i",$lastprocessedyear,$lastprocessedmonth,$lastprocessedday); }
	elsif ($DatabaseBreak eq 'hour')  { $lastprocesseddate=sprintf("%04i%02i%02i%02i",$lastprocessedyear,$lastprocessedmonth,$lastprocessedday,$lastprocessedhour); }


	my @list;
	# Init RobotsSearchIDOrder required for update process
	@list=();

## MJB Check this section:
	if ($LevelForRobotsDetection >= 1)
	{
		foreach (1..$LevelForRobotsDetection) { push @list,"list$_"; }
		push @list,"listgen";		# Always added
	}
	foreach my $key (@list)
	{
		push @RobotsSearchIDOrder,@{"RobotsSearchIDOrder_$key"};
		debug("Add ".@{"RobotsSearchIDOrder_$key"}." elements from RobotsSearchIDOrder_$key into RobotsSearchIDOrder",2);
	}
	debug("RobotsSearchIDOrder has now ".@RobotsSearchIDOrder." elements",1);
	# Init SearchEnginesIDOrder required for update process
	@list=();
	if ($LevelForSearchEnginesDetection >= 1)
	{
		foreach (1..$LevelForSearchEnginesDetection) { push @list,"list$_"; }
		push @list,"listgen";		# Always added
	}
	foreach my $key (@list)
	{
		push @SearchEnginesSearchIDOrder,@{"SearchEnginesSearchIDOrder_$key"};
		debug("Add ".@{"SearchEnginesSearchIDOrder_$key"}." elements from SearchEnginesSearchIDOrder_$key into SearchEnginesSearchIDOrder",2);
	}
	debug("SearchEnginesSearchIDOrder has now ".@SearchEnginesSearchIDOrder." elements",1);

	# Complete HostAliases array
	my $sitetoanalyze=quotemeta(lc($SiteDomain));
	if (! @HostAliases)
	{
		warning("Warning: HostAliases parameter is not defined, $PROG choose \"$SiteDomain localhost 127.0.0.1\".");
		push @HostAliases,qr/^$sitetoanalyze$/i; push @HostAliases,qr/^localhost$/i; push @HostAliases,qr/^127\.0\.0\.1$/i;
	} else {
	    # Add SiteDomain as first value:
	    unshift @HostAliases,qr/^$sitetoanalyze$/i;
	}

	# Optimize arrays
	@HostAliases=&OptimizeArray(\@HostAliases,1); debug("HostAliases precompiled regex list is now @HostAliases",1);
	@SkipHosts=&OptimizeArray(\@SkipHosts,1); debug("SkipHosts precompiled regex list is now @SkipHosts",1);
	@SkipUserAgents=&OptimizeArray(\@SkipUserAgents,1); debug("SkipUserAgents precompiled regex list is now @SkipUserAgents",1);
	@SkipFiles=&OptimizeArray(\@SkipFiles,$URLNotCaseSensitive); debug("SkipFiles precompiled regex list is now @SkipFiles",1);
	@OnlyHosts=&OptimizeArray(\@OnlyHosts,1); debug("OnlyHosts precompiled regex list is now @OnlyHosts",1);
	@OnlyUserAgents=&OptimizeArray(\@OnlyUserAgents,1); debug("OnlyUserAgents precompiled regex list is now @OnlyUserAgents",1);
	@OnlyFiles=&OptimizeArray(\@OnlyFiles,$URLNotCaseSensitive); debug("OnlyFiles precompiled regex list is now @OnlyFiles",1);
	@NotPageFiles=&OptimizeArray(\@NotPageFiles,$URLNotCaseSensitive); debug("NotPageFiles precompiled regex list is now @NotPageFiles",1);
	# Precompile the regex search strings with qr
	@RobotsSearchIDOrder=map{qr/$_/i} @RobotsSearchIDOrder;
	@SearchEnginesSearchIDOrder=map{qr/$_/i} @SearchEnginesSearchIDOrder;
	my $miscquoted="";
	my $defquoted=quotemeta("/$DefaultFile[0]");
	my $sitewithoutwww=lc($SiteDomain); $sitewithoutwww =~ s/www\.//; $sitewithoutwww=quotemeta($sitewithoutwww);
	# Define precompiled regex
	my $regmisc=qr/^$miscquoted/;
	my $regfavico=qr/\/favicon\.ico$/i;
	my $regrobot=qr/^\/robots\.txt$/i;
	$regtruncanchor=qr/#(\w*)$/;
	$regtruncurl=qr/([$URLQuerySeparators])(.*)$/;
	$regext=qr/\.(\w{1,6})$/;
	$regdefault = "";
	if ($URLNotCaseSensitive) { $regdefault=qr/$defquoted$/i; }
	else { $regdefault=qr/$defquoted$/; }
	my $regipv4=qr/^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$/;
	my $regipv6=qr/^[0-9A-F]*:/i;
	my $regvermsie=qr/msie([+_ ]|)([\d\.]*)/i;
	my $regvernetscape=qr/netscape.?\/([\d\.]*)/i;
	my $regverfirefox=qr/firefox\/([\d\.]*)/i;
	my $regversvn=qr/svn\/([\d\.]*)/i;
	my $regvermozilla=qr/mozilla(\/|)([\d\.]*)/i;
	my $regnotie=qr/webtv|omniweb|opera/i;
	my $regnotnetscape=qr/gecko|compatible|opera|galeon|safari/i;
	my $regreferer=qr/^(\w+):\/\/([^\/:]+)(:\d+|)/;
	my $regreferernoquery=qr/^([^$URLQuerySeparators]+)/;
	my $reglocal=qr/^(www\.|)$sitewithoutwww/i;
	$regget=qr/get|out/i;
	$regsent=qr/sent|put|in/i;

	# Define value of $pos_xxx, @fieldlib, $PerlParsingFormat
	&DefinePerlParsingFormat($LogFormat);

	# Processing log
	#------------------------------------------

	if ($EnableLockForUpdate)
	{
		# Trap signals to remove lock
		$SIG{INT} = \&SigHandler;	# 2
		#$SIG{KILL} = \&SigHandler;	# 9
		#$SIG{TERM} = \&SigHandler;	# 15
		# Set AWScloudMap update lock
		&Lock_Update(1);
	}

	debug("Start Update process (lastprocesseddate=$lastprocesseddate)");

    #MJB processLogs($regget, $regsent, $lastprocesseddate, $lastprocessedyear, $lastprocessedmonth, $lastprocessedday, $lastprocessedhour);
    processLogs();

	# Process the Rename - Archive - Purge phase
	my $renameok=1; my $archiveok=1;

	# Open Log file for writing if PurgeLogFile is on
	if ($PurgeLogFile)
	{
		if ($ArchiveLogRecords)
		{
			if ($ArchiveLogRecords == 1)
			{
			    # For backward compatibility
			    $ArchiveFileName="$DirData/${PROG}_archive$FileSuffix.log";
            } else {
                $ArchiveFileName="$DirData/${PROG}_archive$FileSuffix.".&Substitute_Tags($ArchiveLogRecords).".log";
            }
			open(LOG,"+<$LogFile") || error("Enable to archive log records of \"$LogFile\" into \"$ArchiveFileName\" because source can't be opened for read and write: $!<br>\n");
		} else {
			open(LOG,"+<$LogFile");
		}
		binmode LOG;
	}

	# Rename all HISTORYTMP files into HISTORYTXT
	&Rename_All_Tmp_History();

	# Purge Log file if option is on and all renaming are ok
	if ($PurgeLogFile)
	{
		# Archive LOG file into ARCHIVELOG
		if ($ArchiveLogRecords)
		{
			debug("Start of archiving log file");
			open(ARCHIVELOG,">>$ArchiveFileName") || error("Couldn't open file \"$ArchiveFileName\" to archive log: $!");
			binmode ARCHIVELOG;
			while (<LOG>) {
				if (! print ARCHIVELOG $_) { $archiveok=0; last; }
			}
			close(ARCHIVELOG) || error("Archiving failed during closing archive: $!");
			if ($SaveDatabaseFilesWithPermissionsForEveryone) {	chmod 0666,"$ArchiveFileName"; }
			debug("End of archiving log file");
		}
		# If rename and archive ok
		if ($renameok && $archiveok)
		{
			debug("Purge log file");
			truncate(LOG,0) || warning("Warning: $PROG couldn't purge logfile \"$LogFile\"\nChange your logfile permissions to allow write for your web server CGI process or change PurgeLogFile=1 into PurgeLogFile=0 in configure file and think to purge sometimes manually your logfile (just after running an update process to not loose any not already processed records your log file contains).");
		}
		close(LOG);
	}

	if ($EnableLockForUpdate) {
		# Remove lock
		&Lock_Update(0);
		# Restore signals handler
		$SIG{INT} = 'DEFAULT';	# 2
		#$SIG{KILL} = 'DEFAULT';	# 9
		#$SIG{TERM} = 'DEFAULT';	# 15
	}

}
# End of log processing if ($UPdateStats)


#---------------------------------------------------------------------
# SHOW REPORT
#---------------------------------------------------------------------

if (scalar keys %HTMLOutput)
{
	debug("\$HTMLOutput{'urlcloudmap'} =$HTMLOutput{'urlcloudmap'} ",1);
	debug("\$HTMLOutput =".(scalar keys %HTMLOutput),1);

	debug("YearRequired=$YearRequired, MonthRequired=$MonthRequired",2);
	debug("DayRequired=$DayRequired, HourRequired=$HourRequired",2);

	my $max_p; my $max_h; my $max_v;
	my $total_u; my $total_v; my $total_h; my $total_s; my $total_l; my $total_r;
	my $average_u; my $average_v; my $average_p; my $average_h; my $average_k; my $average_s;
	my $rest_p; my $rest_h; my $rest_k; my $rest_e; my $rest_x; my $rest_s; my $rest_l; my $rest_r;
	my $average_nb;

	# Define the NewLinkParams for main chart
	my $NewLinkParams=${QueryString};
	$NewLinkParams =~ s/(^|&|&amp;)update(=\w*|$)//i;
	$NewLinkParams =~ s/(^|&|&amp;)output(=\w*|$)//i;
	$NewLinkParams =~ s/(^|&|&amp;)staticlinks(=\w*|$)//i;
	my $NewLinkTarget='';
	$NewLinkParams =~ s/(&amp;|&)+/&amp;/i;
	$NewLinkParams =~ s/^&amp;//; $NewLinkParams =~ s/&amp;$//;
	if ($NewLinkParams) { $NewLinkParams="${NewLinkParams}&amp;"; }


	# READING DATA
	#-------------
	&Init_HashArray();

	if ($DatabaseBreak eq 'month')
	{
		for (my $ix=12; $ix>=1; $ix--)
		{
    		my $stringforload='';
			my $monthix=sprintf("%02s",$ix);
			if ($MonthRequired eq 'all' || $monthix eq $MonthRequired)
			{
				$stringforload='all';			# Read full history file
			} elsif (($HTMLOutput{'main'} && $ShowMonthStats) || $HTMLOutput{'alldays'}) {
				$stringforload='general time';	# Read general and time sections.
			}
			if ($stringforload)
			{
				&Read_History_With_TmpUpdate($YearRequired,$monthix,'','',0,0,$stringforload);
			}
		}
    }
	if ($DatabaseBreak eq 'day')
	{
		my $stringforload='all';
		my $monthix=sprintf("%02s",$MonthRequired);
		my $dayix=sprintf("%02s",$DayRequired);
		&Read_History_With_TmpUpdate($YearRequired,$monthix,$dayix,'',0,0,$stringforload);
    }
	if ($DatabaseBreak eq 'hour')
	{
		my $stringforload='all';
		my $monthix=sprintf("%02s",$MonthRequired);
		my $dayix=sprintf("%02s",$DayRequired);
		my $hourix=sprintf("%02s",$HourRequired);
		&Read_History_With_TmpUpdate($YearRequired,$monthix,$dayix,$hourix,0,0,$stringforload);
    }

	# HTMLHeadSection
	if(!$PluginMode)
	{
    	&get_html_start();
    }

	# FirstTime LastTime
	my $FirstTime=0;
	my $LastTime=0;
	foreach my $key (keys %FirstTime)
	{
	    my $keyqualified=0;
		if ($MonthRequired eq 'all')                   { $keyqualified=1; }
		if ($key =~ /^$YearRequired$MonthRequired/)    { $keyqualified=1; }
		if ($keyqualified)
		{
		    if ($FirstTime{$key} && ($FirstTime == 0 || $FirstTime > $FirstTime{$key})) { $FirstTime = $FirstTime{$key}; }
		    if ($LastTime < ($LastTime{$key}||0)) { $LastTime = $LastTime{$key}; }
		}
	}

	my $beginmonth=$MonthRequired;my $endmonth=$MonthRequired;
	if ($MonthRequired eq 'all') { $beginmonth=1;$endmonth=12; }
	for (my $month=$beginmonth; $month<=$endmonth; $month++)
	{
		my $monthix=sprintf("%02s",$month);
		$TotalPages+=$MonthPages{$YearRequired.$monthix}||0;
	}

	# Define firstdaytocountaverage, lastdaytocountaverage, firstdaytoshowtime, lastdaytoshowtime
	my $firstdaytocountaverage=$nowyear.$nowmonth."01";				# Set day cursor to 1st day of month
	my $firstdaytoshowtime=$nowyear.$nowmonth."01";					# Set day cursor to 1st day of month
	my $lastdaytocountaverage=$nowyear.$nowmonth.$nowday;			# Set day cursor to today
	my $lastdaytoshowtime=$nowyear.$nowmonth."31";					# Set day cursor to last day of month
	if ($MonthRequired eq 'all')
	{
		$firstdaytocountaverage=$YearRequired."0101";				# Set day cursor to 1st day of the required year
	}
	if (($MonthRequired ne $nowmonth && $MonthRequired ne 'all') || $YearRequired ne $nowyear)
	{
		if ($MonthRequired eq 'all')
		{
			$firstdaytocountaverage=$YearRequired."0101";			# Set day cursor to 1st day of the required year
			$firstdaytoshowtime=$YearRequired."1201";				# Set day cursor to 1st day of last month of required year
			$lastdaytocountaverage=$YearRequired."1231";			# Set day cursor to last day of the required year
			$lastdaytoshowtime=$YearRequired."1231";				# Set day cursor to last day of last month of required year
		} else {
			$firstdaytocountaverage=$YearRequired.$MonthRequired."01";	# Set day cursor to 1st day of the required month
			$firstdaytoshowtime=$YearRequired.$MonthRequired."01";		# Set day cursor to 1st day of the required month
			$lastdaytocountaverage=$YearRequired.$MonthRequired."31";	# Set day cursor to last day of the required month
			$lastdaytoshowtime=$YearRequired.$MonthRequired."31";		# Set day cursor to last day of the required month
		}
	}
	if ($Debug)
	{
		debug("firstdaytocountaverage=$firstdaytocountaverage, lastdaytocountaverage=$lastdaytocountaverage",1);
		debug("firstdaytoshowtime=$firstdaytoshowtime, lastdaytoshowtime=$lastdaytoshowtime",1);
	}

    if (scalar keys %HTMLOutput == 1)
    {
    	if ($HTMLOutput{'urlcloudmap'} )
    	{
    		my $title = '';
    		my $cpt=0;
    		$title=$Message[3];
    		$cpt=(scalar keys %_url_p);
    	    debug("Starting output section, \$cpt=$cpt",1);

		    &BuildKeyListNoSort($MaxRowsInHTMLOutput, $MinHit{'File'}, \%_url_p, \%_url_p);
    		$max_p=1;
    		foreach my $key (@keylist)
    		{
    			if ($_url_p{$key} > $max_p) { $max_p = $_url_p{$key}; }
    		}
			debug("\$max_p=$max_p",1);
			print "\n<div class=\"cloudMapContainer\">";
    		foreach my $key (@keylist)
    		{
    			&ShowCloudMapInfo($_url_p{$key}, $key, $max_p, $_url_title{$key});
    		}
    		print "\n</div>";
            if (! $PluginMode)
            {
                print "\n<H4>$Message[18] $displayCount $Message[28] $cpt $Message[29].</h4>";  # Doubtful if this would be wanted when running in plugInMode
                &get_html_end();
            }
    	}

        my $htmloutput='';
        foreach my $key (keys %HTMLOutput) { $htmloutput=$key; }
    }

}else {
	print "\nJumped lines in file: $lastLineNum";
	if ($lastLineNum) { print "\n Found $lastLineNum already parsed records."; }
	print "\nParsed lines in file: $NbOfLinesParsed";
	print "\n Found $NbOfLinesDropped dropped records,";
	print "\n Found $NbOfLinesCorrupted corrupted records,";
	print "\n Found $NbOfOldLines old records,";
	print "\n Found $NbOfNewLines new qualified records.";
}

#sleep 10;

0;	# Do not remove this line


#-------------------------------------------------------
# ALGORITHM SUMMARY
#
# Read_Config();
# Check_Config() and Init variables
#	&Read_Language_Data($Lang);
# html_head
#
# Get last history file name
# Get value for $LastLine $LastLineNumber $LastLineOffset $LastLineChecksum with
#	&Read_History_With_TmpUpdate(lastyearbeforeupdate,lastmonthbeforeupdate,lastdaybeforeupdate,lasthourbeforeupdate,NOUPDATE,NOPURGE,"general");
#
# &Init_HashArray()
#
# If 'update'
#   Loop on each new line in log file
#     lastlineoffset=lastlineoffsetnext; lastlineoffsetnext=file pointer position
#     If line corrupted, skip --> next on loop
#	  Drop wrong virtual host --> next on loop
#     Drop wrong method/protocol --> next on loop
#     Check date --> next on loop
#     If line older than $LastLine, skip --> next on loop
#     So it's new line
#     $LastLine = time or record
#     Skip if url is /robots.txt --> next on loop
#     Skip line for @SkipHosts --> next on loop
#     Skip line for @SkipFiles --> next on loop
#     Skip line for @SkipUserAgent --> next on loop
#     Skip line for not @OnlyHosts --> next on loop
#     Skip line for not @OnlyFiles --> next on loop
#     Skip line for not @OnlyUserAgent --> next on loop
#     So it's new line approved
#     If other month/year, create/update tmp file and purge data arrays with
#       &Read_History_With_TmpUpdate(lastprocessedyear,lastprocessedmonth,lastprocessedday,lastprocessedhour,UPDATE,PURGE,"all",lastlinenb,lastlineoffset,CheckSum($_));
#     Define a clean Url and Query (set urlwithnoquery, tokenquery and standalonequery and $field[$pos_url])
#     Define PageBool and extension
#     Analyze: Misc tracker --> complete %misc
#     Analyze: Add to favorites --> complete %_misc, countedtraffic=1 (not counted anywhere)
#     Analyze: Cluster
#     Analyze: Extra (must be after 'Define a clean Url and Query')
#     If too many records, we flush data arrays with
#       &Read_History_With_TmpUpdate(lastprocessedyear,lastprocessedmonth,lastprocessedday,lastprocessedhour,UPDATE,PURGE,"all",lastlinenb,lastlineoffset,CheckSum($_));
#   End of loop
#
#   Create/update tmp file
#	  Seek to lastlineoffset in logfile to read and get last line into $_
#	  &Read_History_With_TmpUpdate(lastprocessedyear,lastprocessedmonth,lastprocessedday,lastprocessedhour,UPDATE,PURGE,"all",lastlinenb,lastlineoffset,CheckSum($_))
#   Rename all created tmp files
# End of 'update'
#
# &Init_HashArray()
#
# If 'output'
#   Loop for each month of required year
#     &Read_History_With_TmpUpdate($YearRequired,$monthloop,'','',NOUPDATE,NOPURGE,'all' or 'general time' if not required month)
#   End of loop
#   Show data arrays in HTML page
#   get_html_end
# End of 'output'
#-------------------------------------------------------

#-------------------------------------------------------
# IP Format (d=decimal on 16 bits, x=hexadecimal on 16 bits)
#
# 13.1.68.3						IPv4 (d.d.d.d)
# 0:0:0:0:0:0:13.1.68.3 		IPv6 (x:x:x:x:x:x:d.d.d.d)
# ::13.1.68.3
# 0:0:0:0:0:FFFF:13.1.68.3 		IPv6 (x:x:x:x:x:x:d.d.d.d)
# ::FFFF:13.1.68.3 				IPv6
#
# 1070:0:0:0:0:800:200C:417B 	IPv6
# 1070:0:0:0:0:800:200C:417B 	IPv6
# 1070::800:200C:417B 			IPv6
#-------------------------------------------------------


############### Page Title Retrieval Section ###############
# Generate new lines for urlalias file by doing a http get
# Return: String containing three items
#
sub Generate_Alias_List_Entry
{
	# take in the path & document
	my $urlToGet = shift;
	my $urlScore = shift;

	my $AliasLine = "";
	my $urlPrefix = "http://";
       $pageTitle = "";

	if($useHTTPS)
	{
		$urlPrefix = "https://";
	}

	$AliasLine  = $urlToGet;
	$AliasLine .= "\t";

	# build a full HTTP request to pass to user agent
	my $fullURLToGet = $urlPrefix . $TestDomain . $urlToGet;

	# Create a HTTP request
	print("\nGetting page: '$fullURLToGet'\t{\$urlScore = $urlScore}");
	my $req = new HTTP::Request GET => $fullURLToGet;

	# Pass request to the user agent and get a response back
	my $res = $ua->request($req);

	# Parse returned document for page title
	if ($res->is_success())
	{
		$pageTitle = $res->title;
   		debug ("Got page $pageTitle OK, code:".$res->status_line."\n", 1);
        foreach (@RemoveFromTitle)
        {
            $pageTitle =~ s/$_//;
    		debug("\$pageTitle = $pageTitle \t \$_ = $_ \n", 1);
        }
	} else {
		print "\nFailed to get page, code:".$res->status_line;
		$pageTitle = $MISSING_TITLE_STRING;
	}
	if ($pageTitle eq "")
	{
		$pageTitle = $EMPTY_TITLE_STRING;
	}
	return $AliasLine . $urlScore . "\t" . $pageTitle;
}

##############################
sub processLogs
{
	my $isThisAFile = open(LOG,"$LogFile");
	my $nextLogFile;
    debug("sub processLogs: \$isThisAFile = '$isThisAFile'");

    if($isThisAFile)
    {
        debug("Log file '$LogFile' opened successfully.", 1);
        close LOG;
        processLogFile($LogFile);
    }else{
        if($RecursiveDirs)
        {
            debug("Unable to open $LogFile as a file - doing recursive folder processing instead.", 1);
            print(" - doing recursive folder processing.");
            find(\&processLogFile, $LogFile);       # Recursive processing
        }else{
            debug("Unable to open $LogFile as a file - doing non-recursive folder processing instead.", 1);
            print(" - doing non-recursive folder processing.");
            opendir(BIN, $LogFile) or return;  # Don't want to 'DIE' here - may be able to read other files/folders
            while( defined ($nextLogFile = readdir BIN) )
            {
                next if $nextLogFile =~ /^\.\.?$/;   # ignore the special files:  .  and ..
                if(-f $LogFile."\\".$nextLogFile)
                {
                    processLogFile($nextLogFile);
                }
            }
            closedir BIN
        }
    }

	debug("sub processLogs(".") complete");
}
##############################
sub processLogFile
{
    my $thisFolder;
    my $thisLogFile = shift;
    my $tmp;
    my $ae;
    my $ok;

    if($thisLogFile) # Depends how this function was called.
    {
        print("\nLooking at file: '$LogFile\\$thisLogFile'");
        $thisLogFile = $LogFile."\\".$thisLogFile;
        $thisFolder  = $LogFile;
    }else{
        $thisLogFile = $File::Find::name;
        $thisFolder  = $File::Find::dir;
        print("\nLooking at file: '$_'\n      in folder: '$thisFolder'");
    }

    if($thisLogFile =~ m/\.$/)
    {
        print("\nIgnoring file-system artifact: '$thisLogFile'");
        return;   # ignore these file-system artifacts whatever else is set.
    }
    if(-d)
    {
        print("\nIgnoring folder: '$thisFolder'");
        return;   # $thisLogFile is actually a folder, so ignore it.
    }

    #Check if we have a compressed file:
    if($thisLogFile =~ m/.gz$/)
    {
        print("\n'$thisLogFile' appears to be a gzip archive.");
        $tmp = File::Temp->new();
        $ae = Archive::Extract->new( archive => $thisLogFile );
### ### $ae = Archive::Extract->new(archive => '/path/to/archive',[type => TYPE])  TYPE only needed to over-ride an inaccurate file-extension
        $ok = $ae->extract( to => $tmp->filename );
        die $ae->error unless $ok;          # Is 'DIE' a little too extreme for here? MJB 17/02/2010 18:20:46
        $thisLogFile = $tmp->filename
    }else{
        print("\n'$thisLogFile' will be treated as a plain-text file.");
    	# Open this log file directly:
    }
	debug("Open log file \"$thisLogFile\"");
	open(LOG,"$thisLogFile") || error("\nCouldn't open server log file: \"$thisLogFile\" : $!");
	binmode LOG;	# Avoid premature EOF due to log files corrupted with \cZ or bin chars

	# Define local variables for loop scan
	my @field=();
	my $Counterforflushtest=0;
	my $qualifdrop='';
	my $countedtraffic=0;
	# Reset chrono for benchmark (first call to GetDelaySinceStart)
	&GetDelaySinceStart(1);
	if (! scalar keys %HTMLOutput) { print "\nPhase 1 : First bypass old records, searching new record..."; }

	# Can we try a direct seek access in log ?
	my $line;
	if ($LastLine && $LastLineNumber && $LastLineOffset && $LastLineChecksum)
	{
		# Try a direct seek access to save time
		debug("Try a direct access to LastLine=$LastLine, LastLineNumber=$LastLineNumber, LastLineOffset=$LastLineOffset, LastLineChecksum=$LastLineChecksum");
		seek(LOG,$LastLineOffset,0);
		if ($line=<LOG>)
		{
			chomp $line; $line =~ s/\r$//;
			@field=map(/$PerlParsingFormat/,$line);
			if ($Debug)
			{
				my $string='';
				foreach (0..@field-1) {	$string.="$fieldlib[$_]=$field[$_] "; }
				debug(" Read line after direct access: $string",1);
			}
			my $checksum=&CheckSum($line);
			debug(" LastLineChecksum=$LastLineChecksum, Read line checksum=$checksum",1);
			if ($checksum == $LastLineChecksum )
			{
				if (! scalar keys %HTMLOutput) { print "\nDirect access after last parsed record (after line $LastLineNumber)"; }
				$lastLineNum=$LastLineNumber;
				$lastlineoffset=$LastLineOffset;
				$lastlineoffsetnext=tell LOG;
                $NewLinePhase=1;
			} else {
				if (! scalar keys %HTMLOutput) { print "\nDirect access to last remembered record has fallen on another record.\nSo searching new records from beginning of log file..."; }
				$lastLineNum=0;
				$lastlineoffset=0;
				$lastlineoffsetnext=0;
				seek(LOG,0,0);
			}
		} else {
			if (! scalar keys %HTMLOutput) { print "\nDirect access to last remembered record is out of file.\nSo searching it from beginning of log file..."; }
			$lastLineNum=0;
			$lastlineoffset=0;
			$lastlineoffsetnext=0;
			seek(LOG,0,0);
		}
	} else {
		# No try of direct seek access
		if (! scalar keys %HTMLOutput) { print "\nSearching new records from beginning of log file..."; }
		$lastLineNum=0;
		$lastlineoffset=0;
		$lastlineoffsetnext=0;
	}

	# Loop on each log line:
	while ($line=<LOG>)
	{
		chomp $line; $line =~ s/\r$//;
		if ($UpdateFor && $NbOfLinesParsed >= $UpdateFor)
		{
            debug("PPPPP About to skip out of this block. ", 1);
		    last;
	    }
		$NbOfLinesParsed++;

 		$lastlineoffset=$lastlineoffsetnext; $lastlineoffsetnext=tell LOG;

		if ($ShowSteps)
		{
			if ((++$NbOfLinesShowsteps & $NB_OF_LINES_FOR_BENCHMARK) == 0)
			{
				my $delay=&GetDelaySinceStart(0);
				print "\n$NbOfLinesParsed lines processed (".($delay>0?$delay:1000)." ms, ".int(1000*$NbOfLinesShowsteps/($delay>0?$delay:1000))." lines/second)\n";
			}
		}

        if ($LogFormat eq '2' && $line =~ /^#Fields:/)
        {
            my @fixField=map(/^#Fields: (.*)/, $line);
            if ($fixField[0] !~ /s-kernel-time/) {
                debug("Found new log format: '". $fixField[0] ."'",1);
                &DefinePerlParsingFormat($fixField[0]);
            }
        }

		# Parse line record to get all required fields
		if (! (@field=map(/$PerlParsingFormat/,$line)))
		{
			$NbOfLinesCorrupted++;
			if ($ShowCorrupted) {
				if ($line =~ /^#/ || $line =~ /^!/) { print "\nCorrupted record line ".($lastLineNum+$NbOfLinesParsed)." (comment line): $line"; }
				elsif ($line =~ /^\s*$/) { print "\nCorrupted record line ".($lastLineNum+$NbOfLinesParsed)." (blank line)"; }
				else { print "\nCorrupted record line ".($lastLineNum+$NbOfLinesParsed)." (record format does not match LogFormat parameter): $line"; }
			}
			if ($NbOfLinesParsed >= $NbOfLinesForCorruptedLog && $NbOfLinesParsed == $NbOfLinesCorrupted) { error("Format error",$line,$thisLogFile); }	# Exit with format error
			if ($line =~ /^__end_of_file__/i) { last; }	# For test purpose only
			next;
		}

		if ($Debug)
		{
			my $string='';
			foreach (0..@field-1) {	$string.="$fieldlib[$_]=$field[$_] "; }
			debug(" Correct format line ".($lastLineNum+$NbOfLinesParsed).": $string",1);
		}

		# Drop wrong virtual host name
		#----------------------------------------------------------------------
		if ($pos_vh>=0 && $field[$pos_vh] !~ /^$SiteDomain$/i)
		{
			my $skip=1;
			foreach (@HostAliases)
			{
				if ($field[$pos_vh] =~ /$_/) { $skip=0; last; }
			}
			if ($skip)
			{
				$NbOfLinesDropped++;
				if ($ShowDropped) { print "\nDropped record (virtual hostname '$field[$pos_vh]' does not match SiteDomain='$SiteDomain' nor HostAliases parameters): $line"; }
				next;
			}
		}

		# Drop wrong method/protocol
		#---------------------------
		if ($LogType ne 'M') { $field[$pos_url] =~ s/\s/%20/g; }
		if ($LogType eq 'W'
		    && ($field[$pos_method] eq 'GET'
			|| $field[$pos_method] eq 'POST'
			|| $field[$pos_method] eq 'HEAD'
			|| $field[$pos_method] eq 'PROPFIND'
			|| $field[$pos_method] eq 'CHECKOUT'
			|| $field[$pos_method] eq 'LOCK'
			|| $field[$pos_method] eq 'PROPPATCH'
			|| $field[$pos_method] eq 'OPTIONS'
			|| $field[$pos_method] eq 'MKACTIVITY'
			|| $field[$pos_method] eq 'PUT'
			|| $field[$pos_method] eq 'MERGE'
			|| $field[$pos_method] eq 'DELETE'
			|| $field[$pos_method] eq 'REPORT'
			|| $field[$pos_method] eq 'MKCOL'
			|| $field[$pos_method] eq 'COPY'
			|| $field[$pos_method] =~ /OK/i
			|| $field[$pos_method] =~ /ERR\!/i))
		{
			# HTTP request.	Keep only GET, POST, HEAD, *OK* and ERR! for Webstar. Do not keep OPTIONS, TRACE
		} elsif (($LogType eq 'W' || $LogType eq 'S') && ($field[$pos_method] eq 'GET' || $field[$pos_method] eq 'mms' || $field[$pos_method] eq 'rtsp' || $field[$pos_method] eq 'http' || $field[$pos_method] eq 'RTP')) {
			# Streaming request (windows media server, realmedia or darwin streaming server)
		} elsif ($LogType eq 'F' && ($field[$pos_method] eq 'RETR' || $field[$pos_method] eq 'o' || $field[$pos_method] =~ /$regget/o)) {
			# FTP GET request
		} elsif ($LogType eq 'F' && ($field[$pos_method] eq 'STOR' || $field[$pos_method] eq 'i' || $field[$pos_method] =~ /$regsent/o)) {
			# FTP SENT request
		} elsif ($LogType eq 'C' ) {
			# Accept anything for a CloudMap log format - assume that this has already been sanitised
		} else {
			$NbOfLinesDropped++;
			if ($ShowDropped) { print "\nDropped record (method/protocol '$field[$pos_method]' not qualified when LogType=$LogType): $line"; }
			next;
		}

		$field[$pos_date] =~ tr/,-\/ \t/:::::/s;		# " \t" is used instead of "\s" not known with tr
		my @dateparts=split(/:/,$field[$pos_date]);		# tr and split faster than @dateparts=split(/[\/\-:\s]/,$field[$pos_date])
		# Detected date format: dddddddddd, YYYY-MM-DD HH:MM:SS (IIS), MM/DD/YY\tHH:MM:SS,
		# DD/Month/YYYY:HH:MM:SS (Apache), DD/MM/YYYY HH:MM:SS, Mon DD HH:MM:SS
		if (! $dateparts[1])
		{
			# Unix timestamp
			($dateparts[5],$dateparts[4],$dateparts[3],$dateparts[0],$dateparts[1],$dateparts[2]) = localtime(int($field[$pos_date]));
			$dateparts[1]++;$dateparts[2]+=1900;
		}
		elsif ($dateparts[0] =~ /^....$/) { my $tmp=$dateparts[0]; $dateparts[0]=$dateparts[2]; $dateparts[2]=$tmp; }
		elsif ($field[$pos_date] =~ /^..:..:..:/) { $dateparts[2]+=2000; my $tmp=$dateparts[0]; $dateparts[0]=$dateparts[1]; $dateparts[1]=$tmp; }
		elsif ($dateparts[0] =~ /^...$/)  { my $tmp=$dateparts[0]; $dateparts[0]=$dateparts[1]; $dateparts[1]=$tmp; $tmp=$dateparts[5]; $dateparts[5]=$dateparts[4]; $dateparts[4]=$dateparts[3]; $dateparts[3]=$dateparts[2]; $dateparts[2]=$tmp||$nowyear; }
		if (exists($MonthNum{$dateparts[1]})) { $dateparts[1]=$MonthNum{$dateparts[1]}; }	# Change lib month in num month if necessary
        if ($dateparts[1] <= 0)
        {
            # Date corrupted (for example $dateparts[1]='dic' for december month in a spanish log file)
			$NbOfLinesCorrupted++;
			if ($ShowCorrupted) { print "\nCorrupted record line ".($lastLineNum+$NbOfLinesParsed)." (bad date format for month, may be month are not in english ?): $line"; }
            next;
        }

		# Now @dateparts is (DD,MM,YYYY,HH,MM,SS) and we're going to create $timeRecord=YYYYMMDDHHMMSS
		my $yearrecord=int($dateparts[2]);
		my $monthrecord=int($dateparts[1]);
		my $dayrecord=int($dateparts[0]);
		my $hourrecord=int($dateparts[3]);
		my $daterecord='';
		if    ($DatabaseBreak eq 'month') { $daterecord=sprintf("%04i%02i",$yearrecord,$monthrecord); }
		elsif ($DatabaseBreak eq 'year')  { $daterecord=sprintf("%04i%",$yearrecord); }
		elsif ($DatabaseBreak eq 'day')   { $daterecord=sprintf("%04i%02i%02i",$yearrecord,$monthrecord,$dayrecord); }
		elsif ($DatabaseBreak eq 'hour')  { $daterecord=sprintf("%04i%02i%02i%02i",$yearrecord,$monthrecord,$dayrecord,$hourrecord); }
		my $yearMonthDayRecord=sprintf("$dateparts[2]%02i%02i",$dateparts[1],$dateparts[0]);
		my $timeRecord=((int("$yearMonthDayRecord")*100+$dateparts[3])*100+$dateparts[4])*100+$dateparts[5];

		# Check date
		#-----------------------
		if ($LogType eq 'M' && $timeRecord > $tomorrowtime)
		{
			# Postfix/Sendmail does not store year, so we assume that year is year-1 if record is in future
			$yearrecord--;
			if    ($DatabaseBreak eq 'month') { $daterecord=sprintf("%04i%02i",$yearrecord,$monthrecord); }
			elsif ($DatabaseBreak eq 'year')  { $daterecord=sprintf("%04i%",$yearrecord); }
			elsif ($DatabaseBreak eq 'day')   { $daterecord=sprintf("%04i%02i%02i",$yearrecord,$monthrecord,$dayrecord); }
			elsif ($DatabaseBreak eq 'hour')  { $daterecord=sprintf("%04i%02i%02i%02i",$yearrecord,$monthrecord,$dayrecord,$hourrecord); }
			$yearMonthDayRecord=sprintf("$yearrecord%02i%02i",$dateparts[1],$dateparts[0]);
			$timeRecord=((int("$yearMonthDayRecord")*100+$dateparts[3])*100+$dateparts[4])*100+$dateparts[5];
		}
		if ($timeRecord < 10000000000000 || $timeRecord > $tomorrowtime)
		{
			$NbOfLinesCorrupted++;
			if ($ShowCorrupted) { print "\nCorrupted record (invalid date, timeRecord=$timeRecord): $line"; }
			next;		# Should not happen, kept in case of parasite/corrupted line
		}
		if ($NewLinePhase)
		{
			# TODO NOT_SORTED_RECORD_TOLERANCE does not work around midnight
			if ($timeRecord < ($LastLine - $NOT_SORTED_RECORD_TOLERANCE))
			{
				# Should not happen, kept in case of parasite/corrupted old line
				$NbOfLinesCorrupted++;
				if ($ShowCorrupted) { print "\nCorrupted record (date $timeRecord lower than $LastLine-$NOT_SORTED_RECORD_TOLERANCE): $line"; } next;
			}
		} else {
			if ($timeRecord <= $LastLine)
			{
			    # Already processed
				$NbOfOldLines++;
				next;
			}
			# We found a new line. This will replace comparison "<=" with "<" between timeRecord and LastLine (we should have only new lines now)
			$NewLinePhase=1;	# We will never enter here again
			if ($ShowSteps)
			{
				if ($NbOfLinesShowsteps > 1 && ($NbOfLinesShowsteps & $NB_OF_LINES_FOR_BENCHMARK))
				{
					my $delay=&GetDelaySinceStart(0);
					print "\n".($NbOfLinesParsed-1)." lines processed (".($delay>0?$delay:1000)." ms, ".int(1000*($NbOfLinesShowsteps-1)/($delay>0?$delay:1000))." lines/second)";
				}
				&GetDelaySinceStart(1);	$NbOfLinesShowsteps=1;
			}
			if (! scalar keys %HTMLOutput)
			{
				print "\nPhase 2 : Now process new records (Flush history on disk after ".($LIMIT_FLUSH<<2)." hosts)...";
			}
		}

		# Convert URL for Webstar to common URL
		if ($LogFormat eq '3')
		{
			$field[$pos_url]=~s/:/\//g;
			if ($field[$pos_code] eq '-') { $field[$pos_code]='200'; }
		}

		# Here, field array, timeRecord and yearMonthDayRecord are initialized for log record
		debug("  This is a not already processed record ($timeRecord)",4);

		# We found a new line
		#----------------------------------------
		if ($timeRecord > $LastLine) { $LastLine = $timeRecord; }	# Test should always be true except with not sorted log files

		# Skip for some client host IP addresses, some URLs, other URLs
		if    (@SkipHosts && (&SkipHost($field[$pos_host]) || ($pos_hostr && &SkipHost($field[$pos_hostr]))))   { $qualifdrop="Dropped record (host $field[$pos_host]".($pos_hostr?" and $field[$pos_hostr]":"")." not qualified by SkipHosts)"; }
		elsif (@SkipFiles && &SkipFile($field[$pos_url]))    { $qualifdrop="Dropped record (URL $field[$pos_url] not qualified by SkipFiles)"; }
		elsif (@SkipUserAgents && $pos_agent >= 0 && &SkipUserAgent($field[$pos_agent]))	{ $qualifdrop="Dropped record (user agent '$field[$pos_agent]' not qualified by SkipUserAgents)"; }
		elsif (@SkipReferrers && $pos_referer >= 0 && &SkipReferrer($field[$pos_referer]))    { $qualifdrop="Dropped record (URL $field[$pos_referer] not qualified by SkipReferrers)"; }
		elsif (@OnlyHosts && ! &OnlyHost($field[$pos_host]) && (! $pos_hostr || ! &OnlyHost($field[$pos_hostr]))) { $qualifdrop="Dropped record (host $field[$pos_host]".($pos_hostr?" and $field[$pos_hostr]":"")." not qualified by OnlyHosts)"; }
		elsif (@OnlyFiles && ! &OnlyFile($field[$pos_url]))  { $qualifdrop="Dropped record (URL $field[$pos_url] not qualified by OnlyFiles)"; }
		elsif (@OnlyUserAgents && ! &OnlyUserAgent($field[$pos_agent]))  { $qualifdrop="Dropped record (user agent '$field[$pos_agent]' not qualified by OnlyUserAgents)"; }
		if ($qualifdrop)
		{
			$NbOfLinesDropped++;
			if ($Debug)       { debug("$qualifdrop: $line",1); }
			if ($ShowDropped) { print "\n$qualifdrop: $line"; }
			$qualifdrop='';
			next;
		}

		# Is it in a new break section ?
		#-------------------------------
		if ($daterecord > $lastprocesseddate)
		{
			# A new break to process
			if ($lastprocesseddate > 0)
			{
				# We save data of previous break
				&Read_History_With_TmpUpdate($lastprocessedyear,$lastprocessedmonth,$lastprocessedday,$lastprocessedhour,1,1,"all",($lastLineNum+$NbOfLinesParsed),$lastlineoffset,&CheckSum($line));
				$Counterforflushtest=0;	# We reset counterforflushtest
			}
			$lastprocessedyear=$yearrecord;
			$lastprocessedmonth=$monthrecord;
			$lastprocessedday=$dayrecord;
			$lastprocessedhour=$hourrecord;
			if    ($DatabaseBreak eq 'month') { $lastprocesseddate=sprintf("%04i%02i",$yearrecord,$monthrecord); }
			elsif ($DatabaseBreak eq 'year')  { $lastprocesseddate=sprintf("%04i%",$yearrecord); }
			elsif ($DatabaseBreak eq 'day')   { $lastprocesseddate=sprintf("%04i%02i%02i",$yearrecord,$monthrecord,$dayrecord); }
			elsif ($DatabaseBreak eq 'hour')  { $lastprocesseddate=sprintf("%04i%02i%02i%02i",$yearrecord,$monthrecord,$dayrecord,$hourrecord); }
		}

		$countedtraffic=0;
		$NbOfNewLines++;

		# Define a clean target URL and referrer URL
		# We keep a clean $field[$pos_url] and
		# we store original value for urlwithnoquery, tokenquery and standalonequery
		#---------------------------------------------------------------------------
		if ($URLNotCaseSensitive) { $field[$pos_url]=lc($field[$pos_url]); }
		# Possible URL syntax for $field[$pos_url]: /mydir/mypage.ext?param1=x&param2=y#aaa, /mydir/mypage.ext#aaa, /
		my $urlwithnoquery; my $tokenquery; my $standalonequery; my $anchor='';
		if ($field[$pos_url] =~ s/$regtruncanchor//o) { $anchor=$1; }	# Remove and save anchor
		if ($URLWithQuery)
		{
			$urlwithnoquery=$field[$pos_url];
			my $foundparam=($urlwithnoquery =~ s/$regtruncurl//o);
			$tokenquery=$1||'';
			$standalonequery=$2||'';
			# For IIS setup, if pos_query is enabled we need to combine the URL to query strings
			if (! $foundparam && $pos_query >=0 && $field[$pos_query] && $field[$pos_query] ne '-')
			{
				$foundparam=1;
				$tokenquery='?';
				$standalonequery=$field[$pos_query];
				# Define query
				$field[$pos_url] .= '?'.$field[$pos_query];
			}
 			if ($foundparam)
 			{
				# Keep only params that are defined in URLWithQueryWithOnlyFollowingParameters
				my $newstandalonequery='';
				if (@URLWithQueryWithOnly)
				{
					foreach (@URLWithQueryWithOnly)
					{
						foreach my $p (split(/&/,$standalonequery))
						{
							if ($URLNotCaseSensitive) { if ($p =~ /^$_=/i) { $newstandalonequery.="$p&"; last; } }
							else { if ($p =~ /^$_=/) { $newstandalonequery.="$p&"; last; } }
						}
					}
	 				chop $newstandalonequery;
		 		} elsif (@URLWithQueryWithout) {
    	 			# Remove params that are marked to be ignored in URLWithQueryWithoutFollowingParameters
					foreach my $p (split(/&/,$standalonequery))
					{
						my $found=0;
						foreach (@URLWithQueryWithout)
						{
							#debug("  Check if '$_=' is param '$p' to remove it from query",5);
							if ($URLNotCaseSensitive) { if ($p =~ /^$_=/i) { $found=1; last; } }
							else { if ($p =~ /^$_=/) { $found=1; last; } }
						}
						if (! $found) { $newstandalonequery.="$p&"; }
					}
	 				chop $newstandalonequery;
				} else {
				    $newstandalonequery=$standalonequery;
				}
				# Define query
				$field[$pos_url]=$urlwithnoquery;
				if ($newstandalonequery) { $field[$pos_url].="$tokenquery$newstandalonequery"; }
			}
		} else {
			# Trunc parameters of URL
			$field[$pos_url] =~ s/$regtruncurl//o;
			$urlwithnoquery=$field[$pos_url];
			$tokenquery=$1||'';
			$standalonequery=$2||'';
			# For IIS setup, if pos_query is enabled we need to use it for query strings
			if ($pos_query >=0 && $field[$pos_query] && $field[$pos_query] ne '-')
			{
				$tokenquery='?';
				$standalonequery=$field[$pos_query];
			}
		}
		if ($URLWithAnchor && $anchor) { $field[$pos_url].="#$anchor"; }	# Restore anchor
		# Here now urlwithnoquery is /mydir/mypage.ext, /mydir, /, /page#XXX
		# Here now tokenquery is '' or '?' or ';'
		# Here now standalonequery is '' or 'param1=x'

		# Define page and extension
		#--------------------------
		my $PageBool=1;
		# Extension
		my $extension;
		if ($urlwithnoquery =~ /$regext/o || ($urlwithnoquery =~ /[\\\/]$/ && $DefaultFile[0] =~ /$regext/o))
		{
			$extension=($LevelForFileTypesDetection>=2)?lc($1):'Unknown';
			if ($NotPageList{$extension}) { $PageBool=0; }
		} else {
			$extension='Unknown';
		}

		if (@NotPageFiles && &NotPageFile($field[$pos_url])) { $PageBool=0; }




		# Analyze: Status code (countedtraffic=>3 if error)
		#--------------------------------------------------
		if (! $countedtraffic)
		{
			if ($LogType eq 'W' || $LogType eq 'S')
			{
			    # HTTP record or Stream record
				if ($ValidHTTPCodes{$field[$pos_code]})
				{
				    # Code is valid
					if ($field[$pos_code] == 304) { $field[$pos_size]=0; }
				} else {
				    # Code is not valid
				}
			} elsif ($LogType eq 'F') {
			    # FTP record
			}

			# This block is essential !!! ;)  MJB
			# Analyze: Date - Hour - Pages - Hits - Kilo
			#-------------------------------------------
			if ($PageBool)
			{
				# Replace default page name with / only ('if' is to increase speed when only 1 value in @DefaultFile)
				if (@DefaultFile > 1) { foreach my $elem (@DefaultFile) { if ($field[$pos_url] =~ s/\/$elem$/\//o) { last; } } }
				else { $field[$pos_url] =~ s/$regdefault/\//o; }
				# FirstTime and LastTime are First and Last human visits (so changed if access to a page)
				$FirstTime{$lastprocesseddate}||=$timeRecord;
				$LastTime{$lastprocesseddate}=$timeRecord;
				$DayPages{$yearMonthDayRecord}++;
				$_url_p{$field[$pos_url]}++; 										#Count accesses for page (page)
				# $_url_k{$field[$pos_url]}+=int($field[$pos_size]);
				$_time_p[$hourrecord]++;											#Count accesses for hour (page)
			}
			$_time_h[$hourrecord]++;
			$_time_k[$hourrecord]+=int($field[$pos_size]);
	 		$DayHits{$yearMonthDayRecord}++;						#Count accesses for hour (hit)
	 		$DayBytes{$yearMonthDayRecord}+=int($field[$pos_size]);	#Count accesses for hour (kb)
		}





		my $Host=$field[$pos_host];
		my $HostResolved='';


		# Check cluster
		#--------------
		if ($pos_cluster>=0)
		{
			if ($PageBool) { $_cluster_p{$field[$pos_cluster]}++; }				#Count accesses for page (page)
			$_cluster_h{$field[$pos_cluster]}++;								#Count accesses for page (hit)
			##Del: $_cluster_k{$field[$pos_cluster]}+=int($field[$pos_size]);			#Count accesses for page (kb)
		}

		# Analyze: Extra   ##Del ??
		#---------------
 		foreach my $extranum (1..@ExtraName-1)
 		{
			debug("  Process extra analyze $extranum",1);

 			# Check code
 			my $conditionok=0;
 			if ($ExtraCodeFilter[$extranum])
 			{
	 			foreach my $condnum (0..@{$ExtraCodeFilter[$extranum]}-1) {
					debug("  Check code '$field[$pos_code]' must be '$ExtraCodeFilter[$extranum][$condnum]'",5);
					if ($field[$pos_code] eq "$ExtraCodeFilter[$extranum][$condnum]") { $conditionok=1; last; }
				}
				if (! $conditionok && @{$ExtraCodeFilter[$extranum]}) { next; }	# End for this section
				debug("  No check on code or code is OK. Now we check other conditions.",5);
			}

 			# Check conditions
 			$conditionok=0;
 			foreach my $condnum (0..@{$ExtraConditionType[$extranum]}-1) {
 				my $conditiontype=$ExtraConditionType[$extranum][$condnum];
 				my $conditiontypeval=$ExtraConditionTypeVal[$extranum][$condnum];
 				if ($conditiontype eq 'URL') {
					debug("  Check condition '$conditiontype' must contain '$conditiontypeval' in '$urlwithnoquery'",5);
 					if ($urlwithnoquery =~ /$conditiontypeval/) { $conditionok=1; last; }
 				}
 				elsif ($conditiontype eq 'QUERY_STRING') {
					debug("  Check condition '$conditiontype' must contain '$conditiontypeval' in '$standalonequery'",5);
 					if ($standalonequery =~ /$conditiontypeval/) {	$conditionok=1; last; }
 				}
                elsif ($conditiontype eq 'URLWITHQUERY') {
                    debug("  Check condition '$conditiontype' must contain '$conditiontypeval' in '$urlwithnoquery$tokenquery$standalonequery'",5);
                    if ("$urlwithnoquery$tokenquery$standalonequery" =~ /$conditiontypeval/) { $conditionok=1; last; }
                }
 				elsif ($conditiontype eq 'REFERER') {
					debug("  Check condition '$conditiontype' must contain '$conditiontypeval' in '$field[$pos_referer]'",5);
 					if ($field[$pos_referer] =~ /$conditiontypeval/) { $conditionok=1; last; }
 				}
 				elsif ($conditiontype eq 'UA') {
					debug("  Check condition '$conditiontype' must contain '$conditiontypeval' in '$field[$pos_agent]'",5);
 					if ($field[$pos_agent] =~ /$conditiontypeval/) { $conditionok=1; last; }
 				}
 				elsif ($conditiontype eq 'HOST') {
					debug("  Check condition '$conditiontype' must contain '$conditiontypeval' in '$field[$pos_host]'",5);
 					if ($HostResolved =~ /$conditiontypeval/) { $conditionok=1; last; }
 				}
				elsif ($conditiontype eq 'VHOST') {
					debug("  Check condision '$conditiontype' must contain '$conditiontypeval' in '$field[$pos_vh]'",5);
 					if ($field[$pos_vh] =~ /$conditiontypeval/) { $conditionok=1; last; }
				}
 				elsif ($conditiontype =~ /extra(\d+)/i) {
					debug("  Check condition '$conditiontype' must contain '$conditiontypeval' in '$field[$pos_extra[$1]]'",5);
 					if ($field[$pos_extra[$1]] =~ /$conditiontypeval/) { $conditionok=1; last; }
                }
 				else { error("Wrong value of parameter ExtraSectionCondition$extranum"); }
 			}
			if (! $conditionok && @{$ExtraConditionType[$extranum]}) { next; }	# End for this section
			debug("  No condition or condition is OK. Now we extract value for first column of extra chart.",5);

 			# Determine actual column value to use.
 			my $rowkeyval;
			my $rowkeyok=0;
 			foreach my $rowkeynum (0..@{$ExtraFirstColumnValuesType[$extranum]}-1) {
 				my $rowkeytype=$ExtraFirstColumnValuesType[$extranum][$rowkeynum];
 				my $rowkeytypeval=$ExtraFirstColumnValuesTypeVal[$extranum][$rowkeynum];
				if ($rowkeytype eq 'URL') {
					if ($urlwithnoquery =~ /$rowkeytypeval/) { $rowkeyval = "$1"; $rowkeyok = 1; last; }
				}
 				elsif ($rowkeytype eq 'QUERY_STRING') {
					debug("  Extract value from '$standalonequery' with regex '$rowkeytypeval'.",5);
 					if ($standalonequery =~ /$rowkeytypeval/) { $rowkeyval = "$1"; $rowkeyok = 1; last; }
 				}
                elsif ($rowkeytype eq 'URLWITHQUERY') {
                    if ("$urlwithnoquery$tokenquery$standalonequery" =~ /$rowkeytypeval/) { $rowkeyval = "$1"; $rowkeyok = 1; last; }
                }
 				elsif ($rowkeytype eq 'REFERER') {
 					if ($field[$pos_referer] =~ /$rowkeytypeval/) { $rowkeyval = "$1"; $rowkeyok = 1; last; }
 				}
 				elsif ($rowkeytype eq 'UA') {
 					if ($field[$pos_agent] =~ /$rowkeytypeval/) { $rowkeyval = "$1"; $rowkeyok = 1; last; }
 				}
 				elsif ($rowkeytype eq 'HOST') {
 					if ($HostResolved =~ /$rowkeytypeval/) { $rowkeyval = "$1"; $rowkeyok = 1; last; }
 				}
 				elsif ($rowkeytype eq 'VHOST') {
 					if ($field[$pos_vh] =~ /$rowkeytypeval/) { $rowkeyval = "$1"; $rowkeyok = 1; last; }
 				}
 				elsif ($rowkeytype =~ /extra(\d+)/i) {
 					if ($field[$pos_extra[$1]] =~ /$rowkeytypeval/) { $rowkeyval = "$1"; $rowkeyok = 1; last; }
 				}
 				else { error("Wrong value of parameter ExtraSectionFirstColumnValues$extranum"); }
 			}
			if (! $rowkeyok) { next; }	# End for this section
			debug("  Key val was found: $rowkeyval",5);

 			# Apply function on $rowkeyval
 			if ($ExtraFirstColumnFunction[$extranum])
 			{
 			    # Todo call function on string $rowkeyval
            }

 			# Here we got all values to increase counters
 			if ($PageBool && $ExtraStatTypes[$extranum] =~ /P/i) { ${'_section_' . $extranum . '_p'}{$rowkeyval}++; }
 			${'_section_' . $extranum . '_h'}{$rowkeyval}++;	# Must be set
 			if ($ExtraStatTypes[$extranum] =~ /B/i) { ${'_section_' . $extranum . '_k'}{$rowkeyval}+=int($field[$pos_size]); }
 			if ($ExtraStatTypes[$extranum] =~ /L/i) {
 				if (${'_section_' . $extranum . '_l'}{$rowkeyval}||0 < $timeRecord) { ${'_section_' . $extranum . '_l'}{$rowkeyval}=$timeRecord; }
 			}
			# Check to avoid too large extra sections
			if (scalar keys %{'_section_' . $extranum . '_h'} > $ExtraTrackedRowsLimit) {
				error(<<END_ERROR_TEXT);
The number of values found for extra section $extranum has grown too large.
In order to prevent AWScloudMap from using an excessive amount of memory, the number
of values is currently limited to $ExtraTrackedRowsLimit. Perhaps you should consider
revising extract parameters for extra section $extranum. If you are certain you
want to track such a large data set, you can increase the limit by setting
ExtraTrackedRowsLimit in your AWScloudMap configuration file.
END_ERROR_TEXT
			}
 		}

		# Every 20,000 approved lines after a flush, test to clean too large hash arrays to flush data in tmp file
		if (++$Counterforflushtest >= 20000)
		{
			if ((scalar keys %_host_u) > ($LIMIT_FLUSH<<2) || (scalar keys %_url_p) > $LIMIT_FLUSH)
			{
				# warning("Warning: Try to run AWScloudMap update process more frequently to analyze smaller log files.");
				if ($^X =~ /activestate/i || $^X =~ /activeperl/i)
				{
					# We don't flush if perl is activestate to avoid slowing process because of memory hole
				} else {
					# Clean tmp hash arrays
					%TmpOS = %TmpRefererServer = %TmpRobot = %TmpBrowser = ();
					# We flush if perl is not activestate
					print "\nFlush history file on disk";
					if ((scalar keys %_host_u) > ($LIMIT_FLUSH<<2)) { print " (unique hosts reach flush limit of ".($LIMIT_FLUSH<<2).")"; }
					if ((scalar keys %_url_p)  > $LIMIT_FLUSH)      { print " (unique url reach flush limit of ".($LIMIT_FLUSH).")"; }
					if ($Debug)
					{
						debug("End of set of $Counterforflushtest records: Some hash arrays are too large. We flush and clean some.",2);
						print "\n _url_p:".(scalar keys %_url_p);
					}
					&Read_History_With_TmpUpdate($lastprocessedyear,$lastprocessedmonth,$lastprocessedday,$lastprocessedhour,1,1,"all",($lastLineNum+$NbOfLinesParsed),$lastlineoffset,&CheckSum($_));
					&GetDelaySinceStart(1);	$NbOfLinesShowsteps=1;
				}
			}
			$Counterforflushtest=0;
		}

	}	# End of loop for processing new record.

	if ($Debug)
	{
		debug(" _url_p:".(scalar keys %_url_p),1);
		debug(" _waithost_e:".(scalar keys %_waithost_e)." _waithost_l:".(scalar keys %_waithost_l)." _waithost_s:".(scalar keys %_waithost_s)." _waithost_u:".(scalar keys %_waithost_u),1);
	}

	# Save current processed break section
	# If lastprocesseddate > 0 means there is at least one approved new record in log or at least one existing history file
	if ($lastprocesseddate > 0) {	# TODO: Do not save if we are sure a flush was just already done
		# Get last line
		seek(LOG,$lastlineoffset,0);
		my $line=<LOG>;
 		chomp $line; $line =~ s/\r$//;
		if (! $NbOfLinesParsed) {
			# TODO If there was no lines parsed (log was empty), we only update LastUpdate line with YYYYMMDDHHMMSS 0 0 0 0 0
			&Read_History_With_TmpUpdate($lastprocessedyear,$lastprocessedmonth,$lastprocessedday,$lastprocessedhour,1,1,"all",($lastLineNum+$NbOfLinesParsed),$lastlineoffset,&CheckSum($line));
		}
		else {
			&Read_History_With_TmpUpdate($lastprocessedyear,$lastprocessedmonth,$lastprocessedday,$lastprocessedhour,1,1,"all",($lastLineNum+$NbOfLinesParsed),$lastlineoffset,&CheckSum($line));
		}
	}

	debug("Close log file \"$thisLogFile\"");
	close LOG || error("Command for pipe '$thisLogFile' failed");

}
############ E_N_D ############